"""
 * @file
 * @author  Tommi Kerola <kerola@student.chalmers.se>
 * @author  Linus Hermansson <hlinus@student.chalmers.se>
 * @version 1.0
 *
 * @section LICENSE
 *
 * Copyright [2013] [Tommi Kerola, Linus Hermansson]
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an "AS
 *  IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 *  express or implied.  See the License for the specific language
 *  governing permissions and limitations under the License.
 *
"""

#!/usr/bin/python

"""
Takes the imdb9403_adjlist.txt file and creates a new one where ambiguous entires are present.
The ambiguity is created by merging entries with the same last name.
"""

import itertools
from collections import defaultdict

# The maximum id of the imdb.name SQL table. The ids start at 1.
max_name_id = 2666147

def get_pairs(vals):
    """
    Tranforms a list of items into a list of pairs.
    [1,2,3,4] --> [(1,2), (3,4)]
    """
    args = [iter(vals)] * 2
    return list(itertools.izip_longest(*args))

class AmbiguityCreator(object):

    def __init__(self, curs, imdb9403_adjlist_filename):
        self.idToMergedId = dict()
        self.mergeIdToLabel = dict()
        self.mergedActors = defaultdict(lambda: defaultdict(int))
        self.curs = curs
        self.distict_adjlist_filename = imdb9403_adjlist_filename

    def assign_merged_ids(self, actor_ids):
            first_id = actor_ids[0]
            if first_id in self.idToMergedId:
                first_id = self.idToMergedId[first_id]  # Reuse already assigned merged ids.
            for actor_id in actor_ids:
                self.idToMergedId[actor_id] = first_id  # Assign all actors with common last name the id of the first actor in the group.

    def get_merged_id(self, actor_id):
        assert actor_id in self.idToMergedId, "Tried to get merged id of actor_id {}, but not found! Assign it first.".format(actor_id)
        return self.idToMergedId[actor_id]

    def get_merged_id_label(self, merged_id):
        assert merged_id in self.mergeIdToLabel, "Tried to get label of merged_id {}, but not found! Assign it first.".format(merged_id)
        return self.mergeIdToLabel[merged_id]

    def get_last_name(self, actor_id):
        query = u"SELECT name FROM name WHERE id = ?"

        self.curs.execute(query, [actor_id])
        row = self.curs.fetchall()[0]
        if row is None:
            return None
        last_name = row['name'].split(",")[0]
        return last_name

    def get_actors_with_same_last_name(self, last_name):
        name_pattern = u"{}, %".format(last_name)
        query = u"SELECT id FROM selected_names WHERE name LIKE ?;"
        self.curs.execute(query, [name_pattern])
        rows = self.curs.fetchall()
        ids = map(lambda x: x['id'], rows)
        return ids

    def create_selected_names_table(self):
        """
        Creates prerequisite cache table for creating the ambiguous names.
        This table should be created BEFORE running the ambiguity creator.
        """
        query = u"INSERT INTO selected_names (id, name) VALUES (?, ?)"

        self.curs.execute("SHOW TABLES LIKE \"selected_names\";")
        found = len(self.curs.fetchall()) > 0
        if not found:
            self.curs.execute("CREATE TABLE selected_names (id int, name text);")
            print "Table selected_names not found. Creating it."
        else:
            print "Previous selected_names table found."

        print "Populating selected_names table in imdb SQL database."
        f = open(self.distict_adjlist_filename, "r")
        for line in f:
            vals = line.split()
            if len(vals) == 1:
                continue    # Only include people who have been in movies with others.
            actor_id = int(vals[0])

            self.curs.execute("SELECT name FROM name WHERE id = ?", [actor_id])
            full_name = curs.fetchall()[0]['name']

            self.curs.execute(query, [actor_id, full_name])

            if actor_id % 1000 == 0:
                print "[{}] {:.2f}%".format(actor_id, 100*actor_id/float(max_name_id))
        f.close()

        # Add index
        self.curs.execute("ALTER TABLE selected_names ADD INDEX id (id), ADD INDEX name (name(6));")

    def preprocess_actor(self, actor_id, pairs):
        """
        Assings a merged_id to an actor_id.
        Also labels the merged_id with how many underlying entities it consists of.

        pairs is a list [(nbr_id, movie_id), (nbr_id, movie_id), ...]
        """

        if not actor_id in self.idToMergedId:
            last_name = self.get_last_name(actor_id)
            same_ids = self.get_actors_with_same_last_name(last_name)
            #assert actor_id in same_ids, "actor_id {} not in same_ids: {}".format(actor_id, same_ids)
            if actor_id in same_ids:
                self.assign_merged_ids(same_ids)
                #print "{}".format(" ".join(map(str, map(self.get_merged_id, same_ids))))
                label = len(same_ids) # label entries by how many underlying entities they contain
                for same_id in same_ids:
                    self.mergeIdToLabel[self.get_merged_id(same_id)] = label
            else:
                return # Strange ID that cannot find itself, just remove these.

        # Add actor to merged line
        self.add_pairs_to_merged_actor(self.get_merged_id(actor_id), pairs)

    def add_pairs_to_merged_actor(self, a_merge_id, pairs):
        for (nbr_id,movie_id) in pairs:
            self.mergedActors[a_merge_id][nbr_id] += 1

    def create_ambiguities(self, outfilename):

        num_bachelor_ids = 0
        num_coocs_skipped = 0
        f = open(self.distict_adjlist_filename, "r")
        print "Stage 1: Preprocess IDs."
        for line in f:
            vals = line.split()
            if len(vals) == 1:
                num_bachelor_ids += 1
                continue    # Only include people who have been in movies with others.
            actor_id = int(vals[0])
            pairs = get_pairs(map(int, vals[1:]))

            self.preprocess_actor(actor_id, pairs)
            if actor_id % 1000 == 0:
                print "[{}] {:.2f}%".format(actor_id, 100*actor_id/float(max_name_id))
        f.close()

        outfile = open(outfilename, "w")
        print "Stage 2: Write merged entries to file."
        for (merge_id, cooccurence_pairs) in self.mergedActors.iteritems():

            label = self.mergeIdToLabel[merge_id]

            cooc_str_list, num_skipped = self.create_cooc_str_list(merge_id, cooccurence_pairs)
            num_coocs_skipped += num_skipped

            if len(cooc_str_list) > 0:
                outfile.write("{} {} {}\n".format(merge_id, label, " ".join(cooc_str_list)))
            if actor_id % 1000 == 0:
                print "[{}] {:.2f}%".format(merge_id, 100*merge_id/float(max_name_id))

        outfile.close()

        print "Ambiguity creation complete!"
        print "num_bachelor_ids: {} ({:.2f}%)".format(num_bachelor_ids, 100*num_bachelor_ids/float(max_name_id))
        print "num_coocs_skipped: {}".format(num_coocs_skipped)

    def create_cooc_str_list(self, merge_id, cooccurence_pairs):
        num_coocs_skipped = 0
        # cooccurence_pairs.iteritems() contains (unique_actor_id, num_coocs)
        cooc_amb_list = defaultdict(int)
        # Fix counts for co-occurence pairs, these do not use the merged id yet.
        for (a,count) in cooccurence_pairs.iteritems():
            if not a in self.idToMergedId:
                #print "Warning: merge_id not found for actor_id {} Skipping it.".format(a)
                num_coocs_skipped += 1
                continue
            a_merge_id = self.get_merged_id(a)
            if merge_id != a_merge_id: # Merging ids can created self-edges, avoid these.
                # count is a count of co-occurences, so just add them to the total for this ambiguous id.
                cooc_amb_list[a_merge_id] += count
        cooc_str_list = []
        for (a_merge_id,count) in cooc_amb_list.iteritems():
            cooc_str_list.append("{} {}".format(a_merge_id, count))

        return cooc_str_list, num_coocs_skipped

class AmbiguityCreatorMalin(AmbiguityCreator):
    """ Creates ambiguous actors for Malin's method. It keeps sources.
    """

    def __init__(self, curs, imdb9403_adjlist_filename):
        super(AmbiguityCreatorMalin, self).__init__(curs, imdb9403_adjlist_filename)
        self.mergedActors = defaultdict(list)

    def add_pairs_to_merged_actor(self, a_merge_id, pairs):
        for pair in pairs:
            self.mergedActors[a_merge_id].append(pair)

    def create_cooc_str_list(self, merge_id, cooccurence_pairs):
        num_coocs_skipped = 0
        # cooccurence_pairs is a list of pairs: [(unique_actor_id, movie_id), ...]
        cooc_amb_list = set()
        # Fix counts for co-occurence pairs, these do not use the merged id yet.
        for (a,movie_id) in cooccurence_pairs:
            if not a in self.idToMergedId:
                #print "Warning: merge_id not found for actor_id {} Skipping it.".format(a)
                num_coocs_skipped += 1
                continue
            a_merge_id = self.get_merged_id(a)
            if merge_id != a_merge_id: # Merging ids can created self-edges, avoid these.
                # movie_id is a movie_id in which a starred, so just replace any occurence of older movie_ids.
                cooc_amb_list.add((a_merge_id, movie_id))
        cooc_str_list = []
        for (a_merge_id,movie_id) in cooc_amb_list:
            cooc_str_list.append("{} {}".format(a_merge_id, movie_id))

        return cooc_str_list, num_coocs_skipped

if __name__ == "__main__":
    import oursql
    import time
    import numpy as np
    import sys

    if len(sys.argv) != 4:
        print "usage: {} [our|malin] imdb9403_adjlist.txt outfilename".format(sys.argv[0])
        sys.exit(1)

    julia_id = 2436318
    start_year = 1994
    end_year = 2003
    min_actors_per_movie = 2

    use_malin = sys.argv[1] == "malin"
    imdb9403_adjlist_filename = sys.argv[2]
    outfilename = sys.argv[3]

    conn = oursql.connect(db="imdb", user="root")
    curs = conn.cursor(oursql.DictCursor)

    if use_malin:
        print "Creating adjlist for Malin's method."
        ac = AmbiguityCreatorMalin(curs, imdb9403_adjlist_filename)
    else: # use our method
        print "Creating adjlist for our method."
        ac = AmbiguityCreator(curs, imdb9403_adjlist_filename)

    ac.create_ambiguities(outfilename)
    #ac.create_selected_names_table()

