"""
 * @file
 * @author  Tommi Kerola <kerola@student.chalmers.se>
 * @author  Linus Hermansson <hlinus@student.chalmers.se>
 * @version 1.0
 *
 * @section LICENSE
 *
 * Copyright [2013] [Tommi Kerola, Linus Hermansson]
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an "AS
 *  IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 *  express or implied.  See the License for the specific language
 *  governing permissions and limitations under the License.
 *
"""


#!/usr/bin/python

import os
import sys
import sklearn.cross_validation as cv
import numpy as np
import time
import random

def write_svmlin_files(semi_labels_tr_filename, semi_features_tr_filename, tr_indices, lines, unlabeled_lines=None):
    """
    Writes two new SVMLIN files:
        - labels
        - features
    """
    f_labels = open(semi_labels_tr_filename, "w")
    f_features = open(semi_features_tr_filename, "w")
    for index in tr_indices:
        f_labels.write("{}\n".format(lines[index][0]))
        f_features.write("{}\n".format(" ".join(lines[index][1:])))
    if not unlabeled_lines is None:
        for line in unlabeled_lines:
            f_labels.write("{}\n".format(line[0]))
            f_features.write("{}\n".format(" ".join(line[1:])))

    f_labels.close()
    f_features.close()

def get_test_acc(semi_outputs_tr_filename, tr_indices, lines):
    """
    Gets the test accuracy.
    """

    f = open(semi_outputs_tr_filename, "r")
    result_lines = map(lambda x: int(np.sign(float(x))), f.read().splitlines())
    f.close()

    tr_lines=[]
    for index in tr_indices:
        tr_lines.append(lines[index])
    correct_lines = map(lambda line: int(line[0]), tr_lines)

    num_correct = map(lambda x: 1 if x[0] == x[1] else 0, zip(result_lines, correct_lines))
    if len(num_correct) != 0:
        acc = sum(num_correct)/float(len(num_correct))
    else:
        acc = 0

    return acc

if __name__ == "__main__":

    if len(sys.argv) != 4:
        print "usage: {} halfkernels_labeled.txt halfkernels_unlabeled.txt unlabeled_fraction".format(sys.argv[0])
        sys.exit(1)

    halfkernels_filename = sys.argv[1]
    halfkernels_unlabeled_filename = sys.argv[2]
    unlabeled_fraction = float(sys.argv[3])

    SVMLIN_PATH="/home/kerola/src/svmlin-v1.0/svmlin"
    random.seed(1234)

    # Take halfkernels.txt split it into two parts.
    # - labels.txt
    # - features.txt

    f = open(halfkernels_filename, "r")
    lines = map(lambda line: line.split(), f.read().splitlines())
    f.close()
    f = open(halfkernels_unlabeled_filename, "r")
    unlabeled_lines = map(lambda line: line.split(), f.read().splitlines())
    orig_num_unlabeled = len(unlabeled_lines)
    unlabeled_lines = unlabeled_lines[:int(orig_num_unlabeled * unlabeled_fraction)]
    # Set labels to 0
    for i in range(len(unlabeled_lines)):
        unlabeled_lines[i][0] = "0"
    f.close()

    # Shuffle lines
    random.shuffle(lines)
    random.shuffle(unlabeled_lines)

    # Do 10-fold CV
    num_examples = len(lines)
    num_unlabeled_examples = len(unlabeled_lines)
    lambda_l = 0.001
    lambda_u = 1

    semiSVMs = [2,3]
    lambda_ls = [10 ** i for i in range(-9, 1)]
    lambda_us = [10 ** i for i in range(-9, 1)]
    Rs = [i*0.1 for i in range(1,10)]
    num_folds = 10
    print "fraction of unlabeled_lines: {}".format(unlabeled_fraction)
    print "num_labeled_examples: {}".format(num_examples)
    print "num_unlabeled_examples: {}".format(num_unlabeled_examples)
    print "lambda_ls: {}".format(lambda_ls)
    print "lambda_us: {}".format(lambda_us)
    print "Rs: {}".format(Rs)

    start_time = time.time()

    best_acc = -1
    best_params = ()
    tot_count = len(semiSVMs)*len(lambda_ls)*len(lambda_us)*len(Rs)
    count = 0
    for semiSVM in semiSVMs:
        for lambda_l in lambda_ls:
            for lambda_u in lambda_us:
                for R in Rs:
                    accs = []
                    for fold_index, (tr_indices, te_indices) in enumerate(cv.KFold(num_examples, num_folds)):

                        semi_labels_tr_filename   = "{}.semi.{}.tr.labels".format(halfkernels_filename.split("/")[-1], unlabeled_fraction)
                        semi_features_tr_filename = "{}.semi.{}.tr.features".format(halfkernels_filename.split("/")[-1], unlabeled_fraction)
                        semi_labels_te_filename   = "{}.semi.{}.te.labels".format(halfkernels_filename.split("/")[-1], unlabeled_fraction)
                        semi_features_te_filename = "{}.semi.{}.te.features".format(halfkernels_filename.split("/")[-1], unlabeled_fraction)

                        # Write train and test files
                        write_svmlin_files(semi_labels_tr_filename, semi_features_tr_filename, tr_indices, lines, unlabeled_lines)
                        write_svmlin_files(semi_labels_te_filename, semi_features_te_filename, te_indices, lines) # Do not use unlabeled_lines when testing

                        # Train
                        os.system("{} -A {} -W {} -U {} -R {} {} {} > /dev/null".format(SVMLIN_PATH, semiSVM, lambda_l, lambda_u, R, semi_features_tr_filename, semi_labels_tr_filename))

                        semi_weights_filename = "{}.weights".format(semi_features_tr_filename)
                        semi_outputs_te_filename = "{}.outputs".format(semi_features_te_filename)

                        # Test
                        os.system("{} -f {} {} {} > /dev/null".format(SVMLIN_PATH, semi_weights_filename, semi_features_te_filename, semi_labels_te_filename))

                        acc = get_test_acc(semi_outputs_te_filename, te_indices, lines)
                        accs.append(acc)

                    avg_acc = np.average(accs)
                    if avg_acc > best_acc:
                        best_acc = avg_acc
                        best_params = (avg_acc, semiSVM, lambda_l, lambda_u, R)
                        print best_params
                    sys.stdout.write(".")
                    sys.stdout.flush()
                    count += 1

            # Calc ETA
            # s=vt
            v = count / float(time.time()-start_time)
            if v != 0:
                t = (tot_count-count) / float(v)
                print "ETA: {}".format(int(t))

    print ""

    train_time = time.time()-start_time
    print "best_params: {}".format(best_params)
    print "best_acc {:.2f}%".format(100.0*best_acc)
    print "train_time: {:.2f} secs.".format(train_time)
