"""
 * @file
 * @author  Tommi Kerola <kerola@student.chalmers.se>
 * @author  Linus Hermansson <hlinus@student.chalmers.se>
 * @version 1.0
 *
 * @section LICENSE
 *
 * Copyright [2013] [Tommi Kerola, Linus Hermansson]
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an "AS
 *  IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 *  express or implied.  See the License for the specific language
 *  governing permissions and limitations under the License.
 *
"""


#!/usr/bin/python

"""
Simple comparison of kernel matrix computation for three different ways of computing the direct product kernel:
    * Matrix inversion.
    * Sylvester equations.
    * Tommi & Linus' method.

Written by Tommi Kerola, 2013-04-16
"""

import sys
import numpy as np
import numpy.linalg as LA
import control # contains the dlyap solver
import itertools as it
import time

def create_adj_matrix(graphfilename, directed):
    # Assume edge list format: from to 1
    f = open(graphfilename, "r")
    lines = map(lambda line: map(int, line.split())[:2], f.read().splitlines())
    # a line in lines contains: from to
    f.close()
    max_id = np.max(lines)
    try:
        A = np.zeros((max_id, max_id))
    except MemoryError as me:
        print "max_id: {}".format(max_id)
        print "graphfilename: {}".format(graphfilename)
        raise me
    for a,b in lines:
        A[a-1,b-1] = 1
        if not directed:
            A[b-1,a-1] = 1
    return A

def add_in_edge(adjlist, a, b, directed):
    # format: attribute1 attribute2 list_of_in_nbrs
    if not a in adjlist:
        adjlist[a] = [1, 1, []]
    if not b in adjlist:
        adjlist[b] = [1, 1, []]
    adjlist[a][2].append(b) # in-nbr
    if not directed:
        adjlist[b][2].append(a) # If undirected, add opposite direction as well.

def create_adjlist(graphfilename, directed):
    # Assume edge list format: from to 1
    f = open(graphfilename, "r")
    lines = map(lambda line: map(int, line.split())[:2], f.read().splitlines())
    # a line in lines contains: from to
    f.close()
    adjlist = dict()
    for a,b in lines:
        add_in_edge(adjlist, b, a, directed) # Add in-edge
    return adjlist

def calc_random_walks_sum_adj(lamb,adjlist):
    eps = 10.0**-5
    K = int(np.log(eps) / np.log(lamb))
    u = np.zeros((K+1,1))
    for k in range(K+1):
        u[k] = (lamb ** (k/2.0)) * np.sum([val[0] for val in adjlist.values()])
        for key,(a1,a2,in_nbrs) in adjlist.iteritems():
            adjlist[key][1] = np.sum([adjlist[i][0] for i in in_nbrs])
        for key,(a1,a2,in_nbrs) in adjlist.iteritems():
            adjlist[key][0] = a2
    return u

def get_kernel_val_by_inv(lamb,A1,A2):
    Ax = np.kron(A1,A2)
    I = np.identity(Ax.shape[0])
    M = I - lamb * Ax
    return np.sum(LA.inv(M))

def get_kernel_val_by_sylvester(lamb,A1,A2):
    n = A1.shape[0]
    M0 = np.ones((n,n))
    # X = A X Q^T + C
    # M = \lambda A_1 M A_2^T + M_0
    A = lamb * A1
    Q = A2.T
    C = M0
    M = control.dlyap(A,Q,C)
    def vec(X):
        return np.reshape(X, X.size, order='F')
    return np.sum(vec(M))

def run_dp_kernel():
    if len(sys.argv) != 7:
        print "usage: {} type lambda directed normalize trainfile outfilename".format(sys.argv[0])
        print "       type in (kh, inv, sylv)"
        sys.exit(1)

    comp_type = sys.argv[1]
    lamb = float(sys.argv[2])
    directed = sys.argv[3] == "true"
    normalize = sys.argv[4] == "true"
    f = open(sys.argv[5], "r")
    lines = f.read().splitlines()
    f.close()
    f = open(lines[0], "r")
    labels = map(int, f.read().splitlines())
    f.close()
    graphfilenames = lines[1:]
    svm_gram_matrix_outfilename = sys.argv[6]

    if directed:
        print "Directed edges."
    else:
        print "Undirected edges."
    graphpairs = list(it.combinations(enumerate(graphfilenames), 2))

    n = len(graphfilenames)
    K=np.identity(n)
    start_time = time.time()
    if comp_type == "kh":
        if normalize:
            print "Normalizing TDP kernel."

        print "Using Tommi & Linus' comp.type. O(K|E|) (sparse)"
        U = []
        for i,g in enumerate(graphfilenames):
            sys.stdout.write(".")
            sys.stdout.flush()
            edges = create_adjlist(g, directed)
            u = calc_random_walks_sum_adj(lamb,edges)
            if normalize:
                norm = LA.norm(u) # Norm helps accuracy
                if norm == 0:
                    norm = 1
                U.append(u / norm)
            else:
                U.append(u)
        for (i,g1),(j,g2) in graphpairs:
            K[i,j] = K[j,i] = np.dot(U[i].T,U[j])
    else:
        if comp_type == "inv":
            print "Using matrix inversion. O(n^6) (dense)"
            comp_func = get_kernel_val_by_inv
        elif comp_type == "sylv":
            print "Using Sylvester eqs. O(n^3) (dense)"
            comp_func = get_kernel_val_by_sylvester
        else:
            raise Exception("Invalid computation type: {}".format(comp_type))

        num_pairs = len(graphpairs)
        for index,((i,g1),(j,g2)) in enumerate(graphpairs):
            sys.stdout.write(".")
            if index % 100 == 0:
                elapsed = time.time()-start_time
                speed = 1
                if elapsed != 0:
                    speed = index / elapsed
                if speed != 0:
                    sys.stdout.write(" {:.2f}% (ETA: {:d}) ".format(100*float(index)/num_pairs, int((num_pairs-index)/speed)))
            sys.stdout.flush()
            A1 = create_adj_matrix(g1, directed)
            A2 = create_adj_matrix(g2, directed)
            # Make sure the A's are of the same size
            if A1.shape < A2.shape:
                A1.resize(A2.shape)
            else:
                A2.resize(A1.shape)

            K[i,j] = K[j,i] = comp_func(lamb,A1,A2)
    print ""
    print "Kernel precomputation complete! ({} secs)".format(time.time() - start_time)

    if comp_type == "kh":
        # Print half-kernels to file in LIBSVM format.
        halfkernelfile_outname = "{}.halfkernels".format(svm_gram_matrix_outfilename)
        f = open(halfkernelfile_outname, "w")
        for i in range(n):
            vals = " ".join(["{}:{}".format(j+1, U[i][j][0]) for j in range(len(U[i]))])
            f.write("{} {}\n".format(labels[i], vals))
        f.close()
        print "Wrote halfkernels to \"{}\"".format(halfkernelfile_outname)

    # Print kernel matrix to file in LIBSVM format.
    f = open(svm_gram_matrix_outfilename, "w")
    for i in range(n):
        vals = " ".join(["{}:{}".format(j+1, K[i,j]) for j in range(n)])
        f.write("{} 0:{} {}\n".format(labels[i], i+1, vals))
    f.close()
    print "Wrote kernel matrix to \"{}\"".format(svm_gram_matrix_outfilename)

if __name__ == "__main__":
    run_dp_kernel()

