"""
 * @file
 * @author  Tommi Kerola <kerola@student.chalmers.se>
 * @author  Linus Hermansson <hlinus@student.chalmers.se>
 * @version 1.0
 *
 * @section LICENSE
 *
 * Copyright [2013] [Tommi Kerola, Linus Hermansson]
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an "AS
 *  IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 *  express or implied.  See the License for the specific language
 *  governing permissions and limitations under the License.
 *
"""

#!/usr/bin/python

"""
Does various tests on the IMDB dataset output files to make sure that the extraction was performed without errors.
"""

import sys

def check_graph_is_undirected(inputfilename):
    """
    Checks that in the adjacency list file, for each edge from -> to, there exists an edge to -> from.

    Input: imdb_amb_adjlist_kernel.txt
    """

    id_to_nbrs = dict()
    f = open(inputfilename, "r")
    lines = f.read().splitlines()
    f.close()
    print "Stage 1: Preprocess IDs."
    num_lines = len(lines)
    for index,line in enumerate(lines):
        vals = line.split()
        actor_id = vals[0]
        nbrs = vals[2:][::2]
        assert not actor_id in id_to_nbrs, "Error: actor_id {} already added to id_to_nbrs!".format(actor_id)
        id_to_nbrs[actor_id] = nbrs

        if index % 1000 == 0:
            print "[{}] {:.2f}%".format(actor_id, 100*index/float(num_lines))

    print "Stage 2: Check undirectedness."
    for index,line in enumerate(lines):
        vals = line.split()
        actor_id = vals[0]
        nbrs = vals[2:][::2]
        for nbr in nbrs:
            if not actor_id in id_to_nbrs[nbr]:
                print "Warning: Found edge from {} -> {}, but not other way around!".format(actor_id, nbr)

        if index % 1000 == 0:
            print "[{}] {:.2f}%".format(actor_id, 100*index/float(num_lines))

def check_actors_with_only_one_movie(inputfilename):
    """
    Checks in the adjacency list file, if there are any actors that have only made one movie.

    Input: imdb_amb_adjlist_malin.txt
    """

    f = open(inputfilename, "r")
    lines = f.read().splitlines()
    f.close()
    num_lines = len(lines)
    totcount = 0
    poscount = 0
    negcount = 0

    print "Stage 1: Check number of movies per actor."
    for index,line in enumerate(lines):
        vals = line.split()
        actor_id = vals[0]
        num_ents = int(vals[1])
        movies = set(vals[2:][1::2]) # The movies this actor has starred in.
        if len(movies) == 1:
            print "Warning: Actor {} has only starred in {} movies!".format(actor_id, len(movies))
            totcount += 1
            if num_ents > 1:
                poscount += 1
            else:
                negcount += 1
        elif len(movies) < 1:
            print "Error: Actor {} has starred in NO movies!".format(actor_id)
            sys.exit(1)

        if index % 1000 == 0:
            print "[{}] {:.2f}%".format(actor_id, 100*index/float(num_lines))

    print "Num actors with only one movie: {} ({:.2f}%)".format(totcount, 100.0 * totcount / float(num_lines))
    print "of which +1: {:.2f}%".format(100.0 * poscount / float(totcount))
    print "         -1: {:.2f}%".format(100.0 * negcount / float(totcount))

def check_actors_with_self_edge(inputfilename):
    """
    Checks in the adjacency list file, if there are any actors that have edges to themselves.
    These can occur when creating ambiguous actors if a movie contains several actors with the same last name.

    Input: imdb_amb_adjlist_malin.txt
    """

    f = open(inputfilename, "r")
    lines = f.read().splitlines()
    f.close()
    num_lines = len(lines)
    totcount = 0
    poscount = 0
    negcount = 0

    print "Stage 1: Check number of actors with self-edges."
    for index,line in enumerate(lines):
        vals = line.split()
        actor_id = vals[0]
        num_ents = int(vals[1])
        nbrs = vals[2:][::2] # The actors this actor has starred with.
        if actor_id in nbrs:
            print "Warning: Actor {} has self-edge!".format(actor_id)
            totcount += 1
            if num_ents > 1:
                poscount += 1
            else:
                negcount += 1

        if index % 1000 == 0:
            print "[{}] {:.2f}%".format(actor_id, 100*index/float(num_lines))

    print "Num actors with self-edges: {} ({:.2f}%)".format(totcount, 100.0 * totcount / float(num_lines))
    totcount = max(totcount, 1)
    print "of which +1: {:.2f}%".format(100.0 * poscount / float(totcount))
    print "         -1: {:.2f}%".format(100.0 * negcount / float(totcount))

if __name__ == "__main__":

    if len(sys.argv) != 2:
        print "usage: {} imdb_file".format(sys.argv[0])
        sys.exit(1)

    inputfilename = sys.argv[1]
    #check_graph_is_undirected(inputfilename)
    #check_actors_with_only_one_movie(inputfilename)
    check_actors_with_self_edge(inputfilename)
