"""
 * @file
 * @author  Tommi Kerola <kerola@student.chalmers.se>
 * @author  Linus Hermansson <hlinus@student.chalmers.se>
 * @version 1.0
 *
 * @section LICENSE
 *
 * Copyright [2013] [Tommi Kerola, Linus Hermansson]
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an "AS
 *  IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 *  express or implied.  See the License for the specific language
 *  governing permissions and limitations under the License.
 *
"""


#!/usr/bin/python

import numpy as np
import numpy.linalg as LA
import matplotlib.pyplot as plt
import sys
from collections import defaultdict

def adj_mat(n):
    return np.mat(np.random.random_integers(0,1, (n,n)))
    # return np.mat(np.random.choice([0,1], (n,n)))

def get_max_in_degree(A):
    indeg = np.sum(A, 0)
    return np.max(indeg)

def get_max_out_degree(A):
    outdeg = np.sum(A, 1)
    return np.max(outdeg)

def get_max_degree(A):
    return np.max([get_max_out_degree(A), get_max_in_degree(A)])

def get_good_lambda(A):
    # lambda < 1/a, where a >= max_degree(A)
    a = get_max_degree(A)
    lamb = 1.0 / (a + 1.0)
    return lamb

def exact_kernel(lamb, A):
    I = np.eye(A.shape[0])
    inv = LA.inv( I - lamb * A )
    #print "inv:",inv
    return np.sum( inv )

def trunc_kernel(K, lamb, A):
    s = np.eye(A.shape[0])
    total = s
    us = [np.sum(s)]
    for k in range(1, K+1):
        s = s * (lamb * A)
        total += s
        us.append(np.sum(total))
    return us

plot_convergence = False
num_iters = 100

if plot_convergence:
    K_max = 40
    Vs = [25, 75, 125, 200]
else:
    K_max = 100
    Vs = range(5, 200, 1)

best_Ks = defaultdict(list)
K_range = range(0, K_max+1)
for i in range(num_iters):
    for n in Vs:
        A = adj_mat(n)
        lamb = get_good_lambda(A)
        print "{}: {}".format(n, lamb)

        sys.stdout.write(".")
        sys.stdout.flush()
        exact = exact_kernel(lamb, A)
        sys.stdout.write(".")
        sys.stdout.flush()
        approx_list = trunc_kernel(K_max, lamb, A)
        sys.stdout.write(".")
        sys.stdout.flush()

        if plot_convergence:
            sys.stdout.write("P")
            sys.stdout.flush()
            plt.plot(K_range, [a-exact for a in approx_list], "--", label="approx-exact ({})".format(n))
            print "."

        alpha = 0.99
        best_K = n
        for (K, a) in enumerate(approx_list):
            if a/float(exact) >= alpha:
                best_K = K
                print "best_K:",best_K
                break
        best_Ks[n].append(best_K)

if plot_convergence:
    plt.xlabel("K")
    plt.ylabel("Kernel value diff.")
    plt.legend(loc='lower right')
    plt.show()

print "num_iters:",num_iters

plt.plot(Vs, [np.average(best_Ks[n]) for n in Vs], "-")
plt.xlabel("|V|")
plt.ylabel("K_conv")
plt.show()

