"""
 * @file
 * @author  Tommi Kerola <kerola@student.chalmers.se>
 * @author  Linus Hermansson <hlinus@student.chalmers.se>
 * @version 1.0
 *
 * @section LICENSE
 *
 * Copyright [2013] [Tommi Kerola, Linus Hermansson]
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an "AS
 *  IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 *  express or implied.  See the License for the specific language
 *  governing permissions and limitations under the License.
 *
"""



#!/usr/bin/python

"""
Implementation of the Nested Variational EM algorithm
(Airoldi et al. 2008)


"""

import sys
import numpy as np
import scipy
import scipy.special

if len(sys.argv) != 1:
    print "usage: {}".format(sys.argv[0])
    sys.exit(1)

# Computes E_q [ log \pi_k ]
# Args: k - the index of \pi
#       alpha - the alpha vector
def E_q_log_pi_k(k, alpha):
    return scipy.special.psi(alpha[k]) - scipy.special.psi(np.sum(alpha))

def get_variational(K, phi_pq, phi_qp, Y_pq, gamma_p, gamma_q, B, alpha):
    phi_pq[:] = phi_qp[:] = 1/K
    i = 0
    while 1:
        for g in range(K):
            P = [( (B[g,h]**Y_pq) * ((1-B[g,h])**(1-Y_pq)) )**phi_qp[h] for h in range(K)]
            phi_pq[g] = np.exp(E_q_log_pi_k(g, alpha)) * P.prod()
        phi_pq /= phi_pq.sum()

        for h in range(K):
            P = [( (B[g,h]**Y_pq) * ((1-B[g,h])**(1-Y_pq)) )**phi_pq[g] for g in range(K)]
            phi_qp[h] = np.exp(E_q_log_pi_k(h, alpha)) * P.prod()
        phi_qp /= phi_qp.sum()

        converged = i > 100
        i+=1
        if converged:
            break

def partially_update_gamma(K, gamma_p, phi_p, phi_q, alpha):
    for k in range(K):
        gamma_p[k] = alpha[k] + phi_p.sum(axis=0)[k] + phi_q.sum(axis=0)[k]

def partially_update(K, gamma_p, gamma_q, B, phi_p, phi_q, alpha):
    gamma_p = partially_update_gamma(K, gamma_p, phi_p, phi_q, alpha)
    gamma_q = partially_update_gamma(K, gamma_q, phi_q, phi_p, alpha)
    alpha = get_alpha_newton_raphson()
    B = partially_update_B()
    return gamma_p, gamma_q, alpha, B

# Nested variational EM algorithm.
# Args: N - the number of vertices
#       K - the number of groups
#       Y - the NxN adjacency matrix
def nested_variational_EM(N,K,Y):
    phi = np.ones((N,N,K))
    gamma = np.ones((N,K)) * 2*N/K
    alpha = np.ones((K,1)) * 0.25
    B = np.ones((K,K)) * 0.5
    i = 0
    while 1:
        for p in range(N):
            for q in range(N):
                # phi[p,q] and phi[q,p]
                phi[p,q], phi[q,p] = get_variational(K, phi[p,q], p[q,p], Y[p,q], gamma[p], gamma[q], B, alpha)
                gamma[p], gamma[q], alpha, B = partially_update(K, gamma[p], gamma[q], B, phi[p], phi[q], alpha)

        converged = i > 100
        i+=1
        if converged:
            break
