/**
 * @file
 * @author  Tommi Kerola <kerola@student.chalmers.se>
 * @author  Linus Hermansson <hlinus@student.chalmers.se>
 * @version 1.0
 *
 * @section LICENSE
 *
 * Copyright [2013] [Tommi Kerola, Linus Hermansson]
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an "AS
 *  IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 *  express or implied.  See the License for the specific language
 *  governing permissions and limitations under the License.
 *
 */

#include <iostream>
#include <fstream>
#include <sstream>
#include <cstdlib>
#include <cstring>
#include "shortest_path_types.hpp"

const int INFINITY_LENGTH = 1000000;
const int k = 10;
//const int N = 5;

int **createTwoDimensionalArray(int m, int n) {
	int **p, i;
	p = (int **)malloc(m*sizeof(int*));/* this will store base order of all the row in p */
	for(i = 0; i < m; i++)
	p[i] = (int *)malloc(n*sizeof(int));/* this will create m row of n elements */
	return p;
}

int generateBinnedAnswer(char* filepath, int* result, bool useGraphChiOutputFormat);

int main(int argc,char** argv) {

	int *B1 = new int[k];
	int *B2 = new int[k];
	counter_type similarity = 0;

	//algo--------------------------------------------------------------

	// When using GraphLab, vertices that are not connected to any node
	// will no be printed. However, in GraphChi they are, so we need to
	// handle both these cases.
	bool useGraphChiOutputFormat = false;

	if(argc < 2) {
		std::cout << "usage: " << argv[0] << " graph [graphchi|graphlab]\n";
		return 1;
	}

	// GraphLab and GraphChi differ in output format styles.
	// Therefore, we must alter the format depending on which library we use.
	if(argc == 4) {
		if(strcmp(argv[3], "graphchi") == 0) {
	    		useGraphChiOutputFormat = true;
	    		std::cout << "Using GraphChi output format.\n";
		} 
		else if(strcmp(argv[3], "graphlab") == 0) {
	    		useGraphChiOutputFormat = false;
	    		std::cout << "Using GraphLab output format.\n";
		}
	}
	
	


	generateBinnedAnswer(argv[1], B1, useGraphChiOutputFormat);
	generateBinnedAnswer(argv[2], B2, useGraphChiOutputFormat);

	std::cout << "First Graph: " <<std::endl;
	for(int i = 0; i < k; i++)
	{
		std::cout << "#Shortest paths of length: " << i+1  << " = " <<B1[i] << std::endl;	
	}

	std::cout << "Second Graph: " <<std::endl;
	for(int i = 0; i < k; i++)
	{
		std::cout << "#Shortest paths of length: " << i+1  << " = " <<B2[i] << std::endl;	
	}

	for(int i = 0; i < k; i++)
	{
		similarity += B1[i]*B2[i];
	}

	std::cout << "Final Similarity measurment: " << similarity << std::endl;

	return 0;

}

int generateBinnedAnswer(char* filepath, int* result, bool useGraphChiOutputFormat)
{	
	std::ifstream file(filepath);
	int maxId = 0;

	std::cout << "Finding the maximum id of any node..." << std::endl;	

	int firstInt;
	int secondInt;

	if (file.is_open()) {
        	std::string line;
		while (std::getline(file, line)) {

            		if(line[0] == '#') // Skip comments
                		continue;

            		std::istringstream iss(line);
			if(!(iss >> firstInt >> secondInt))
                		break; // error

			if(firstInt > maxId)
				maxId = firstInt;

			if(secondInt > maxId)
				maxId = secondInt;
		}
		file.close();
	}

	const int N = maxId+1;

	std::cout << "Starting memory allocation, with N = " << N << std::endl;
	
	int **cost = createTwoDimensionalArray(N, N);
	int **adjacency = createTwoDimensionalArray(N, N);
	int **distance = createTwoDimensionalArray(N, N);
	
	std::cout << "Memory allocation complete, generating graph with N = " << N << std::endl;

	for(int i = 0; i < N; i++)
		for(int j = 0; j < N; j++)
			adjacency[i][j] = 0;

	file.open(filepath);
	if (file.is_open()) {
        std::string line;
		while (std::getline(file, line)) {

            if(line[0] == '#') // Skip comments
                continue;

            std::istringstream iss(line);
			if(!(iss >> firstInt >> secondInt))
                break; // error

			adjacency[firstInt][secondInt] = 1;
		}
		file.close();
	}
/*
	for(int i = 0; i < N; i++) {
		for(int j = 0; j < N; j++) {
            std::cout << adjacency[i][j] << " ";
        }
        std::cout << std::endl;
    }
*/

//-------------------------------
// Simple undirected graph with all edges of length 1.
//-------------------------------
	
	for(int i = 0; i < N; i++)
		for(int j = 0; j < N; j++)
			if(i != j)
				distance[i][j] = 1;
			else
				distance[i][j] = 0;

//-------------------------------
// ENDOF Simple undirected graph
//-------------------------------

	

	std::cout << "Graph Generated! Now running the algorithm" << std::endl;
	for(int i = 0; i < N; i++) {
		for(int j = 0; j < N; j++) {
			if(adjacency[i][j] == 1 && i != j)
				cost[i][j] = distance[i][j];
			else if(i == j)
                cost[i][j] = 0;
            else
                cost[i][j] = INFINITY_LENGTH;
		}
	}

	double loop_counter = 0;

	for(int h = 0; h < N; h++)
	{
		if(h % (N /10 ) == 0)
		{
			std::cout << loop_counter*10 << "% \n";
			loop_counter++;
		}
		for(int i = 0; i < N; i++)
		{
			for(int j = 0; j < N; j++)
			{
				if(cost[i][h] + cost[h][j] < cost[i][j])
					cost[i][j] = cost[i][h] + cost[h][j];
			}
		}
	}
	

	std::cout << "\n";

	std::ofstream outfile;

	outfile.open("result.fw");

	for(int i = 0; i < N; i++) {
		// Only print a line for a node $i$ if it is connected to some node $j$, or if some node $j$ is connected to this node $i$.
		// This is needed since GraphLab does not print information for vertices that have no edges.
		bool okToPrint = false;
			for(int j = 0; j < N; j++)
				if(i != j && (adjacency[j][i] == 1 || adjacency[i][j] == 1))
		        okToPrint = true;

		if(useGraphChiOutputFormat || okToPrint) {
			    outfile << i << " ==> \t";

		    for(int j = 0; j < N; j++)
		        if(cost[j][i] != INFINITY_LENGTH)
		            outfile << "(" << j << ":" << cost[j][i] << ") ";

		    outfile << std::endl;
		}
	}


	for(int i = 0; i < k; i++)
	{
			result[i] = 0;		
	}


	for(int i = 0; i < N; i++)
	{
		for(int j = 0; j < N; j++)
		{
			if(cost[j][i] <= k && cost[j][i] != 0)
			{
				result[ cost[j][i] - 1 ] += 1;
			}
		}
	}
	
	return 0;

}


