/**
 * @file
 * @author  Tommi Kerola <kerola@student.chalmers.se>
 * @author  Linus Hermansson <hlinus@student.chalmers.se>
 * @version 1.0
 *
 * @section LICENSE
 *
 * Copyright [2013] [Tommi Kerola, Linus Hermansson]
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an "AS
 *  IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 *  express or implied.  See the License for the specific language
 *  governing permissions and limitations under the License.
 *
 * @section DESCRIPTION
 *
 * Utility class for loading a GraphLab graph.
 */
#ifndef GRAPH_LOADER_HPP
#define GRAPH_LOADER_HPP

#include "graphlabchi.hpp"
#include <boost/bind.hpp>

/**
 * \brief Loads several graphs into one single graph, representing each graph as a cluster.
 */
template<class GraphType>
class ClusterGraphLoader {
    public:

        ClusterGraphLoader() {}

        ~ClusterGraphLoader() {
            if(graph) {
                graph->clear();
                SAFE_DELETE(graph);
            }
        }

        bool load(std::vector<std::string> graph_files, std::string format,
                graphlab::distributed_control& dc, graphlab::command_line_options& clopts) {

            cluster_id = 0;
            offset = 0;
            graph = new GraphType(dc, clopts);
            for(size_t i=0;i<graph_files.size();i++) {

                max_vertex_id = 0;
                graph->load(graph_files[i], boost::bind(&ClusterGraphLoader::snap_parser<GraphType>, this, _1, _2, _3));
                dc.cout() << "graph " << i << ", max_vertex_id: " << max_vertex_id << ", offset: " << offset << std::endl;
                offset += max_vertex_id+1;
                cluster_id++;

            }
            graph->finalize();
            if (graph->num_vertices() == 0) {
                throw MtltQuitException("Could not load graphs. Please check the graph files.");
            }

            dc.cout() << "num_clusters: " << cluster_id << std::endl;
            dc.cout() 
                << "========== Graph statistics on proc " << dc.procid() 
                << " ==============="
                << "\n Num vertices: " << graph->num_vertices()
                << "\n Num edges: " << graph->num_edges()
                << "\n Num replica: " << graph->num_replicas()
                << "\n Replica to vertex ratio: " 
                << float(graph->num_replicas())/graph->num_vertices()
                << "\n --------------------------------------------" 
                << "\n Num local own vertices: " << graph->num_local_own_vertices()
                << "\n Num local vertices: " << graph->num_local_vertices()
                << "\n Replica to own ratio: " 
                << (float)graph->num_local_vertices()/graph->num_local_own_vertices()
                << "\n Num local edges: " << graph->num_local_edges()
                //<< "\n Begin edge id: " << graph->global_eid(0)
                << "\n Edge balance ratio: " 
                << float(graph->num_local_edges())/graph->num_edges()
                << std::endl;

            return true;
        }

        /**
         * \brief Parse files in the Stanford Network Analysis Package format.
         *
         * example:
         *
         *  # some comment
         *  # another comment
         *  1 2
         *  3 4
         *  1 4
         *
         */
        template <typename Graph>
            bool snap_parser(Graph& graph, const std::string& srcfilename,
                    const std::string& str) {

                if (str.empty()) return true;
                else if (str[0] == '#') {
                    std::cout << str << std::endl;
                } else {
                    size_t source, target;
                    double weight = 1;
                    char* targetptr;
                    source = strtoul(str.c_str(), &targetptr, 10);
                    if(targetptr == NULL)
                        return false;
                    target = strtoul(targetptr, &targetptr, 10);
                    if(targetptr != NULL)
                        weight = strtoul(targetptr, NULL, 10);

                    if(source > max_vertex_id)
                        max_vertex_id = source;
                    if(target > max_vertex_id)
                        max_vertex_id = target;

                    if(source != target) {
                        graph.add_vertex(source+offset, typename Graph::vertex_data_type(cluster_id, source == 0));
                        graph.add_vertex(target+offset, typename Graph::vertex_data_type(cluster_id, target == 0));
                        graph.add_edge(source+offset, target+offset, typename Graph::edge_data_type(weight));
                    }
                }
                return true;
            } // end of snap parser

        inline GraphType* get_graph() {return graph;}

    private:
        GraphType* graph;
        long max_vertex_id;
        long offset;
        long cluster_id;
};

template<class GraphType>
class GraphLoader {
    public:

        GraphLoader() {}

        ~GraphLoader() {
            if(graph) {
                graph->clear();
                SAFE_DELETE(graph);
            }
        }

        bool load(std::string graph_dir, std::string format,
                graphlab::distributed_control& dc, graphlab::command_line_options& clopts) {

            graph = new GraphType(dc, clopts);
            graph->load_format(graph_dir, format);
            graph->finalize();

            /*
            dc.cout() 
                << "========== Graph statistics on proc " << dc.procid() 
                << " ==============="
                << "\n Num vertices: " << graph->num_vertices()
                << "\n Num edges: " << graph->num_edges()
                << "\n Num replica: " << graph->num_replicas()
                << "\n Replica to vertex ratio: " 
                << float(graph->num_replicas())/graph->num_vertices()
                << "\n --------------------------------------------" 
                << "\n Num local own vertices: " << graph->num_local_own_vertices()
                << "\n Num local vertices: " << graph->num_local_vertices()
                << "\n Replica to own ratio: " 
                << (float)graph->num_local_vertices()/graph->num_local_own_vertices()
                << "\n Num local edges: " << graph->num_local_edges()
                //<< "\n Begin edge id: " << graph->global_eid(0)
                << "\n Edge balance ratio: " 
                << float(graph->num_local_edges())/graph->num_edges()
                << std::endl;
            */

            if (graph->num_vertices() == 0) {
                dc.cout() << "Could not load graph: \"" << graph_dir << "\"\n";
                return false;
            }

            return true;
        }

        inline GraphType* get_graph() {return graph;}

    private:
        GraphType* graph;
};


#endif
