/**
 * @file
 * @author  Tommi Kerola <kerola@student.chalmers.se>
 * @author  Linus Hermansson <hlinus@student.chalmers.se>
 * @version 1.0
 *
 * @section LICENSE
 *
 * Copyright [2013] [Tommi Kerola, Linus Hermansson]
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an "AS
 *  IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 *  express or implied.  See the License for the specific language
 *  governing permissions and limitations under the License.
 *
 */

#ifndef GRAPHCHI_ENGINE_WRAPPER_HPP
#define GRAPHCHI_ENGINE_WRAPPER_HPP

#include "graphlabchi.hpp"

template<class VertexType, class EdgeType, class VertexProgram>
class GraphChiEngineWrapper : GraphEngineWrapper<VertexType, EdgeType> {

    public:

        typedef graphlab::distributed_graph<VertexType, EdgeType> GraphType;

        /**
         * \brief Constructs a new graph engine.
         */
        GraphChiEngineWrapper(std::string filename, std::string format,
                graphlab::distributed_control& dc, 
                graphlab::command_line_options& clopts) 
            : GraphEngineWrapper<VertexType, EdgeType>() {

            load_graph(filename, format, dc, clopts);
        }

        /**
         *  \brief Maps the mapfunction to each vertex and returns a result.
         *  May not change the vertex data.
         *
         *  \return The reduction value.
         */
        template <typename ReductionType, typename VertexMapperType>
        virtual ReductionType map_reduce_vertices(VertexMapperType mapfunction) {
        }

        /**
         *  \brief Maps the mapfunction to each vertex.
         *  May modify the vertex.
         */
        template <typename VertexMapperType>
        virtual void transform_vertices(VertexMapperType mapfunction) {

            if(vertices == NULL) {
                vertices = load_vertices();
            }
        }

        /**
         * \brief Start the engine execution.
         *
         * \return the reason for termination
         */
        virtual execution_status::status_enum start() {

            typedef graphlab::GraphLabWrapper<VertexProgram> GLWrapper;
            GLWrapper wrapperProgram;
            graphchi::graphchi_engine<bool, typename GLWrapper::EdgeDataType> engine(graph_filename, nshards, use_scheduler, _m); 
            engine.set_modifies_inedges(false);
            engine.set_modifies_outedges(false);
            engine.run(wrapperProgram, niters);

            vertices = wrapperProgram.vertexInmemoryArray;
        }

        /**
         *  \brief Signals all the vertices.
         */
        virtual void signal_all() = 0;

        /**
         * \brief Get the elapsed time in seconds since start was last
         * called.
         * 
         * \return elapsed time in seconds
         */
        virtual float elapsed_seconds() const = 0;

    private:

        /**
         * \brief Loads the graph.
         *
         * \return Whether the graph was loaded successfully or not.
         */
        bool load_graph(std::string filename) {
            
            graph_filename  = filename;
            niters          = graphchi::get_option_int("niters", std::numeric_limits<int>::max()); // Loop until converged.
            use_scheduler   = true;                                                 // Whether the scheduler should be enabled or not.
            filetype        = graphchi::get_option_string("filetype", "edgelist");  // Note: --filetype (double dashes) cannot be used since GraphChi
                                                                                    // expects a parameter on this format.
            // Process input file - if not already preprocessed
            nshards         = graphchi::convert_if_notexists<EdgeType>(filename, graphchi::get_option_string("nshards", "auto"));

            return nshards > 0;
        }

        /**
         * Loads vertices from file.
         */
        std::vector<VertexType>* load_vertices() {
            std::string filename = graphchi::filename_vertex_data<int>(graph_filename);
            FILE*  f = fopen(filename.c_str(), "w");
            for(int i=0; i<numvertices; i++) {
                int value = vertex_value_of(i); // TODO 
                fwrite(&value, sizeof(int), 1, f);
            } 
            fclose(f);
        }

        std::string graph_filename;
        std::string filetype;
        int niters;
        int nshards;
        bool use_scheduler;
        size_t num_vertices;
        size_t num_edges;

        std::vector<VertexType>* vertices;

};

#endif
