/**
 * @file
 * @author  Tommi Kerola <kerola@student.chalmers.se>
 * @author  Linus Hermansson <hlinus@student.chalmers.se>
 * @version 1.0
 *
 * @section LICENSE
 *
 * Copyright [2013] [Tommi Kerola, Linus Hermansson]
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an "AS
 *  IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 *  express or implied.  See the License for the specific language
 *  governing permissions and limitations under the License.
 *
 * @section DESCRIPTION
 *
 * Graphlab engine wrapper.
 */
#ifndef GRAPHLAB_ENGINE_WRAPPER_HPP
#define GRAPHLAB_ENGINE_WRAPPER_HPP

#include "graphlabchi.hpp"

template<class VertexType, class EdgeType>
class GraphLabEngineWrapper : GraphEngineWrapper<VertexType, EdgeType> {

    public:

        typedef graphlab::distributed_graph<VertexType, EdgeType> GraphType;

        /**
         * \brief Constructs a new graph engine.
         */
        GraphLabEngineWrapper(std::string filename, std::string format,
                graphlab::distributed_control& dc, 
                graphlab::command_line_options& clopts) 
            : GraphEngineWrapper<VertexType, EdgeType>() {

            load_graph(filename, format, dc, clopts);
        }

        /**
         *  \brief Maps the mapfunction to each vertex and returns a result.
         *  May not change the vertex data.
         *
         *  \return The reduction value.
         */
        template <typename ReductionType, typename VertexMapperType>
        virtual ReductionType map_reduce_vertices(VertexMapperType mapfunction) {
        }

        /**
         *  \brief Maps the mapfunction to each vertex.
         *  May modify the vertex.
         */
        template <typename VertexMapperType>
        virtual void transform_vertices(VertexMapperType mapfunction) {
        }

        /**
         * \brief Start the engine execution.
         *
         * \return the reason for termination
         */
        virtual execution_status::status_enum start() = 0;

        /**
         *  \brief Signals all the vertices.
         */
        virtual void signal_all() = 0;

        /**
         * \brief Get the elapsed time in seconds since start was last
         * called.
         * 
         * \return elapsed time in seconds
         */
        virtual float elapsed_seconds() const = 0;

    private:

        /**
         * \brief Loads the graph.
         *
         * \return Whether the graph was loaded successfully or not.
         */
        bool load_graph(std::string graph_dir, std::string format,
                graphlab::distributed_control& dc,
                graphlab::command_line_options& clopts) {

            graph = new GraphType(dc, clopts);
            graph->load_format(graph_dir, format);
            graph->finalize();
            if (graph->num_vertices() == 0) {
                dc.cout() << "Could not load graph: \"" << graph_dir << "\"\n";
                return false;
            }

            return true;
        }

        GraphType* graph;

};

#endif
