/**
 * @file
 * @author  Tommi Kerola <kerola@student.chalmers.se>
 * @author  Linus Hermansson <hlinus@student.chalmers.se>
 * @version 1.0
 *
 * @section LICENSE
 *
 * Copyright [2013] [Tommi Kerola, Linus Hermansson]
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an "AS
 *  IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 *  express or implied.  See the License for the specific language
 *  governing permissions and limitations under the License.
 *
 * @section DESCRIPTION
 *
 * Random walk graph half-kernel.
 */
#ifndef HALF_RANDOM_WALK_KERNEL_HPP
#define HALF_RANDOM_WALK_KERNEL_HPP

#include "sparse_vector.hpp"
#include "half_graph_kernel.hpp"

#include <boost/bind.hpp>
#include <cmath>

typedef unsigned long long int ULLONG;

class HalfRandomWalkKernel : public HalfGraphKernel<rw_vertex_data, double> {
    
    public:

        typedef graphlab::distributed_graph<rw_vertex_data, double> GraphType;
    
        /**
         * \brief Constructs a new half graph kernel for comparing the two graphs g1 and g2.
         *
         * params:
         * g1     - the graph
         * dc     - graphlab::distributed_control object.
         * lambda - kernel parameter, the path length decay factor. lambda \in [0,1]
         */
        HalfRandomWalkKernel(GraphType& _g1,
                graphlab::distributed_control& _dc,
                std::string _engine_type,
                graphlab::command_line_options& _clopts,
                double _lambda,
                bool _count_from_distinguished_vertex_only) :
            HalfGraphKernel<rw_vertex_data, double>(_g1, _dc, _engine_type, _clopts), lambda(_lambda), count_from_distinguished_vertex_only(_count_from_distinguished_vertex_only) {

        }

        virtual ~HalfRandomWalkKernel() {
        }

        std::vector<SparseVector> calculate(size_t num_graphs) {

            this->dc.cout() << "Running RW kernel." << std::endl;
            if(count_from_distinguished_vertex_only)
                this->dc.cout() << "Counting walks from distinguished vertex only." << std::endl;
            else
                this->dc.cout() << "Counting walks from all vertices." << std::endl;

            graphlab::omni_engine<RandomWalkProgram> g1_engine(this->dc, this->g1, engine_type, clopts);

            g1_engine.transform_vertices(boost::bind(&HalfRandomWalkKernel::init_vertex_data, this, _1, _2));

            const double LIMIT = 1e-5; // Loop until lambda^i becomes smaller than this value.
            int n = (int) (log(LIMIT) / log(lambda)); // Calculate suitable number of times to loop.
            //SparseVector u1(n);

            std::vector<SparseVector> half_kernels(num_graphs, SparseVector(n));
            this->dc.cout() << "n = " << n << std::endl;
            for(int i=0; i<n; i++) {

                // Get the sums for this walk length.
		// g1_map will contain all the indexes of the clusters and each index will be the sum
		// of all the num_walks in that cluster
                SummableMap<double> g1_map = g1_engine.map_reduce_vertices<SummableMap<double> >(
                                    boost::bind(&HalfRandomWalkKernel::get_vertex_data, this, _1, _2));
                double lambda_val = pow(lambda, i);

                // Run each vertex program one step.
                g1_engine.signal_all();
                g1_engine.start();

                for(std::map<size_t, double>::iterator it=g1_map.map.begin();it!=g1_map.map.end();it++) {
                    size_t cluster_id = it->first;
                    double num_walks = it->second;
                    half_kernels[cluster_id](i) = num_walks * sqrt(lambda_val);
                }
            }

            return half_kernels;
        }

    private:

        /**
         * \brief Get the data of a vertex.
         */
        SummableMap<double> get_vertex_data(graphlab::omni_engine<RandomWalkProgram>::icontext_type& context, const random_walk_graph_type::vertex_type& vertex) {
            SummableMap<double> result;
            result.map[vertex.data().cluster_id] = vertex.data().num_walks;
            return result;
        }

        /**
         * \brief Initializes the data of a vertex.
         */
        void init_vertex_data(graphlab::omni_engine<RandomWalkProgram>::icontext_type& context, random_walk_graph_type::vertex_type& vertex) {
            //vertex.data().num_walks = MTLT_DIRECTED_GRAPH ? vertex.num_in_edges() : vertex.num_in_edges() + vertex.num_out_edges();
            if(count_from_distinguished_vertex_only) {
                if(vertex.data().distinguished_vertex) { // Only count walks from the "Chris Andeson" vertex.
                    //std::cout << "Found distinguished vertex! id: " << vertex.id() << " cluster_id: " << vertex.data().cluster_id << std::endl;
                    vertex.data().num_walks = 1;
                }
            } else { // No distinguished vertex, count walks from all nodes.
                vertex.data().num_walks = 1;
            }
        }
        
        /*
         * VARIABLES
         */
        double lambda; // path length decay factor. lambda \in [0,1]
        bool count_from_distinguished_vertex_only;  // Whether we should only count walks from the distinguished "Chris Anderson" vertex with id 0.
};

#endif
