/**
 * @file
 * @author  Tommi Kerola <kerola@student.chalmers.se>
 * @author  Linus Hermansson <hlinus@student.chalmers.se>
 * @version 1.0
 *
 * @section LICENSE
 *
 * Copyright [2013] [Tommi Kerola, Linus Hermansson]
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an "AS
 *  IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 *  express or implied.  See the License for the specific language
 *  governing permissions and limitations under the License.
 *
 * @section DESCRIPTION
 *
 * Shortest path graph half-kernel.
 */
#ifndef HALF_SHORTEST_PATH_KERNEL_HPP
#define HALF_SHORTEST_PATH_KERNEL_HPP

#include "sparse_vector.hpp"
#include <boost/bind.hpp>

#include "half_graph_kernel.hpp"
//#include "shortest_path_program.hpp"
#include "shortest_path_types.hpp"
#include "histogram.hpp"

class ShortestPathProgram;

class HalfShortestPathKernel : public HalfGraphKernel<vertex_data, edge_data> {
    
    typedef graphlab::distributed_graph<vertex_data, edge_data> GraphType;
    
    public:

        /**
         * Constructs a new half graph kernel for comparing the two graphs g1 and g2.
         *
         * params:
         * g1     - the graph
         * dc     - graphlab::distributed_control object.
         * k      - kernel parameter, the maximum length of a path.
         */
        HalfShortestPathKernel(GraphType& _g1,
                graphlab::distributed_control& _dc,
                std::string _engine_type,
                graphlab::command_line_options& _clopts,
                unsigned int _k,
                std::string _binning_type = "linear",
                size_t _num_bins = M_E,
                bool _inv_edge_lengths = false,
                bool _log_edge_lengths = false) :
            HalfGraphKernel<vertex_data, edge_data>(_g1, _dc, _engine_type, _clopts), k(_k), binning_type(_binning_type), num_bins(_num_bins), inv_edge_lengths(_inv_edge_lengths), log_edge_lengths(_log_edge_lengths) {

            ShortestPathProgram::max_path_length = _k;
        }

        /**
         * Calculates the normalized vector the half-kernel.
         */
        std::vector<SparseVector> calculate(size_t num_graphs) {

            this->dc.cout() << "Running RSP kernel." << std::endl;
            return create_binned_vectors(this->g1, num_graphs);
        }

    private:

        /**
         * Creates a vector with binned shortest path values for a graph.
         */
        std::vector<SparseVector> create_binned_vectors(GraphType& graph, size_t num_graphs) {

            if(log_edge_lengths) {
                dc.cout() << "Setting log edges lenghts." << std::endl;
                MaxValue<double> max_edge_length = graph.map_reduce_edges<MaxValue<double> >(HalfShortestPathKernel::get_max_edge_length);
                dc.cout() << "Found max_edge_length = " << max_edge_length.get_value() << std::endl;
                graph.transform_edges(boost::bind(&HalfShortestPathKernel::set_edge_log_length, _1, max_edge_length.get_value()));
            }

            if(inv_edge_lengths) {
                dc.cout() << "Inverting edges lenghts." << std::endl;
                graph.transform_edges(HalfShortestPathKernel::set_edge_inv_length);
            }

            graphlab::omni_engine<ShortestPathProgram> engine(this->dc, graph, engine_type, this->clopts);
            engine.signal_all();
            engine.start();

            distance_type max_sp_length = graph.map_reduce_vertices<MaxValue<distance_type> >(HalfShortestPathKernel::get_max_sp_length).get_value();
            dc.cout() << "Found max_sp_length = " << max_sp_length << std::endl;

            this->dc.cout() << "binning_type: " << binning_type << std::endl;
            this->dc.cout() << "num_bins: " << num_bins << std::endl;
            boost::function<float(float)> binning_func;
            if(binning_type == "lin") {
                binning_func = boost::bind(&Histogram::linear_map, _1, num_bins, max_sp_length);
            } else if(binning_type == "log") {
                binning_func = boost::bind(&Histogram::logarithmic_map, _1, num_bins, max_sp_length);
            } else if(binning_type == "revlog") {
                binning_func = boost::bind(&Histogram::reverse_logarithmic_map, _1, num_bins, max_sp_length);
            } else if(binning_type == "default") {
                binning_func = boost::bind(&Histogram::default_map, _1);
                this->dc.cout() << "binning_default_map" << std::endl;
            } else {
                throw MtltQuitException("Invalid binning type specified: " + binning_type);
            }
            
            // -------------------------
            // Finished calculating shortest paths, now create the histogram.
            // -------------------------
            SummableMap<Histogram> smh = graph.map_reduce_vertices<SummableMap<Histogram> >(
                    boost::bind(&HalfShortestPathKernel::create_sp_histogram, _1, binning_func));

            std::vector<SparseVector> half_kernels(num_graphs, SparseVector(k));
            size_t max_u_size = 0;
            for(SummableMap<Histogram>::map_type::iterator it=smh.map.begin();it!=smh.map.end();it++) {

                const size_t cluster_id = it->first;
                const Histogram h = it->second;
                const SparseVector u1 = h.to_vector();
                max_u_size = std::max(max_u_size, u1.size());
                half_kernels[cluster_id] = u1;
            }
            dc.cout() << "Found max_u_size: " << max_u_size << std::endl;
            for(size_t i=0;i<num_graphs;i++) {
                half_kernels[i].conservativeResize(max_u_size);
            }

            dc.cout() << "half_kernels[0] =\n" << half_kernels[0] << std::endl;

            return half_kernels;
        }
        
        static MaxValue<double> get_max_edge_length(const GraphType::edge_type& edge) {
            return MaxValue<double>(edge.data().length);
        }

        static void set_edge_log_length(GraphType::edge_type& edge, double max_edge_length) {
            edge.data().length = std::max(0.0, log(max_edge_length) - log(edge.data().length));
        }

        static void set_edge_inv_length(GraphType::edge_type& edge) {
            edge.data().length = 1.0/edge.data().length;
        }

        static MaxValue<distance_type> get_max_sp_length(const sp_graph_type::vertex_type& vertex) {
            sp_map_type source_dists = vertex.data().source_dists;
            distance_type max_len = 0;
            FOR_EACH(sp_map_type, source_dists, it) {

                distance_type sp_dist = it->second;
                max_len = std::max(max_len, sp_dist);
            }

            return MaxValue<distance_type>(max_len);
        }

        /**
         * \brief Creates a histogram of binned shortests path values.
         */
        static SummableMap<Histogram> create_sp_histogram(const sp_graph_type::vertex_type& vertex, boost::function<float(float)> binFunc) {

            Histogram h;

            sp_map_type source_dists = vertex.data().source_dists;
            FOR_EACH(sp_map_type, source_dists, it) {

                distance_type sp_dist = it->second;
                
                h.add(sp_dist, binFunc);
            }

            SummableMap<Histogram> sm;
            sm.map[vertex.data().cluster_id] = h;

            return sm;
        }

        /*
         * VARIABLES
         */
        unsigned int k; // The length of the restricted shortest path
        std::string binning_type;
        size_t num_bins;
        bool inv_edge_lengths;
        bool log_edge_lengths;
};

#endif
