/**
 * @file
 * @author  Tommi Kerola <kerola@student.chalmers.se>
 * @author  Linus Hermansson <hlinus@student.chalmers.se>
 * @version 1.0
 *
 * @section LICENSE
 *
 * Copyright [2013] [Tommi Kerola, Linus Hermansson]
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an "AS
 *  IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 *  express or implied.  See the License for the specific language
 *  governing permissions and limitations under the License.
 *
 * @section DESCRIPTION
 *
 * Various helper macros and functions.
 */
#ifndef HELPER_HPP
#define HELPER_HPP

#define FOR_EACH(MY_STD_COL_TYPE, MY_STD_COL_VAR, MY_ITERATOR) \
        MY_STD_COL_TYPE::iterator MY_ITERATOR; \
        for(MY_ITERATOR = MY_STD_COL_VAR.begin(); \
		MY_ITERATOR != MY_STD_COL_VAR.end(); \
		MY_ITERATOR++)

#define FOR_EACH_CONST(MY_STD_COL_TYPE, MY_STD_COL_VAR, MY_ITERATOR) \
        MY_STD_COL_TYPE::const_iterator MY_ITERATOR; \
        for(MY_ITERATOR = MY_STD_COL_VAR.begin(); \
		MY_ITERATOR != MY_STD_COL_VAR.end(); \
		MY_ITERATOR++)

#define SAFE_DELETE(x) if(x) delete x; x = NULL;

/**
 * \brief A collection that is summable.
 */
template<typename ValueType>
class SummableMap
#ifdef USING_GRAPHCHI
: graphlab::IS_POD_TYPE
#endif
{
    public:

        typedef typename std::map<size_t, ValueType> map_type;
        map_type map;
        SummableMap() { }

        SummableMap& operator+=(const SummableMap& other) {

            // Add the other message's data to this message.
            FOR_EACH_CONST(typename map_type, other.map, other_it) {
                size_t    other_id    = other_it->first;
                ValueType    other_val   = other_it->second;

                // Add own value if found.
                typename map_type::iterator my_it = map.find(other_id);
                if(my_it != map.end()) {
                    my_it->second += other_val;
                } else {
                    map[other_id] = other_val;
                }
            }

            return *this;
        }

#ifdef USING_GRAPHCHI
#else
        void save(graphlab::oarchive& oarc) const {
            oarc << map;
        }
        void load(graphlab::iarchive& iarc) {
            iarc >> map;
        }
#endif
};


/**
 * \brief A value that when summed, keeps just the maximum of the two terms.
 */
template<typename ValueType>
class MaxValue
#ifdef USING_GRAPHCHI
: graphlab::IS_POD_TYPE
#endif
{
    public:

        MaxValue() :initialized(false) { }
        MaxValue(ValueType val) :initialized(true), value(val) { }

        MaxValue& operator+=(const MaxValue& other) {

            if(other.initialized) {
                if(initialized)
                    value = std::max(value, other.value);
                else
                    value = other.value;
            }

            return *this;
        }

        inline ValueType get_value() { return value; }

#ifdef USING_GRAPHCHI
#else
        void save(graphlab::oarchive& oarc) const {
            oarc << value << initialized;
        }
        void load(graphlab::iarchive& iarc) {
            iarc >> value >> initialized;
        }
#endif
    private:
        ValueType value;
        bool initialized;
};

class Classification {
    public:
        typedef std::pair<size_t, double> rank_type;

        Classification() :
            num_true_pos(0), num_false_pos(0), num_true_neg(0), num_false_neg(0), num_time_trials(0) {}
        Classification(size_t _num_true_pos, size_t _num_false_pos, size_t _num_true_neg, size_t _num_false_neg) :
            num_true_pos(_num_true_pos), num_false_pos(_num_false_pos), num_true_neg(_num_true_neg), num_false_neg(_num_false_neg), num_time_trials(0) {}
        ~Classification() {}

        Classification& operator+=(const Classification& other) {
            num_true_pos += other.num_true_pos;
            num_false_pos += other.num_false_pos;
            num_true_neg += other.num_true_neg;
            num_false_neg += other.num_false_neg;
            train_time_millis += other.train_time_millis;
            num_time_trials += other.num_time_trials;

            return *this;
        }

        /**
         * Sets the train time in milliseconds.
         */
        inline void set_train_time(double _train_time_millis) {
            train_time_millis = _train_time_millis;
            num_time_trials = 1;
        }

        inline void add_prediction_result(size_t index, double pred) {
            if(pred >= 0)
                pos_preds.push_back(rank_type(index, pred));
            else
                neg_preds.push_back(rank_type(index, pred));
        }

        std::vector<rank_type> get_ranked_pos_predictions() {
            // Put small values at the beginning of the list
            std::sort(pos_preds.begin(), pos_preds.end(),
                    boost::bind(&rank_type::second, _1) < boost::bind(&rank_type::second, _2));
            return pos_preds;
        }

        std::vector<rank_type> get_ranked_neg_predictions() {
            // Put large values at the beginning of the list
            std::sort(neg_preds.begin(), neg_preds.end(),
                    boost::bind(&rank_type::second, _1) > boost::bind(&rank_type::second, _2));
            return neg_preds;
        }

        /**
         * \brief Returns a sorted rank list with examples sorted according to their absolute value.
         */
        std::vector<rank_type> get_ranked_abs_predictions() {

            std::vector<rank_type> abs_preds;
            abs_preds.insert(abs_preds.begin(), pos_preds.begin(), pos_preds.end());
            abs_preds.insert(abs_preds.begin(), neg_preds.begin(), neg_preds.end());
            // Put small values at the beginning of the list
            std::sort(abs_preds.begin(), abs_preds.end(),
                    boost::bind(&Classification::abs_rank_comp, _1) < boost::bind(&Classification::abs_rank_comp, _2));
            return abs_preds;
        }

        inline double get_accuracy() {
            double tot = num_true_pos+num_false_pos+num_true_neg+num_false_neg;
            if(tot == 0)
                return 0;
            return (num_true_pos+num_true_neg)/tot;
        }

        inline double get_precision() {
            double tot = num_true_pos+num_false_pos;
            if(tot == 0)
                return 0;
            return (num_true_pos)/tot;
        }

        inline double get_negative_predictive_value() {
            double tot = num_true_neg+num_false_neg;
            if(tot == 0)
                return 0;
            return (num_true_neg)/tot;
        }

        inline double get_recall() {
            double tot = num_true_pos+num_false_neg;
            if(tot == 0)
                return 0;
            return (num_true_pos)/tot;
        }

        inline double get_specificity() {
            double tot = num_true_neg+num_false_pos;
            if(tot == 0)
                return 0;
            return (num_true_neg)/tot;
        }

        inline double get_avg_train_time_millis() {
            if(num_time_trials == 0)
                return 0;
            return train_time_millis / num_time_trials;
        }
    private:
    
        static inline double abs_rank_comp(rank_type r) {
            return fabs(r.second);
        }

        size_t num_true_pos, num_false_pos, num_true_neg, num_false_neg;
        double train_time_millis;
        size_t num_time_trials;
        std::vector<rank_type> pos_preds, neg_preds;
};

#endif
