/**
 * @file
 * @author  Tommi Kerola <kerola@student.chalmers.se>
 * @author  Linus Hermansson <hlinus@student.chalmers.se>
 * @version 1.0
 *
 * @section LICENSE
 *
 * Copyright [2013] [Tommi Kerola, Linus Hermansson]
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an "AS
 *  IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 *  express or implied.  See the License for the specific language
 *  governing permissions and limitations under the License.
 *
 * @section DESCRIPTION
 *
 * Simple histogram implementation.
 */
#ifndef HISTOGRAM_HPP
#define HISTOGRAM_HPP

typedef int value_type;
typedef unsigned long long int counter_type;

class Histogram
#ifdef USING_GRAPHCHI
: graphlab::IS_POD_TYPE
#endif
{
    public:

        typedef std::map<value_type, counter_type> bin_type;

        Histogram() {}

        Histogram& operator+=(const Histogram& other) {

            // Add the other histogram's data to this message.
            FOR_EACH_CONST(bin_type, other.bins, other_it) {

                value_type     other_val    = other_it->first;
                counter_type   other_count  = other_it->second;

                // If we already have a value equal to other_val, set it to the sum of the two.
                bin_type::const_iterator my_it = bins.find(other_val);
                if(my_it != bins.end()) { // If found

                    counter_type my_count = my_it->second;
                    other_count += my_count;
                }

                bins[other_val] = other_count;
            }

            return *this;
        }

        /**
         * \brief Adds x to the correct bin in this histogram.
         * 
         * x            - the value to add to a bin
         * map_to_index - function pointer to a mapping funtion
         *                to use for mapping x to the correct bin.
         */
        template<typename MapperType>
        void add(float x, MapperType map_to_index) {

           value_type index = map_to_index(x); 
           if(bins.count(index))
               bins[index] += 1.0;
           else
               bins[index] = 1.0;

        }

        void add(float x) {
            add(x, Histogram::default_map);
        }

        /**
         * \brief Converts the histogram to a vector.
         */
        SparseVector to_vector() const {

            // Get the last key
            size_t n = bins.rbegin()->first+1;
            SparseVector u(n);
            FOR_EACH_CONST(bin_type, bins, it) {
                u(it->first) = it->second;
            }

            return u;
        }

#ifdef USING_GRAPHCHI
#else
        void save(graphlab::oarchive& oarc) const {
            oarc << bins;
        }
        void load(graphlab::iarchive& iarc) {
            iarc >> bins;
        }
#endif

        /**
         * \brief Trivial mapping functions, just returns x.
         */
        static value_type default_map(float x) { return value_type(x); }

        /**
         * \brief Discrete logarithmic map. Denser grid (left) to finer grid (right).
         */
        static value_type logarithmic_map(float x, size_t num_bins, float max_len) {
            assert(x >= 0);
            assert(max_len >= 0);
            const float eps = 1E-5;
            // Map to discrete value
            return (value_type)floor(pow((num_bins+1), (x/(max_len+eps)) ))-1;
        }

        /**
         * \brief Discrete reverse logarithmic map. Finer grid (left) to denser grid (right).
         */
        static value_type reverse_logarithmic_map(float x, size_t num_bins, float max_len) {
            assert(x >= 0);
            assert(max_len >= 0);
            const float eps = 1E-5;
            // Map to discrete value
            return num_bins - (value_type)floor(pow((num_bins+1), ((max_len-x)/(max_len+eps)) ));
        }

        /**
         * \brief Discrete linear map.
         */
        static value_type linear_map(float x, size_t num_bins, float max_len) {
            const float eps = 1E-5;
            // Map to discrete value
            return (value_type)floor(num_bins * x / (max_len + eps));
        }

    private:

        // std::map containing the bin values of the histogram.
        bin_type bins;

};

#endif
