/**
 * @file
 * @author  Tommi Kerola <kerola@student.chalmers.se>
 * @author  Linus Hermansson <hlinus@student.chalmers.se>
 * @version 1.0
 *
 * @section LICENSE
 *
 * Copyright [2013] [Tommi Kerola, Linus Hermansson]
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an "AS
 *  IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 *  express or implied.  See the License for the specific language
 *  governing permissions and limitations under the License.
 *
 */

#include "main.hpp"

MTLT_MAIN_DECLARATION {

    std::string saveprefix = "result";

#ifdef USING_GRAPHCHI
    graphchi::graphchi_init(argc, argv);

    graphchi::metrics m("mtlt2013");

    // Parameters
    std::string filename    = graphchi::get_option_string("--graph"); // Base filename
    int niters              = graphchi::get_option_int("--niters", 20);
    bool scheduler          = true;                    // Whether the scheduler should be enabled or not.
    std::string filetype     = graphchi::get_option_string("filetype", "edgelist"); // Note: --filetype (double dashes) cannot be used since GraphChi
    // expects a parameter on this format.
    ShortestPathProgram::DIRECTED_SSSP =
        "true" == graphchi::get_option_string(
                "--directed", ShortestPathProgram::DIRECTED_SSSP ? "true" : "false");

    std::cout << "Directed graph: " << (ShortestPathProgram::DIRECTED_SSSP ? "Yes" : "No") << std::endl;

    // Process input file - if not already preprocessed
    int nshards             = graphchi::convert_if_notexists<edge_data>(filename, graphchi::get_option_string("--nshards", "auto"));

    std::vector<vertex_data> * vertices =
        run_graphlab_vertexprogram<ShortestPathProgram>(filename, nshards, niters, scheduler, m, false, false);

    //	error_aggregator final_error =
    //	            run_graphlab_edge_aggregator<pagerank_program, error_aggregator>(filename, nshards,
    //	                                                        error_aggregator::map, error_aggregator::finalize, vertices, m);
    //	std::cout << "Final train error: " << final_error.train_error << std::endl;

    if(saveprefix != "") {
        std::cout << "Saving results to file with prefix \"" << saveprefix << "\"." << std::endl;
        std::ofstream outfile;
        std::string save_filename = saveprefix + "_1_of_1";
        outfile.open(save_filename.c_str()); // Use the same savefile filename as GraphLab produces.

        int id = 0;
        std::vector<vertex_data>::iterator it;
        for(it = vertices->begin();
                it != vertices->end();
                it++) {
            outfile << get_vertex_sp_string<vertex_data>(id++, *it);
        }

        outfile.close();
    }

    SAFE_DELETE(vertices);

    std::cout << "GraphChi finished!\n";
    //	graphchi::metrics_report(m);
#else
    graphlab::mpi_tools::init(argc, argv);
    // Distributed communication layer which is needed by a number
    // of the core GraphLab objects, whether you are running
    // distributed or not.
    graphlab::distributed_control dc;

    dc.cout() << "Hello, master's thesis." << std::endl;

    // Parse command line options
    graphlab::command_line_options clopts("mtlt2013");
    std::string graph1_dir, graph2_dir, engine_type = "sync";
    std::string format = "adj";
    clopts.attach_option("graph1", graph1_dir, "The 1st graph file. Required ");
    clopts.add_positional("graph1");
    clopts.attach_option("graph2", graph2_dir, "The 2nd graph file. Required ");
    clopts.add_positional("graph2");
    clopts.attach_option("format", format, "The graph file format");
    clopts.attach_option("enginetype", engine_type, "Type of GraphLab engine (sync, async or semi).");
    clopts.attach_option("directed", ShortestPathProgram::DIRECTED_SSSP,
            "Treat edges as directed.");

    if(!clopts.parse(argc, argv)) {
        dc.cout() << "Error in parsing command line arguments." << std::endl;
        return EXIT_FAILURE;
    }
    if (graph1_dir == "") {
        dc.cout() << "1st graph not specified. Cannot continue\n";
        return EXIT_FAILURE;
    }
    if (graph2_dir == "") {
        dc.cout() << "2nd graph not specified. Cannot continue\n";
        return EXIT_FAILURE;
    }

    dc.cout() << "Loading graph in format: "<< format << std::endl;

    // Create graphs
    GraphLoader<sp_graph_type> sp_loader1;
    sp_loader1.load(graph1_dir, format, dc, clopts);
    GraphLoader<sp_graph_type> sp_loader2;
    sp_loader2.load(graph2_dir, format, dc, clopts);

    GraphLoader<random_walk_graph_type> rw_loader1;
    rw_loader1.load(graph1_dir, format, dc, clopts);
    GraphLoader<random_walk_graph_type> rw_loader2;
    rw_loader2.load(graph2_dir, format, dc, clopts);

    double kernel_value;

    float lambda = 0.6;
    dc.cout() << "Calculating Random Walk Kernel with parameter lambda = " << lambda << std::endl;
    RandomWalkKernel random_walk_kernel(*rw_loader1.get_graph(), *rw_loader2.get_graph(), dc, engine_type, clopts, lambda);
    kernel_value = random_walk_kernel.calculate();
    dc.cout() << "Kernel calculation complete. Got value: " << kernel_value << std::endl;

    int k=10;
    dc.cout() << "Calculating Restricted SP Kernel with parameter k = " << k << std::endl;
    ShortestPathKernel sp_kernel(*sp_loader1.get_graph(), *sp_loader2.get_graph(), dc, engine_type, clopts, k);
    kernel_value = sp_kernel.calculate();
    dc.cout() << "Kernel calculation complete. Got value: " << kernel_value << std::endl;

    //	if(saveprefix != "")
    //	{
    //	    dc.cout() << "Saving results to file with prefix \"" << saveprefix << "\"." << std::endl;
    //        graph.save(saveprefix,
    //                   shortest_path_writer(),
    //                   false, // do not gzip
    //                   true, // save vertices,
    //                   false, // do not save edges
    //                   1);      // number of files to split into
    //	}

    graphlab::mpi_tools::finalize();
#endif

    return 0;
}
