/**
 * @file
 * @author  Tommi Kerola <kerola@student.chalmers.se>
 * @author  Linus Hermansson <hlinus@student.chalmers.se>
 * @version 1.0
 *
 * @section LICENSE
 *
 * Copyright [2013] [Tommi Kerola, Linus Hermansson]
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an "AS
 *  IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 *  express or implied.  See the License for the specific language
 *  governing permissions and limitations under the License.
 *
 * @section DESCRIPTION
 *
 * Vertex program for calculating matrix powers.
 */
#ifndef MATRIX_POWER_PROGRAM_HPP
#define MATRIX_POWER_PROGRAM_HPP

#include "graphlabchi.hpp"
#include "helper.hpp"

#include "shortest_path_utility.hpp"

/**
 * \brief This class is used as the gather type.
 * It contains a sum of polynomials.
 */
class poly_sum_type
#ifdef USING_GRAPHCHI
: graphlab::IS_POD_TYPE
#endif
{
    public:

        std::map<graphlab::vertex_id_type, double> values;

        poly_sum_type() { }

        /**
         * \brief Creates a poly_sum_type with the specified index set to value.
         */
        poly_sum_type(graphlab::vertex_id_type index,
                double value)
        {
            values[index] = value;
        }

        poly_sum_type& operator+=(const poly_sum_type& other) {

            std::map<graphlab::vertex_id_type, double>::const_iterator other_it;
            for(other_it=other.values.begin();other_it!=other.values.end();other_it++) {
                graphlab::vertex_id_type    other_id   = other_it->first;
                double                      other_val  = other_it->second;

                std::map<graphlab::vertex_id_type, double>::const_iterator my_it = values.find(other_id);
                if(my_it != values.end()) { // If found
                    double my_val = my_it->second;

                    other_val *= my_val;
                }

                values[other_id] = other_val;
            }

            return *this;
        }

        /**
         *  \brief Returns the sum of all the mapped values;
         */
        double get_sum() const {

            double total = 0;
            std::map<graphlab::vertex_id_type, double>::const_iterator it;
            for(it=values.begin();it!=values.end();it++) {
                total += it->second;
            }
            return total;
        }

#ifdef USING_GRAPHCHI
#else
        void save(graphlab::oarchive& oarc) const {
            oarc << values;
        }
        void load(graphlab::iarchive& iarc) {
            iarc >> values;
        }
#endif
};

/**
 * \brief Vertex with a value and a row, col index.
 */
class matrix_vertex_data
#ifdef USING_GRAPHCHI
: graphlab::IS_POD_TYPE
#endif
{
    public:

        double value;
        graphlab::vertex_id_type row;
        graphlab::vertex_id_type col;

        matrix_vertex_data(double _value = 0, graphlab::vertex_id_type _row = -1, graphlab::vertex_id_type _col = -1) :
            value(_value), row(_row), col(_col) { }

#ifdef USING_GRAPHCHI
#else
        void save(graphlab::oarchive& oarc) const {
            oarc << value << row << col;
        }
        void load(graphlab::iarchive& iarc) {
            iarc >> value >> row >> col;
        }
#endif

}; // end of vertex data

/**
 * \brief The edge indicates the adjacency matrix index and it is is a row or col edge.
 */
class matrix_edge_data
#ifdef USING_GRAPHCHI
: graphlab::IS_POD_TYPE
#endif
{
    public:

        graphlab::vertex_id_type index;
        bool is_row_edge;

        matrix_edge_data(graphlab::vertex_id_type _index = -1, bool _is_row_edge = true) : index(_index), is_row_edge(_is_row_edge) { }

#ifdef USING_GRAPHCHI
#else
        void save(graphlab::oarchive& oarc) const {
            oarc << index << is_row_edge;
        }
        void load(graphlab::iarchive& iarc) {
            iarc >> index >> is_row_edge;
        }
#endif
}; // end of edge data

/**
 * \brief The edge indicates the adjacency matrix index and it is is a row or col edge.
 */
class matrix_vertex_msg
#ifdef USING_GRAPHCHI
: graphlab::IS_POD_TYPE
#endif
{
    public:

        bool initialize;

        matrix_vertex_msg(bool _initialize = false) : initialize(_initialize) { }

        matrix_vertex_msg& operator+=(const matrix_vertex_msg& other) {

            // Make sure initialization always takes precedence.
            if(!initialize)
                initialize = other.initialize;

            return *this;
        }

#ifdef USING_GRAPHCHI
#else
        void save(graphlab::oarchive& oarc) const {
            oarc << initialize;
        }
        void load(graphlab::iarchive& iarc) {
            iarc >> initialize;
        }
#endif
}; // end of edge data

typedef graphlab::distributed_graph<matrix_vertex_data, matrix_edge_data> matrix_power_graph_type;

/**
 * \brief Vertex program for calculating matrix powers.
 */
class MatrixPowerProgram :
    public graphlab::ivertex_program<
    matrix_power_graph_type,
    poly_sum_type,
    matrix_vertex_msg> // gather_type NOTE: Cannot be graphlab::empty when using GraphChi.
#ifdef USING_GRAPHCHI
    , graphlab::IS_POD_TYPE
#endif
{
    public:

        bool should_set_value;

        MatrixPowerProgram() : should_set_value(false) {}
        virtual ~MatrixPowerProgram() {}

        void init(icontext_type& context, const vertex_type& vertex,
                const matrix_vertex_msg& msg) {
            if(msg.initialize)
                should_set_value = true;
        } 

        edge_dir_type gather_edges(icontext_type& context,
                const vertex_type& vertex) const {
            return graphlab::IN_EDGES;
        }

        gather_type gather(icontext_type& context, const vertex_type& vertex,
                edge_type& edge) const {

            const vertex_type other = get_other_vertex(edge, vertex);

            std::cout << "[" << vertex.id() << "] GATHER edge.data(): " << edge.data().index << " other.data(): " << other.data().value << std::endl;
            poly_sum_type result(edge.data().index, other.data().value);
            
            if(edge.data().is_row_edge && edge.data().index == vertex.data().col)
                result += poly_sum_type(vertex.data().col, vertex.data().value);
            else if(!edge.data().is_row_edge && edge.data().index == vertex.data().row)
                result += poly_sum_type(vertex.data().row, vertex.data().value);

            return result;
        }

        void apply(icontext_type& context, vertex_type& vertex,
                const gather_type& total) {

            double new_value = total.get_sum();
            std::cout << "[" << vertex.id() << "] APPLY new_value: " << new_value << " vertex.data().value: " << vertex.data().value << std::endl;

            // Set new value.
            vertex.data().value = new_value;

            std::cout << "[" << vertex.id() << "] SET initial value 1.0 " << std::endl;
            if(should_set_value)
                vertex.data().value = 1.0;
        }

        edge_dir_type scatter_edges(icontext_type& context,
                const vertex_type& vertex) const {
            return graphlab::NO_EDGES;
        }

        /**
         * \brief The scatter function does nothing.
         */
        void scatter(icontext_type& context, const vertex_type& vertex,
                edge_type& edge) const {
        } // end of scatter

#ifdef USING_GRAPHCHI
#else
        void save(graphlab::oarchive& oarc) const {
        }
        void load(graphlab::iarchive& iarc) {
        }
#endif
};

#endif

