/**
 * @file
 * @author  Tommi Kerola <kerola@student.chalmers.se>
 * @author  Linus Hermansson <hlinus@student.chalmers.se>
 * @version 1.0
 *
 * @section LICENSE
 *
 * Copyright [2013] [Tommi Kerola, Linus Hermansson]
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an "AS
 *  IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 *  express or implied.  See the License for the specific language
 *  governing permissions and limitations under the License.
 *
 * @section DESCRIPTION
 *
 * Helper class for muting stdout and stderr output.
 */
#ifndef OUTPUT_MUTER_HPP
#define OUTPUT_MUTER_HPP

#include <streambuf>
#include <ostream>

class OutputMuter {

    public:
        OutputMuter(bool _mute = false) : muted(_mute), finfo(NULL), fnull(NULL) {

            if(_mute) {
                muted = false;
                mute();
            }
        }

        ~OutputMuter() {
            cleanup();
        }

        void mute() {
            if(!muted) {
                // Redirect cout and cerr.
                cleanup();
                oldCoutStreamBuf = std::cout.rdbuf();
                oldCerrStreamBuf = std::cerr.rdbuf();

                finfo = new std::ostream(oldCoutStreamBuf);
                fnull = new std::ofstream("/dev/null");

                std::cout.rdbuf( fnull->rdbuf() );
                std::cerr.rdbuf( fnull->rdbuf() );
                muted = true;
                cout() << "[NOTE] Disabling normal stdout and stderr." << std::endl;
            }
        }

        void restore() {
            if(muted) {
                // Restore old cout and cerr.
                std::cout.rdbuf( oldCoutStreamBuf );
                std::cerr.rdbuf( oldCerrStreamBuf );
                cleanup();
                muted = false;
                cout() << "[NOTE] stdout and stderr restored." << std::endl;
            }
        }

        std::ostream& cout() const {
            if(muted)
                return *finfo;
            else
                return std::cout;
        }

    private:
        void cleanup() {

            if(finfo != NULL)
                delete finfo;
            finfo = NULL;
            if(fnull != NULL)
                delete fnull;
            fnull = NULL;
        }

        bool muted;
        std::streambuf* oldCoutStreamBuf;
        std::streambuf* oldCerrStreamBuf;
        std::ostream*   finfo;
        std::ofstream*  fnull;
};

#endif
