// Distributed under GNU General Public License (see license.txt for details).
//
//  Copyright (c) 2007 Shai Shalev-Shwartz.
//  All Rights Reserved.
//
// Main for the PEGASOS algorithm.
//
//
// 
// Pegasos: Fly high in the primal 
// main file
//
// 
/**
 * @file
 * @author  Tommi Kerola <kerola@student.chalmers.se>
 * @author  Linus Hermansson <hlinus@student.chalmers.se>
 * @version 1.0
 *
 * @section LICENSE
 *
 * Copyright [2013] [Tommi Kerola, Linus Hermansson / Chalmers University of Technology]
 * (this interface file only, the Pegasos algorithm code belongs to Shai Shalev-Shwartz)
 *
 * [License text here]
 *
 * @section DESCRIPTION
 *
 * Interface to train the Pegasos SVM algorithm.
 */
#ifndef PEGASOS_TRAIN_H
#define PEGASOS_TRAIN_H

//*****************************************************************************
// Included Files
//*****************************************************************************
#include "cmd_line.h"
#include "simple_sparse_vec_hash.h"
#include "pegasos_optimize.h"

class PegasosInterface {
    public:
    /**
     *  Creates an interface to the Pegasos SVM subgradient solver.
     *  Default parameter values are:
     *  lambda = 0.001
     *  max_iter = int(100/lambda)
     *  exam_per_iter = 1
     */
    PegasosInterface(double _lambda, int _max_iter, int _exam_per_iter) :
        lambda(_lambda), max_iter(_max_iter), exam_per_iter(_exam_per_iter) {

        experiments_file = "noExperimentsFile";
        model_filename = "noModelFile";
    }

    ~PegasosInterface() {}

    /**
     *  Trains Pegasos kernelized version and computes validation set error.
     */
    double train_kernelized(std::vector<SparseVector>& trainset_half_kernels, std::vector<SparseVector>& valset_half_kernels,
                        std::vector<int> train_labels, std::vector<int> val_labels) {

/*
        // -------------------------------------------------------------
        // ---------------------- add bias elements---------------------
        // -------------------------------------------------------------
        for(std::vector<std::vector<std::vector<double> > >::iterator it=trainset.begin();it!=trainset.end();it++)
            add_bias_element(*it);
        for(std::vector<std::vector<std::vector<double> > >::iterator it=valset.begin();it!=valset.end();it++)
            add_bias_element(*it);
*/
        // -------------------------------------------------------------
        // ---------------------- Main Learning function ---------------
        // -------------------------------------------------------------
        long trainTime,calc_obj_time;
        double obj_value,norm_value,loss_value,zero_one_error,test_loss,test_error;
        LearnKernelizedReturnLast(trainset_half_kernels,train_labels,valset_half_kernels,val_labels,
                lambda,max_iter,exam_per_iter,
                model_filename,
                trainTime,calc_obj_time,obj_value,norm_value,
                loss_value,zero_one_error,
                test_loss,test_error,
                0,0.0,0,0.0);

        // -------------------------------------------------------------
        // ---------------------- Print Results ------------------------
        // -------------------------------------------------------------
        std::cout
            << "[train] time: " << trainTime << "s" << std::endl
            << "[train] time for calculating objective: " << calc_obj_time << "s" << std::endl 
            << "[train] norm: " << norm_value << std::endl 
            << "[train] primal objective: " << obj_value << std::endl
            << "[train] avg loss: " << loss_value << std::endl  
            << "[train] avg 1-0 error: " << zero_one_error << std::endl 
            << " [test] avg loss: " << test_loss << std::endl 
            << " [test] avg 1-0 error: " << test_error << std::endl
            <<  std::endl;

        return test_error;
    }
    /**
     *  Trains Pegasos and computes validation set error.
     */
    double train(std::vector<std::vector<std::vector<double> > >& trainset, std::vector<std::vector<std::vector<double> > >& valset, int num_line_values) {

/*
        // -------------------------------------------------------------
        // ---------------------- add bias elements---------------------
        // -------------------------------------------------------------
        for(std::vector<std::vector<std::vector<double> > >::iterator it=trainset.begin();it!=trainset.end();it++)
            add_bias_element(*it);
        for(std::vector<std::vector<std::vector<double> > >::iterator it=valset.begin();it!=valset.end();it++)
            add_bias_element(*it);
*/

        // -------------------------------------------------------------
        // ---------------------- read the data ------------------------
        // -------------------------------------------------------------
        uint dimension = 0;
        std::vector<simple_sparse_vector> Dataset;
        std::vector<int> Labels;
        long readingTime;
        ReadDataInternal(trainset,num_line_values,Dataset,Labels,dimension,readingTime);

        uint testDimension = 0;
        std::vector<simple_sparse_vector> testDataset;
        std::vector<int> testLabels;
        long testReadingTime;
        ReadDataInternal(valset,num_line_values,testDataset,testLabels,testDimension,testReadingTime);
        std::cerr << readingTime+testReadingTime << " = Time for reading the data" <<  std::endl;

        // -------------------------------------------------------------
        // ---------------------- Main Learning function ---------------
        // -------------------------------------------------------------
        long trainTime,calc_obj_time;
        double obj_value,norm_value,loss_value,zero_one_error,test_loss,test_error;
        LearnReturnLast(Dataset,Labels,dimension,testDataset,testLabels,
                lambda,max_iter,exam_per_iter,
                model_filename,
                trainTime,calc_obj_time,obj_value,norm_value,
                loss_value,zero_one_error,
                test_loss,test_error,
                0,0.0,0,0.0);

        // -------------------------------------------------------------
        // ---------------------- Print Results ------------------------
        // -------------------------------------------------------------
        std::cout << trainTime << " = Time for training\n" 
            << calc_obj_time << " = Time for calculate objective\n" 
            << norm_value  << " = Norm of solution\n" 
            << loss_value << " = avg Loss of solution\n"  
            << zero_one_error  << " = avg zero-one error of solution\n" 
            << obj_value << " = primal objective of solution\n" 
            << test_loss << " = avg Loss over test\n"  
            << test_error  << " = avg zero-one error over test\n" 	    
            <<  std::endl;

        return test_error;
    }

    private:

        /**
         * \brief Adds a constant 1.0 to a dataset to act as a bias in Pegasos.
         */
        void add_bias_element(std::vector<std::vector<double> >& dataset) {

            for(std::vector<std::vector<double> >::iterator it=dataset.begin();it!=dataset.end();it++)
                it->push_back(1.0); // Add constant 1.0 to act as bias.
        }

        std::string model_filename;
        std::string experiments_file;
        double lambda;
        int max_iter;
        int exam_per_iter;
        //uint num_iter_to_avg = 100;
};

#endif
