/**
 * @file
 * @author  Tommi Kerola <kerola@student.chalmers.se>
 * @author  Linus Hermansson <hlinus@student.chalmers.se>
 * @version 1.0
 *
 * @section LICENSE
 *
 * Copyright [2013] [Tommi Kerola, Linus Hermansson]
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an "AS
 *  IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 *  express or implied.  See the License for the specific language
 *  governing permissions and limitations under the License.
 *
 * @section DESCRIPTION
 *
 * Serial half-kernelized Pegasos implementation with active learning.
 */
#ifndef PEGASOS_HALF_KERNEL_ACTIVE_LEARNING_HPP
#define PEGASOS_HALF_KERNEL_ACTIVE_LEARNING_HPP

#include "sparse_vector.hpp"

#include <boost/bind.hpp>
#include <cmath>

#include "graphlabchi.hpp"

#include "ouput_muter.hpp"
#include "shortest_path_utility.hpp"

#include "pegasos_interface.hpp"

typedef unsigned long long int ULLONG;

class PegasosHalfKernelActiveLearning : public PegasosInterface {

    public:

        /**
         * \brief Construct a new Pegasos kernelized active learning object.
         */
        PegasosHalfKernelActiveLearning(double _pegasos_lambda,
                        size_t _num_iters,
                        size_t _minibatch_size,
                        graphlab::distributed_control& _dc,
                        graphlab::command_line_options& _clopts,
                        std::string _engine_type)
                :   PegasosInterface(_pegasos_lambda, _num_iters, _minibatch_size, _dc, _clopts, _engine_type)
                    {
        }

        virtual ~PegasosHalfKernelActiveLearning() {
        }
        
        Classification train(std::vector<SparseVector>& trainset_half_kernels, std::vector<SparseVector>& valset_half_kernels,
                        std::vector<int> train_labels, std::vector<int> val_labels) {

            dc.cout() << "Training Pegasos (" << num_iters << " iters)" << std::endl;
            num_graphs = trainset_half_kernels.size();
            dimension = trainset_half_kernels[0].size();

            const size_t b = 4;

            dc.cout() << "Pegasos active learning with b = " << b << std::endl;

            graphlab::timer timer;
            timer.start();

            std::vector<SparseVector> active_half_kernels;
            std::vector<int> active_labels;
            std::vector<SparseVector> remaining_half_kernels(trainset_half_kernels);
            std::vector<int> remaining_labels(train_labels);

            bool found_pos = false, found_neg = false;
            // Make sure we have at least one example from each class in the trainset.
            for(std::vector<int>::iterator it=train_labels.begin(); it!=train_labels.end(); it++) {
                if(*it == 1)
                    found_pos = true;
                if(*it == -1)
                    found_neg = true;
                if(found_pos && found_neg)
                    break;
            }
            assert(found_pos);
            assert(found_neg);

            // Add b random examples from each class to begin with
            for(size_t i=0;i<b;i++) {

                size_t index;

                // Positive class
                do {
                    index = size_t(rand()) % remaining_half_kernels.size(); 
                } while(remaining_labels[index] != 1);
                add_to_active(index, active_half_kernels, active_labels, remaining_half_kernels, remaining_labels);

                // Negative class
                do {
                    index = size_t(rand()) % remaining_half_kernels.size();
                } while(remaining_labels[index] != -1);
                add_to_active(index, active_half_kernels, active_labels, remaining_half_kernels, remaining_labels);
            }

            SparseVector gamma;
            bool should_stop = false;
            while(1) {

                std::pair<SparseVector, Classification> result = train_one_pass(active_half_kernels, remaining_half_kernels, valset_half_kernels, active_labels, remaining_labels, val_labels);

                if(should_stop) {
                    gamma = result.first;
                    break;
                }

                Classification c = result.second;
                
                // Get ranks
                std::vector<std::pair<size_t, double> > ranks = c.get_ranked_abs_predictions();

                // Add b elements closest to the hyperplane
                int removed_inds[b];
                for(size_t i=0;i<b;i++) {
                    size_t index = ranks[i].first;
                    const double dist_to_hyperplane = fabs(ranks[i].second);

                    removed_inds[i] = index;
                    // Update index to account for add_to_active removing indices
                    const size_t old_index = index;
                    for(size_t j=0;j<i;j++) {
                        if(old_index > removed_inds[j])
                            index--;
                    }

                    if(dist_to_hyperplane >= 1.0 || remaining_half_kernels.size() == 0) {
                        should_stop = true;
                        break;
                    }

                    dc.cout() << "Adding " << index << " to active (" << dist_to_hyperplane << ")" << std::endl;

                    add_to_active(index, active_half_kernels, active_labels, remaining_half_kernels, remaining_labels);
                }

                dc.cout()
                    << "num_active: " << active_half_kernels.size() << " of " << trainset_half_kernels.size() << std::endl
                    << "train_acc: " << 1.0-zero_one_error << std::endl
                    << " test_acc: " << 1.0-test_error <<  std::endl
                    << std::endl;

            }

            train_time = timer.current_time();

            double eta = 1.0/(pegasos_lambda*num_iters);
            dc.cout() << "eta:" << eta << std::endl;

            timer.start();

            // Calculate squared norm value (w^T*w)
            calc_obj_value(gamma, train_labels, train_labels, eta, pegasos_lambda, trainset_half_kernels, trainset_half_kernels, loss_value, zero_one_error, norm_value, obj_value);

            calc_obj_time = timer.current_time();

            // Calculate test_loss and test_error
            Classification c = calc_loss_and_error(gamma, val_labels, train_labels, eta, valset_half_kernels, trainset_half_kernels, test_loss, test_error);
            c.set_train_time(train_time*1000);

            // -------------------------------------------------------------
            // ---------------------- Print Results ------------------------
            // -------------------------------------------------------------
            dc.cout() << train_time << " = Time for training\n" 
                << calc_obj_time << " = Time for calculate objective\n" 
                << norm_value  << " = Norm of solution\n" 
                << loss_value << " = avg Loss of solution\n"  
                << zero_one_error  << " = avg zero-one error of solution\n" 
                << obj_value << " = primal objective of solution\n" 
                << test_loss << " = avg Loss over test\n"  
                << test_error  << " = avg zero-one error over test\n" 	    
                <<  std::endl;

            return c;
        }

    private:

        std::pair<SparseVector, Classification> train_one_pass(
                std::vector<SparseVector>& active_half_kernels, // the set of training examples selected so far
                std::vector<SparseVector>& remaining_half_kernels,
                std::vector<SparseVector>& valset_half_kernels,
                std::vector<int> active_labels,
                std::vector<int> remaining_labels,
                std::vector<int> val_labels) {

            const size_t num_active = active_half_kernels.size();

            // Get new gamma
            SparseVector gamma(dimension);

            for(size_t t=0; t < num_iters; t++) {

                // Select random index to update
                const size_t index = size_t(rand()) % num_active; 

                const double eta = 1.0 / (pegasos_lambda * (t+1));
                const SparseVector ui = active_half_kernels[index];
                const int yi = active_labels[index];
                const double loss = eta * yi * ui.dot(gamma);

                if(loss < 1)
                    gamma += ui*yi;

            }

            const double eta = 1.0/(pegasos_lambda*num_iters);
            Classification train_c = calc_loss_and_error(gamma, remaining_labels, active_labels, eta, remaining_half_kernels, active_half_kernels, loss_value, zero_one_error);
            
            // Calculate test error just to be informative.
            calc_loss_and_error(gamma, val_labels, active_labels, eta, valset_half_kernels, active_half_kernels, test_loss, test_error);

            return std::pair<SparseVector, Classification>(gamma, train_c);
        }

        /**
         * \brief Adds the element of the specified index of remaining_half_kernels to active_half_kernels and
         * removes the element from the remaining ones.
         */
        inline void add_to_active(size_t index,
                std::vector<SparseVector>& active_half_kernels, std::vector<int>& active_labels,
                std::vector<SparseVector>& remaining_half_kernels, std::vector<int>& remaining_labels) {

                active_half_kernels.push_back(remaining_half_kernels[index]);
                active_labels.push_back(remaining_labels[index]);
                remaining_half_kernels.erase(remaining_half_kernels.begin()+index);
                remaining_labels.erase(remaining_labels.begin()+index);
        }

        /*
         * VARIABLES
         */
};

#endif
