/**
 * @file
 * @author  Tommi Kerola <kerola@student.chalmers.se>
 * @author  Linus Hermansson <hlinus@student.chalmers.se>
 * @version 1.0
 *
 * @section LICENSE
 *
 * Copyright [2013] [Tommi Kerola, Linus Hermansson]
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an "AS
 *  IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 *  express or implied.  See the License for the specific language
 *  governing permissions and limitations under the License.
 *
 * @section DESCRIPTION
 *
 * Pegasos interface base class.
 */
#ifndef PEGASOS_INTERFACE_HPP
#define PEGASOS_INTERFACE_HPP

#include "sparse_vector.hpp"

#include <boost/bind.hpp>
#include <cmath>

#include "graphlabchi.hpp"
#include "graphlab/rpc/dht.hpp"

#include "ouput_muter.hpp"
#include "shortest_path_utility.hpp"

typedef unsigned long long int ULLONG;

class PegasosInterface {

    public:

        /**
         * \brief Construct a new Pegasos kernelized parallel object.
         */
        PegasosInterface(double _pegasos_lambda,
                        size_t _num_iters,
                        size_t _minibatch_size,
                        graphlab::distributed_control& _dc,
                        graphlab::command_line_options& _clopts,
                        std::string _engine_type)
                :   dc(_dc), clopts(_clopts), engine_type(_engine_type), num_iters(_num_iters), minibatch_size(_minibatch_size),
                    pegasos_lambda(_pegasos_lambda) {
        }

        virtual ~PegasosInterface() {
        }
        
        /**
         * \brief Trains the svm and reports validation error.
         */
        virtual Classification train(std::vector<SparseVector>& trainset_half_kernels, std::vector<SparseVector>& valset_half_kernels,
                        std::vector<int> train_labels, std::vector<int> val_labels) = 0;

    protected:

        Classification calc_obj_value(
                // in
                const SparseVector& gamma, const std::vector<int>& labels, const std::vector<int>& train_labels, const double eta, const double lambda,
                const std::vector<SparseVector>& dataset, const std::vector<SparseVector>& train_dataset,
                // out
                double& loss, double& error, double& norm_value, double& obj_value) {

            // Calculate squared norm value.
            norm_value = (gamma*eta).dot(gamma*eta); // Be careful of numerical difficulties here.

            // Calculate objective value
            obj_value = norm_value * lambda / 2.0;
            norm_value = sqrt(norm_value); // turn squared norm into norm
            Classification c = calc_loss_and_error(gamma, labels, train_labels, eta, dataset, train_dataset, loss, error);
            obj_value += loss;
            return c;
        }

        Classification calc_loss_and_error(
                // in
                const SparseVector& gamma, const std::vector<int>& labels, const std::vector<int>& train_labels, double eta,
                const std::vector<SparseVector>& dataset, const std::vector<SparseVector>& train_dataset,
                // out
                double& loss, double& error) {

            //std::cout << "Began calculating loss and error!" << std::endl;

            std::vector<std::pair<size_t, double> > predictions;

            size_t tp=0,fp=0,tn=0,fn=0;
            size_t num_sv=0;
            loss = 0.0;
            error = 0.0;
            dc.cout() << "SV: ";
            for (size_t i=0; i < dataset.size(); ++i) {
                const SparseVector ui = dataset.at(i);
                const double prediction = eta * ui.dot(gamma);
                double cur_loss = 1 - labels[i]*prediction;
//                std::cout << "wx < 1: " << labels[i]*prediction << " < 1" << std::endl;
                if (cur_loss < 0.0) cur_loss = 0.0;
                if (cur_loss >= 1.0) {
                    error += 1.0/dataset.size();
                }

                predictions.push_back(std::pair<size_t, double>(i, prediction));

                if(labels[i]*prediction < 1.0) { // support vectors cross the margin
                    num_sv++;
                    dc.cout() << i << ", ";
                }
                if(labels[i] == 1) { // Ambiguous
                    if(prediction > 0)
                        tp++;
                    else
                        fn++;
                } else { // -1, Unambiguous
                    if(prediction < 0)
                        tn++;
                    else
                        fp++;
                }

                loss += cur_loss/dataset.size();
            }
            dc.cout() << std::endl << "num_sv: " << num_sv << std::endl;
            Classification c(tp, fp, tn, fn);
            for(std::vector<std::pair<size_t, double> >::iterator it=predictions.begin();it!=predictions.end();it++)
                c.add_prediction_result(it->first, it->second);

            return c;
        }

        /*
         * VARIABLES
         */
        size_t num_iters;
        size_t minibatch_size;
        double pegasos_lambda;
        size_t num_graphs;
        size_t dimension;
        graphlab::distributed_control& dc;
        graphlab::command_line_options& clopts;
        std::string engine_type;

        // Distributed objects.
        //graphlab::dht<size_t, SparseVector> half_kernels_with_labels;
        SparseVector* half_kernels_with_labels;

        // Stats
        long train_time;
        long calc_obj_time;
        double obj_value;
        double norm_value;
        double loss_value;
        double zero_one_error;
        double test_loss;
        double test_error;

};

#endif
