/**
 * @file
 * @author  Tommi Kerola <kerola@student.chalmers.se>
 * @author  Linus Hermansson <hlinus@student.chalmers.se>
 * @version 1.0
 *
 * @section LICENSE
 *
 * Copyright [2013] [Tommi Kerola, Linus Hermansson]
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an "AS
 *  IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 *  express or implied.  See the License for the specific language
 *  governing permissions and limitations under the License.
 *
 * @section DESCRIPTION
 *
 * Serial naive kernelized Pegasos implementation (as seen in Shalev-Shwartz 2011)
 */
#ifndef PEGASOS_NAIVE_KERNEL_SINGLE_HPP
#define PEGASOS_NAIVE_KERNEL_SINGLE_HPP

#include "sparse_vector.hpp"

#include <boost/bind.hpp>
#include <cmath>

#include "graphlabchi.hpp"

#include "ouput_muter.hpp"
#include "shortest_path_utility.hpp"

#include "pegasos_interface.hpp"

typedef unsigned long long int ULLONG;

class PegasosNaiveKernelSingle : public PegasosInterface {

    public:

        /**
         * \brief Construct a new Pegasos kernelized parallel object.
         */
        PegasosNaiveKernelSingle(double _pegasos_lambda,
                        size_t _num_iters,
                        size_t _minibatch_size,
                        graphlab::distributed_control& _dc,
                        graphlab::command_line_options& _clopts,
                        std::string _engine_type)
                :   PegasosInterface(_pegasos_lambda, _num_iters, _minibatch_size, _dc, _clopts, _engine_type)
                    {
        }

        virtual ~PegasosNaiveKernelSingle() {
        }
        
        Classification train(std::vector<SparseVector>& trainset_half_kernels, std::vector<SparseVector>& valset_half_kernels,
                        std::vector<int> train_labels, std::vector<int> val_labels) {

            dc.cout() << "Training Pegasos (" << num_iters << " iters)" << std::endl;
            num_graphs = trainset_half_kernels.size();
            dimension = trainset_half_kernels[0].size();

            graphlab::timer timer;
            timer.start();

            SparseVector alpha(num_graphs);

            for(size_t t=0; t < num_iters; t++) {

                // Select random index to update
                const size_t index = size_t(rand()) % num_graphs; 

                const double eta = 1.0 / (pegasos_lambda * (t+1));

                // calc loss
                double loss = 0.0;
                for(size_t j=0; j<num_graphs; j++) {
                    if(alpha[j] != 0) {
                        double kernel_val = trainset_half_kernels[index].dot(trainset_half_kernels[j]);
                        loss += (eta * alpha[j]) * train_labels[j] * kernel_val;
                    }
                }

                loss *= train_labels[index];

                if(loss < 1.0)
                    alpha[index] += 1.0;

            }

            train_time = timer.current_time_millis();

            SparseVector gamma(dimension);
            for(size_t j=0; j<num_graphs; j++) {
                if(alpha[j] != 0) {
                    gamma += trainset_half_kernels[j] * alpha[j] * train_labels[j];
                }
            }
            double eta = 1.0/(pegasos_lambda*num_iters);
            dc.cout() << "eta:" << eta << std::endl;

            timer.start();

            // Calculate squared norm value (w^T*w)
            calc_obj_value(gamma, train_labels, train_labels, eta, pegasos_lambda, trainset_half_kernels, trainset_half_kernels, loss_value, zero_one_error, norm_value, obj_value);

            calc_obj_time = timer.current_time_millis();

            // Calculate test_loss and test_error
            Classification c = calc_loss_and_error(gamma, val_labels, train_labels, eta, valset_half_kernels, trainset_half_kernels, test_loss, test_error);
            c.set_train_time(train_time*1000);

            // -------------------------------------------------------------
            // ---------------------- Print Results ------------------------
            // -------------------------------------------------------------
            dc.cout() << train_time << " = Time for training\n" 
                << calc_obj_time << " = Time for calculate objective\n" 
                << norm_value  << " = Norm of solution\n" 
                << loss_value << " = avg Loss of solution\n"  
                << zero_one_error  << " = avg zero-one error of solution\n" 
                << obj_value << " = primal objective of solution\n" 
                << test_loss << " = avg Loss over test\n"  
                << test_error  << " = avg zero-one error over test\n" 	    
                <<  std::endl;

            return c;
        }

    private:

        /*
         * VARIABLES
         */
};

#endif
