/**
 * @file
 * @author  Tommi Kerola <kerola@student.chalmers.se>
 * @author  Linus Hermansson <hlinus@student.chalmers.se>
 * @version 1.0
 *
 * @section LICENSE
 *
 * Copyright [2013] [Tommi Kerola, Linus Hermansson]
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an "AS
 *  IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 *  express or implied.  See the License for the specific language
 *  governing permissions and limitations under the License.
 *
 * @section DESCRIPTION
 *
 * Parallel kernelized Pegasos implementation.
 */
#ifndef PARALLEL_PEGASOS_HPP
#define PARALLEL_PEGASOS_HPP

#include "sparse_vector.hpp"

#include <boost/bind.hpp>
#include <cmath>

#include "graphlabchi.hpp"
#include <graphlab/rpc/dht.hpp>

typedef unsigned long long int ULLONG;

/**
 * \brief This class is used as the gather type.
 * It contains a collection that is summable.
 */
class beta_map
#ifdef USING_GRAPHCHI
: graphlab::IS_POD_TYPE
#endif
{
    public:

        std::map<size_t, double> betas;
        beta_map() { }

        beta_map& operator+=(const beta_map& other) {

            typedef std::map<size_t, double> betas_type;
            // Add the other message's data to this message.
            FOR_EACH_CONST(betas_type, other.betas, other_it) {
                size_t    other_id    = other_it->first;
                double    other_val   = other_it->second;

                betas[other_id] = other_val;
            }

            return *this;
        }

#ifdef USING_GRAPHCHI
#else
        void save(graphlab::oarchive& oarc) const {
            oarc << betas;
        }
        void load(graphlab::iarchive& iarc) {
            iarc >> betas;
        }
#endif
};

typedef graphlab::distributed_graph<beta_map, graphlab::empty> pegasos_graph_type;

/**
 * \brief Vertex program for calculating kernelized Pegasos alpha values.
 */
class PegasosProgram :
    public graphlab::ivertex_program<
    pegasos_graph_type,
    int,        // gather_type NOTE: Cannot be graphlab::empty when using GraphChi.
    size_t>     // message_type
#ifdef USING_GRAPHCHI
    , graphlab::IS_POD_TYPE
#endif
{
    public:

        static size_t t;
        static size_t num_graphs;
        static double pegasos_lambda;
        static graphlab::dht<size_t, size_t>* alphas_p;
        static graphlab::dht<size_t, SparseVector>* half_kernels_p;
        static graphlab::dht<size_t, int>* labels_p;

        size_t index; // The index that this node should compute.

        PegasosProgram() {}
        virtual ~PegasosProgram() {}

        void init(icontext_type& context,
                const vertex_type& vertex,
                const message_type& msg) {

            index = msg;
        }

        edge_dir_type gather_edges(icontext_type& context,
                const vertex_type& vertex) const {
            return graphlab::NO_EDGES;
        }

        gather_type gather(icontext_type& context, const vertex_type& vertex,
                edge_type& edge) const {
            return 0;
        }

        void apply(icontext_type& context, vertex_type& vertex,
                const gather_type& total) {

            SparseVector uin = half_kernels_p->get(index).second; // Normalized half-kernels.
            int yi = labels_p->get(index).second; 

            double beta = 0;
            for(size_t j=0; j<num_graphs; j++) {

                int aj = alphas_p->get(j).second; 
                if(aj != 0) {
                    SparseVector ujn = half_kernels_p->get(j).second;
                    int yj = labels_p->get(j).second; 

                    double kernel_value = uin.dot(ujn);
                    double extra_loss = yi * (1 / (pegasos_lambda * (t+1+index))) * aj * yj * kernel_value;
                    beta += extra_loss;
                }
            }

            // Set betas
            vertex.data().betas[index] = beta;
        }

        edge_dir_type scatter_edges(icontext_type& context,
                const vertex_type& vertex) const {
            return graphlab::NO_EDGES;
        }

        /**
         * \brief The scatter function does nothing.
         */
        void scatter(icontext_type& context, const vertex_type& vertex,
                edge_type& edge) const {
        } // end of scatter

#ifdef USING_GRAPHCHI
#else
        void save(graphlab::oarchive& oarc) const {
            oarc << index;
        }
        void load(graphlab::iarchive& iarc) {
            iarc >> index;
        }
#endif
};

size_t PegasosProgram::t;
size_t PegasosProgram::num_graphs;
double PegasosProgram::pegasos_lambda;
graphlab::dht<size_t, size_t>* PegasosProgram::alphas_p;
graphlab::dht<size_t, SparseVector>* PegasosProgram::half_kernels_p;
graphlab::dht<size_t, int>* PegasosProgram::labels_p;

class PegasosParallel {

    public:

        /**
         * \brief Construct a new Pegasos kernelized parallel object.
         */
        PegasosParallel(double _pegasos_lambda,
                        size_t _num_iters,
                        size_t _minibatch_size,
                        graphlab::distributed_control& _dc,
                        graphlab::command_line_options& _clopts,
                        std::string _engine_type)
                :   dc(_dc), clopts(_clopts), engine_type(_engine_type), num_iters(_num_iters), minibatch_size(_minibatch_size),
                    pegasos_lambda(_pegasos_lambda), node_graph_p(NULL),
                    alphas(graphlab::dht<size_t, size_t>(dc)),
                    half_kernels(graphlab::dht<size_t, SparseVector>(dc)),
                    labels(graphlab::dht<size_t, int>(dc)) {

            dc.barrier();
        }

        virtual ~PegasosParallel() {
        }
        
        double train(std::vector<SparseVector>& trainset_half_kernels, std::vector<SparseVector>& valset_half_kernels,
                        std::vector<int> train_labels, std::vector<int> val_labels) {

            num_graphs = trainset_half_kernels.size();

            create_node_graph();

            // Init DHT's
            alphas.clear();
            half_kernels.clear();
            labels.clear();
            for(size_t i=0; i<num_graphs; i++) {
                alphas.set(i, 0);
                half_kernels.set(i, trainset_half_kernels[i]);
                labels.set(i, train_labels[i]);
            }

            PegasosProgram::num_graphs = num_graphs;
            PegasosProgram::pegasos_lambda = pegasos_lambda;
            PegasosProgram::alphas_p = &alphas;
            PegasosProgram::half_kernels_p = &half_kernels;
            PegasosProgram::labels_p = &labels;

            // Redirect cout and cerr.
            std::streambuf* oldCoutStreamBuf = std::cout.rdbuf();
            std::ofstream fout("/dev/null");
            std::cout.rdbuf( fout.rdbuf() );
            std::streambuf* oldCerrStreamBuf = std::cerr.rdbuf();
            std::cerr.rdbuf( fout.rdbuf() );
            std::iostream finfo(oldCoutStreamBuf);

            graphlab::timer timer;
            graphlab::timer part_timer;
            double part1=0;
            double part2=0;
            double part3=0;
            double part4=0;
            double part5=0;
            timer.start();

            size_t counter = 0;
            for(size_t t=0; t < num_iters; t+=minibatch_size) {

                part_timer.start();
                PegasosProgram::t = t;

                // Choose indices uniformly, at random.
                std::vector<size_t> selected_indices;
                for(size_t j=0; j<minibatch_size; j++)
                    selected_indices.push_back(rand() % num_graphs);

                part1 += part_timer.current_time_millis();
                part_timer.start();

                // Init vertices
                graphlab::omni_engine<PegasosProgram> engine(dc, *node_graph_p, engine_type, clopts);
                for(size_t j=0; j<minibatch_size; j++)
                    engine.signal(j, selected_indices[j]);

                part2 += part_timer.current_time_millis();
                part_timer.start();

                // Start engine
                engine.start();

                part3 += part_timer.current_time_millis();
                part_timer.start();

                // Get betas
                beta_map betas = engine.map_reduce_vertices<beta_map>(boost::bind(&PegasosParallel::get_beta, this, _1, _2));

                part4 += part_timer.current_time_millis();
                part_timer.start();

                // Update alphas
                do_post_update(selected_indices, betas, t, num_graphs);

                part5 += part_timer.current_time_millis();

                if((counter+1) % 100 == 0) {

                    finfo << ".";
                    finfo.flush();
                    if((counter+1) % (30*100) == 0)
                        finfo << " " << 100.0*(double)(t+1)/num_iters << "%" << std::endl;
                }
                counter++;
            }

            train_time = timer.current_time();

            // Restore old cout and cerr.
            std::cout.rdbuf( oldCoutStreamBuf );
            std::cerr.rdbuf( oldCerrStreamBuf );

            std::cout << "part1: " << part1/(train_time*1000.0) << std::endl;
            std::cout << "part2: " << part2/(train_time*1000.0) << std::endl;
            std::cout << "part3: " << part3/(train_time*1000.0) << std::endl;
            std::cout << "part4: " << part4/(train_time*1000.0) << std::endl;
            std::cout << "part5: " << part5/(train_time*1000.0) << std::endl;

            double eta = 1.0/(pegasos_lambda*num_iters);
            double alpha_update_ratio = 0;
            std::cout << "eta:" << eta << std::endl;
            std::cout << "alpha: ";
            for(size_t i=0; i<num_graphs;i++) {
                std::cout << alphas.get(i).second << " ";
                alpha_update_ratio += alphas.get(i).second;
            }
            std::cout << std::endl;

            if(num_iters != 0)
                alpha_update_ratio /= num_iters;
            std::cout << "alpha_update_ratio: " << alpha_update_ratio << std::endl;

            timer.start();

            // Calculate squared norm value (w^T*w)
            calc_obj_value(alphas, train_labels, train_labels, eta, pegasos_lambda, trainset_half_kernels, trainset_half_kernels, loss_value, zero_one_error, norm_value, obj_value);

            calc_obj_time = timer.current_time();

            // Calculate test_loss and test_error
            calc_loss_and_error(alphas, val_labels, train_labels, eta, valset_half_kernels, trainset_half_kernels, test_loss, test_error);

            // -------------------------------------------------------------
            // ---------------------- Print Results ------------------------
            // -------------------------------------------------------------
            std::cout << train_time << " = Time for training\n" 
                << calc_obj_time << " = Time for calculate objective\n" 
                << norm_value  << " = Norm of solution\n" 
                << loss_value << " = avg Loss of solution\n"  
                << zero_one_error  << " = avg zero-one error of solution\n" 
                << obj_value << " = primal objective of solution\n" 
                << test_loss << " = avg Loss over test\n"  
                << test_error  << " = avg zero-one error over test\n" 	    
                <<  std::endl;

            return test_error;
        }

    private:

        void create_node_graph() {

            node_graph_p = new graphlab::distributed_graph<beta_map, graphlab::empty>(dc, clopts);
            for(size_t i=0; i<minibatch_size; i++) {
                node_graph_p->add_vertex(i);
            }

            node_graph_p->finalize();

        }

        beta_map get_beta(  graphlab::omni_engine<PegasosProgram>::icontext_type& context,
                            graphlab::distributed_graph<beta_map, graphlab::empty>::vertex_type vertex) {
            return vertex.data();
        }

        void do_post_update(const std::vector<size_t>& selected_indices, const beta_map& betas, size_t t, size_t num_graphs) {

            std::vector<size_t> c(minibatch_size, 0);
            for(size_t i=0; i<minibatch_size; i++) {
                double loss = betas.betas.at(selected_indices[i]);
                for(size_t j=0; j<i; j++) {
                    if(c[j] != 0) {
                        SparseVector uin = half_kernels.get(selected_indices[i]).second; // Normalized half-kernels.
                        SparseVector ujn = half_kernels.get(selected_indices[j]).second;
                        int yi = labels.get(selected_indices[i]).second; 
                        int yj = labels.get(selected_indices[j]).second; 

                        double kernel_value = uin.dot(ujn);
                        double extra_loss = yi * (1 / (pegasos_lambda * (t+1+i))) * yj * kernel_value;
                        loss += c[j] * extra_loss;
                    }
                }

                if(loss < 1) {
                    c[i] = 1;
                }
            }


            // c is computed, now update alphas.
            for(size_t i=0; i<minibatch_size; i++) {
                if(c[i] != 0) {
                    size_t j = selected_indices[i];
                    size_t prev_alpha = alphas.get(j).second;
                    alphas.set(j, prev_alpha + 1);
                }
            }

        }

        void calc_obj_value(
                // in
                const graphlab::dht<size_t, size_t>& alpha, const std::vector<int>& Labels, const std::vector<int>& train_labels, const double eta, const double lambda,
                const std::vector<SparseVector>& dataset, const std::vector<SparseVector>& train_dataset,
                // out
                double& loss, double& error, double& norm_value, double& obj_value) {

            norm_value = 0;
            for(size_t i=0; i<dataset.size(); i++) {
                norm_value += pow(eta*alpha.get(i).second*Labels[i], 2); // kernel_val is 1 here, so don't include it
            }
            for(size_t i=0; i<dataset.size(); i++) {
                for(size_t j=i+1; j<train_dataset.size(); j++) {
                    double kernel_val = dataset[i].dot(train_dataset[j]);
                    norm_value += 2 * (eta*alpha.get(i).second)*(eta*alpha.get(j).second)*Labels[i]*train_labels[j]*kernel_val; // We need to be careful with the order of multiplication here to avoid numerical difficulties.
                }
            }

            // Calculate objective value
            obj_value = norm_value * lambda / 2.0;
            norm_value = sqrt(norm_value); // turn squared norm into norm
            calc_loss_and_error(alpha, Labels, train_labels, eta, dataset, train_dataset, loss, error);
            obj_value += loss;
        }

        void calc_loss_and_error(
                // in
                const graphlab::dht<size_t, size_t>& alpha, const std::vector<int>& Labels, const std::vector<int>& train_labels, double eta,
                const std::vector<SparseVector>& dataset, const std::vector<SparseVector>& train_dataset,
                // out
                double& loss, double& error) {

            //std::cout << "Began calculating loss and error!" << std::endl;

            loss = 0.0;
            error = 0.0;
            for (size_t i=0; i < dataset.size(); ++i) {
                double prediction = 0;
                for(size_t j=0; j<train_dataset.size(); j++) {
                    if(alpha.get(j).second != 0) {
                        double kernel_val = dataset.at(i).dot(train_dataset.at(j));
                        prediction += eta * alpha.get(j).second * train_labels[j] * kernel_val;
                    }
                }   
                //std::cout << "y: " << Labels[i] << ", p: " << prediction << ", x: ("<< dataset.at(i)(0) << ", "<< dataset.at(i)(1) <<")" << std::endl; 
                double cur_loss = 1 - Labels[i]*prediction;
                if (cur_loss < 0.0) cur_loss = 0.0;
                if (cur_loss >= 1.0) error += 1.0/dataset.size();
                loss += cur_loss/dataset.size();
            }
        }

        /*
         * VARIABLES
         */
        size_t num_iters;
        size_t minibatch_size;
        double pegasos_lambda;
        size_t num_graphs;
        graphlab::distributed_control& dc;
        graphlab::command_line_options& clopts;
        std::string engine_type;
        graphlab::distributed_graph<beta_map, graphlab::empty>* node_graph_p;

        // Distributed objects.
        graphlab::dht<size_t, size_t> alphas;
        graphlab::dht<size_t, SparseVector> half_kernels;
        graphlab::dht<size_t, int> labels;

        // Stats
        long train_time;
        long calc_obj_time;
        double obj_value;
        double norm_value;
        double loss_value;
        double zero_one_error;
        double test_loss;
        double test_error;

};

#endif
