/**
 * @file
 * @author  Tommi Kerola <kerola@student.chalmers.se>
 * @author  Linus Hermansson <hlinus@student.chalmers.se>
 * @version 1.0
 *
 * @section LICENSE
 *
 * Copyright [2013] [Tommi Kerola, Linus Hermansson]
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an "AS
 *  IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 *  express or implied.  See the License for the specific language
 *  governing permissions and limitations under the License.
 *
 * @section DESCRIPTION
 *
 * Parallel kernelized Pegasos implementation.
 */
#ifndef PARALLEL_PEGASOS_HPP
#define PARALLEL_PEGASOS_HPP

#include "sparse_vector.hpp"

#include <boost/bind.hpp>
#include <cmath>

#include "graphlabchi.hpp"
#include "graphlab/rpc/dht.hpp"

#include "ouput_muter.hpp"
#include "shortest_path_utility.hpp"

typedef unsigned long long int ULLONG;
const int center_id = -2;

/**
 * \brief This class is used as the gather type.
 * It contains a collection that is summable.
 */
class beta_map
#ifdef USING_GRAPHCHI
: graphlab::IS_POD_TYPE
#endif
{
    public:

        std::map<size_t, double> betas;
        beta_map() { }

        beta_map& operator+=(const beta_map& other) {

            typedef std::map<size_t, double> betas_type;
            // Add the other message's data to this message.
            FOR_EACH_CONST(betas_type, other.betas, other_it) {
                size_t    other_id    = other_it->first;
                double    other_val   = other_it->second;

                betas[other_id] = other_val;
            }

            return *this;
        }

#ifdef USING_GRAPHCHI
#else
        void save(graphlab::oarchive& oarc) const {
            oarc << betas;
        }
        void load(graphlab::iarchive& iarc) {
            iarc >> betas;
        }
#endif
};

typedef graphlab::distributed_graph<SparseVector, graphlab::empty> pegasos_graph_type;

/**
 * \brief Vertex program for calculating kernelized Pegasos gamma.
 */
class PegasosProgram :
    public graphlab::ivertex_program<
    pegasos_graph_type,
    SparseVector,        // gather_type NOTE: Cannot be graphlab::empty when using GraphChi.
    size_t>     // message_type
#ifdef USING_GRAPHCHI
    , graphlab::IS_POD_TYPE
#endif
{
    public:

        static size_t num_iters;
        static size_t num_graphs;
        static size_t num_machines;
        static double pegasos_lambda;
        static SparseVector* half_kernels_with_labels_p;

        PegasosProgram() {}
        virtual ~PegasosProgram() {}

        void init(icontext_type& context,
                const vertex_type& vertex,
                const message_type& msg) {

        }

        edge_dir_type gather_edges(icontext_type& context,
                const vertex_type& vertex) const {
            return graphlab::IN_EDGES;
        }

        gather_type gather(icontext_type& context, const vertex_type& vertex,
                edge_type& edge) const {
            const vertex_type other = get_other_vertex(edge, vertex);
            if(vertex.id() == center_id)
                return other.data() * (1/num_machines);
            else // normal node
                return other.data();
        }

        void apply(icontext_type& context, vertex_type& vertex,
                const gather_type& total) {

            size_t t = 1+context.iteration();

            if(vertex.id() == center_id) {

                SparseVector gamma;
                if(total.isZero())
                    gamma = vertex.data();
                else
                    gamma = total + (vertex.data() * (1.0 / num_machines));
                vertex.data() = gamma;

            } else {

                // Select random index to update
                size_t index = rand() % num_graphs; 

                // Get new gamma
                SparseVector gamma = total;

                double eta = 1.0 / (pegasos_lambda * t);
                SparseVector ui = half_kernels_with_labels_p[index];
                double loss = eta * ui.dot(gamma);

                if(loss < 1)
                    gamma += ui;

                vertex.data() = gamma;

            }
        }

        edge_dir_type scatter_edges(icontext_type& context,
                const vertex_type& vertex) const {
            if(vertex.id() == center_id) {

                size_t t = 1+context.iteration();
                if(t*2 < num_iters)
                    return graphlab::OUT_EDGES;
                else
                    return graphlab::NO_EDGES;
            } else { // normal node
                return graphlab::OUT_EDGES;
            }
        }

        /**
         * \brief The scatter function does nothing.
         */
        void scatter(icontext_type& context, const vertex_type& vertex,
                edge_type& edge) const {
            const vertex_type other = get_other_vertex(edge, vertex);
            context.signal(other);
        } // end of scatter

        /**
         * \brief Signal all normal vertices.
         */
        static graphlab::empty signal_normal(icontext_type& context,
                const vertex_type& vertex) {
            if(vertex.id() != center_id) context.signal(vertex);
            return graphlab::empty();
        } // end of signal_normal 


#ifdef USING_GRAPHCHI
#else
        void save(graphlab::oarchive& oarc) const {
        }
        void load(graphlab::iarchive& iarc) {
        }
#endif
};

size_t PegasosProgram::num_iters;
size_t PegasosProgram::num_graphs;
size_t PegasosProgram::num_machines;
double PegasosProgram::pegasos_lambda;
SparseVector* PegasosProgram::half_kernels_with_labels_p;

class PegasosParallel {

    public:

        /**
         * \brief Construct a new Pegasos kernelized parallel object.
         */
        PegasosParallel(double _pegasos_lambda,
                        size_t _num_iters,
                        size_t _minibatch_size,
                        graphlab::distributed_control& _dc,
                        graphlab::command_line_options& _clopts,
                        std::string _engine_type)
                :   dc(_dc), clopts(_clopts), engine_type(_engine_type), num_iters(_num_iters), minibatch_size(_minibatch_size),
                    pegasos_lambda(_pegasos_lambda), node_graph_p(NULL)
                   // , half_kernels_with_labels(graphlab::dht<size_t, SparseVector>(_dc))
                     {
        }

        virtual ~PegasosParallel() {
        }
        
        double train(std::vector<SparseVector>& trainset_half_kernels, std::vector<SparseVector>& valset_half_kernels,
                        std::vector<int> train_labels, std::vector<int> val_labels) {

            dc.cout() << "Training Pegasos (" << num_iters << " iters)" << std::endl;
            dc.cout() << "nprocs: " << dc.numprocs() << std::endl;
            dc.cout() << "ncpus: " << clopts.get_ncpus() << std::endl;
            num_graphs = trainset_half_kernels.size();
            num_machines = minibatch_size;
            dimension = trainset_half_kernels[0].size();

            dc.cout() << " - Creating node graph." << std::endl;
            create_node_graph();

            dc.cout() << " - Init DHT's." << std::endl;
            // Init DHT's
            half_kernels_with_labels = new SparseVector[num_graphs];
            for(size_t i=0; i<num_graphs; i++) {
                half_kernels_with_labels[i] = trainset_half_kernels[i] * train_labels[i];
            }

            dc.cout() << " - Setting PegasosProgram values." << std::endl;
            PegasosProgram::num_iters = num_iters;
            PegasosProgram::num_graphs = num_graphs;
            PegasosProgram::num_machines = num_machines;
            PegasosProgram::pegasos_lambda = pegasos_lambda;
            PegasosProgram::half_kernels_with_labels_p = half_kernels_with_labels;

            // Redirect cout and cerr.
            OutputMuter muter(false);

            graphlab::timer timer;
            timer.start();

            graphlab::omni_engine<PegasosProgram> engine(dc, *node_graph_p, engine_type, clopts);

            engine.transform_vertices(boost::bind(&PegasosParallel::set_initial_gamma, this, _1, _2));
            engine.map_reduce_vertices<graphlab::empty>(&PegasosProgram::signal_normal);

            // Start engine
            engine.signal_all();
            engine.start();

            // Get gamma
            SparseVector gamma = engine.map_reduce_vertices<SparseVector>(boost::bind(&PegasosParallel::get_gamma, this, _1, _2));

            train_time = timer.current_time();
            // Restore old cout and cerr.
            muter.restore();

            delete[] half_kernels_with_labels;

            double eta = 1.0/(pegasos_lambda*num_iters);
            dc.cout() << "eta:" << eta << std::endl;

            timer.start();

            // Calculate squared norm value (w^T*w)
            calc_obj_value(gamma, train_labels, train_labels, eta, pegasos_lambda, trainset_half_kernels, trainset_half_kernels, loss_value, zero_one_error, norm_value, obj_value);

            calc_obj_time = timer.current_time();

            // Calculate test_loss and test_error
            calc_loss_and_error(gamma, val_labels, train_labels, eta, valset_half_kernels, trainset_half_kernels, test_loss, test_error);

            // -------------------------------------------------------------
            // ---------------------- Print Results ------------------------
            // -------------------------------------------------------------
            dc.cout() << train_time << " = Time for training\n" 
                << calc_obj_time << " = Time for calculate objective\n" 
                << norm_value  << " = Norm of solution\n" 
                << loss_value << " = avg Loss of solution\n"  
                << zero_one_error  << " = avg zero-one error of solution\n" 
                << obj_value << " = primal objective of solution\n" 
                << test_loss << " = avg Loss over test\n"  
                << test_error  << " = avg zero-one error over test\n" 	    
                <<  std::endl;

            return test_error;
        }

    private:

        void create_node_graph() {

            node_graph_p = new graphlab::distributed_graph<SparseVector, graphlab::empty>(dc, clopts);

            if(dc.procid() == 0) {
                for(size_t i=0; i<num_machines; i++) {
                    node_graph_p->add_edge(i, center_id);
                    node_graph_p->add_edge(center_id, i);
                }
                /*
                if(num_machines > 1) {
                    for(size_t i=0; i<num_machines; i++) {
                        for(size_t j=i+1; j<num_machines; j++) {
                            node_graph_p->add_edge(i, j);
                        }
                    }
                } else { // num_machines == 1
                    node_graph_p->add_vertex(0);
                }
                */
            }

            node_graph_p->finalize();

        }

        void set_initial_gamma(  graphlab::omni_engine<PegasosProgram>::icontext_type& context,
                            graphlab::distributed_graph<SparseVector, graphlab::empty>::vertex_type vertex) {
            vertex.data().conservativeResize(dimension);
        }

        SparseVector get_gamma(  graphlab::omni_engine<PegasosProgram>::icontext_type& context,
                            graphlab::distributed_graph<SparseVector, graphlab::empty>::vertex_type vertex) {
            return vertex.data() * (1.0/num_machines);
        }

        void calc_obj_value(
                // in
                const SparseVector& gamma, const std::vector<int>& Labels, const std::vector<int>& train_labels, const double eta, const double lambda,
                const std::vector<SparseVector>& dataset, const std::vector<SparseVector>& train_dataset,
                // out
                double& loss, double& error, double& norm_value, double& obj_value) {

            // Calculate squared norm value.
            norm_value = (gamma*eta).dot(gamma*eta); // Be careful of numerical difficulties here.

            // Calculate objective value
            obj_value = norm_value * lambda / 2.0;
            norm_value = sqrt(norm_value); // turn squared norm into norm
            calc_loss_and_error(gamma, Labels, train_labels, eta, dataset, train_dataset, loss, error);
            obj_value += loss;
        }

        void calc_loss_and_error(
                // in
                const SparseVector& gamma, const std::vector<int>& Labels, const std::vector<int>& train_labels, double eta,
                const std::vector<SparseVector>& dataset, const std::vector<SparseVector>& train_dataset,
                // out
                double& loss, double& error) {

            //std::cout << "Began calculating loss and error!" << std::endl;

            loss = 0.0;
            error = 0.0;
            for (size_t i=0; i < dataset.size(); ++i) {
                const SparseVector ui = dataset.at(i);
                double prediction = eta * ui.dot(gamma);
                double cur_loss = 1 - Labels[i]*prediction;
                if (cur_loss < 0.0) cur_loss = 0.0;
                if (cur_loss >= 1.0) error += 1.0/dataset.size();
                loss += cur_loss/dataset.size();
            }
        }

        /*
         * VARIABLES
         */
        size_t num_iters;
        size_t minibatch_size;
        size_t num_machines;
        double pegasos_lambda;
        size_t num_graphs;
        size_t dimension;
        graphlab::distributed_control& dc;
        graphlab::command_line_options& clopts;
        std::string engine_type;
        graphlab::distributed_graph<SparseVector, graphlab::empty>* node_graph_p;

        // Distributed objects.
        SparseVector* half_kernels_with_labels;

        // Stats
        long train_time;
        long calc_obj_time;
        double obj_value;
        double norm_value;
        double loss_value;
        double zero_one_error;
        double test_loss;
        double test_error;

};

#endif
