/**
 * @file
 * @author  Tommi Kerola <kerola@student.chalmers.se>
 * @author  Linus Hermansson <hlinus@student.chalmers.se>
 * @version 1.0
 *
 * @section LICENSE
 *
 * Copyright [2013] [Tommi Kerola, Linus Hermansson]
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an "AS
 *  IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 *  express or implied.  See the License for the specific language
 *  governing permissions and limitations under the License.
 *
 * @section DESCRIPTION
 *
 * Product graph kernel.
 */
#ifndef PRODUCT_KERNEL_HPP
#define PRODUCT_KERNEL_HPP

#include "graph_kernel.hpp"

#include <boost/bind.hpp>
#include <cmath>

typedef unsigned long long int ULLONG;

class ProductKernel : public GraphKernel<vertex_type, edge_type> {
    
    public:

        typedef graphlab::distributed_graph<vertex_type, edge_type> GraphType;
    
        /**
         * \brief Constructs a new graph kernel for comparing the two graphs g1 and g2.
         *
         * params:
         * g1, g2 - the graphs
         * dc     - graphlab::distributed_control object.
         * lambda - kernel parameter, the path length decay factor. lambda \in [0,1]
         */
        ProductKernel(GraphType& _g1, GraphType& _g2,
                graphlab::distributed_control& _dc,
                std::string _engine_type,
                graphlab::command_line_options& _clopts,
                float _lambda) :
            GraphKernel<vertex_type, edge_type>(_g1, _g2, _dc, _engine_type, _clopts), lambda(_lambda) {

        }

        virtual ~ProductKernel() {
        }

        /**
         * \brief Calculates the value of the kernel.
         */
        double calculate() {

            graphlab::omni_engine<MatrixPowerProgram>* g1_adj_engine = create_adj_graph(this->g1);
            graphlab::omni_engine<MatrixPowerProgram>* g2_adj_engine = create_adj_graph(this->g2);

            const float LIMIT = 0.005; // Loop until lambda^i becomes smaller than this value.
            int n = (int) (log(LIMIT) / log(lambda)); // Calculate suitable number of times to loop.
            this->dc.cout() << "lambda: " << lambda << ", LIMIT: " << LIMIT << ", n: " << n << std::endl;
            double total = 0;
            for(int i=0; i<=n; i++) {

                // Get the sums for this walk length.
                double g1_val = g1_adj_engine->map_reduce_vertices<double>(
                                    boost::bind(&ProductKernel::get_vertex_data, this, _1, _2));
                double g2_val = g2_adj_engine->map_reduce_vertices<double>(
                                    boost::bind(&ProductKernel::get_vertex_data, this, _1, _2));
                double lambda_val = pow(lambda, i);

                // Run each vertex program one step.
                g1_adj_engine->signal_all();
                g1_adj_engine->start();
                g2_adj_engine->signal_all();
                g2_adj_engine->start();


                this->dc.cout() << "i=" << i << ", lambda_val=" << lambda_val << ", g1_val=" << g1_val << ", g2_val=" << g2_val << std::endl; 

                total += lambda_val * g1_val * g2_val;
            }
            
            SAFE_DELETE(g1_adj_engine);
            SAFE_DELETE(g2_adj_engine);
            return total / (this->g1.num_edges()*this->g2.num_edges());
        }

    private:

        /**
         * \brief Get the data of a vertex.
         */
        double get_vertex_data(graphlab::omni_engine<MatrixPowerProgram>::icontext_type& context, const matrix_power_graph_type::vertex_type& vertex) {
            return vertex.data().value;
        }

        /**
         * \brief Creates an adjacency graph from the specified base graph.
         */
        graphlab::omni_engine<MatrixPowerProgram>* create_adj_graph(GraphType& graph) {

            matrix_power_graph_type adj_graph(this->dc, this->clopts);
            temp_adj_graph = &adj_graph;
            temp_num_verts = graph.num_vertices();

            this->dc.cout() << "create_adj_graph: Adding grid vertices." << std::endl;
            // O(n^3)
            for(ULLONG i=0; i<temp_num_verts; i++) {
                for(ULLONG j=0; j<temp_num_verts; j++) {

                    graphlab::vertex_id_type a = temp_num_verts * i + j; // The vertex

                    // Note: Set all vertices to 0.0 at this phase. 
                    // The vertices that should correctly have 1.0 
                    // will be given this at the next step in transform_edges.
                    temp_adj_graph->add_vertex(a, matrix_vertex_data(0.0, i, j));
                      
                    // Row
                    for(ULLONG k=0; k<temp_num_verts; k++) {

                        if(k == i)
                            continue;

                        graphlab::vertex_id_type b = temp_num_verts * k + j; // Create in-edge from this vertex

                        graphlab::vertex_id_type pair_index = k; // indicates which edge the value from this edge should be paired up with in the matrix_power_program.
                        temp_adj_graph->add_edge(b, a, matrix_edge_data(pair_index, true)); // row edge from b --> a
                    }

                    // Col
                    for(ULLONG l=0; l<temp_num_verts; l++) {

                        if(l == j)
                            continue;

                        graphlab::vertex_id_type b = temp_num_verts * i + l; // Create in-edge from this vertex

                        graphlab::vertex_id_type pair_index = l; // indicates which edge the value from this edge should be paired up with in the matrix_power_program.
                        temp_adj_graph->add_edge(b, a, matrix_edge_data(pair_index, false)); // col edge from b --> a
                    }
                }
            }

            this->dc.barrier(); // Make sure all machines have created the vertices before starting to create edges.

            adj_graph.finalize();
            temp_adj_graph = NULL;
            temp_adj_engine = new graphlab::omni_engine<MatrixPowerProgram>(
                                    this->dc, adj_graph, engine_type, clopts);
            graph.transform_edges(boost::bind(
                        &ProductKernel::signal_matrix_vertices, this, _1));
            return temp_adj_engine;
        }

        /**
         * \brief Creates edges in the adj_graph.
         */
        void signal_matrix_vertices(graphlab::distributed_graph<vertex_data, edge_data>::edge_type& edge) {

            ULLONG i = edge.source().id(); // row
            ULLONG j = edge.target().id(); // col
            
            graphlab::vertex_id_type a = temp_num_verts * i + j; // The vertex
            temp_adj_engine->signal(a, matrix_vertex_msg(true));
        }

        
        /**
         * \brief Creates edges in the adj_graph.
         */
        /*
        void create_adj_graph_edge(graphlab::distributed_graph<vertex_data, edge_data>::edge_type& edge) {

            ULLONG i = edge.source().id(); // row
            ULLONG j = edge.target().id(); // col
            
            graphlab::vertex_id_type a = temp_num_verts * i + j; // The vertex
            this->dc.cout() << "create_adj_graph_edge: Adding vertex [" << i << "," << j << "] = 1.0" << std::endl;
            temp_adj_graph->add_vertex(a, matrix_vertex_data(1.0, i, j)); // 1.0 indicates that there is an edge at this position in the adjacency matrix.

            // Row
            for(ULLONG k=0; k<temp_num_verts; k++) {

                if(k == i)
                    continue;

                graphlab::vertex_id_type b = temp_num_verts * k + j; // Create in-edge from this vertex

                graphlab::vertex_id_type pair_index = k; // indicates which edge the value from this edge should be paired up with in the matrix_power_program.
                temp_adj_graph->add_edge(b, a, matrix_edge_data(pair_index, true)); // row edge from b --> a
            }

            // Col
            for(ULLONG l=0; l<temp_num_verts; l++) {

                if(l == j)
                    continue;

                graphlab::vertex_id_type b = temp_num_verts * i + l; // Create in-edge from this vertex

                graphlab::vertex_id_type pair_index = l; // indicates which edge the value from this edge should be paired up with in the matrix_power_program.
                temp_adj_graph->add_edge(b, a, matrix_edge_data(pair_index, false)); // col edge from b --> a
            }

        }*/

        /*
         * VARIABLES
         */
        float lambda; // path length decay factor. lambda \in [0,1]

        matrix_power_graph_type* temp_adj_graph;
        graphlab::omni_engine<MatrixPowerProgram>* temp_adj_engine;
        ULLONG temp_num_verts;
};

#endif
