/**
 * @file
 * @author  Tommi Kerola <kerola@student.chalmers.se>
 * @author  Linus Hermansson <hlinus@student.chalmers.se>
 * @version 1.0
 *
 * @section LICENSE
 *
 * Copyright [2013] [Tommi Kerola, Linus Hermansson]
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an "AS
 *  IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 *  express or implied.  See the License for the specific language
 *  governing permissions and limitations under the License.
 *
 * @section DESCRIPTION
 *
 * Random walk graph kernel.
 */
#ifndef RANDOM_WALK_KERNEL_HPP
#define RANDOM_WALK_KERNEL_HPP

#include "sparse_vector.hpp"
#include "graph_kernel.hpp"

#include <boost/bind.hpp>
#include <cmath>

typedef unsigned long long int ULLONG;

class RandomWalkKernel : public GraphKernel<ULLONG, graphlab::empty> {
    
    public:

        typedef graphlab::distributed_graph<ULLONG, graphlab::empty> GraphType;
    
        /**
         * \brief Constructs a new graph kernel for comparing the two graphs g1 and g2.
         *
         * params:
         * g1, g2 - the graphs
         * dc     - graphlab::distributed_control object.
         * lambda - kernel parameter, the path length decay factor. lambda \in [0,1]
         */
        RandomWalkKernel(GraphType& _g1, GraphType& _g2,
                graphlab::distributed_control& _dc,
                std::string _engine_type,
                graphlab::command_line_options& _clopts,
                double _lambda) :
            GraphKernel<ULLONG, graphlab::empty>(_g1, _g2, _dc, _engine_type, _clopts), lambda(_lambda) {

        }

        virtual ~RandomWalkKernel() {
        }

        /**
         * \brief Calculates the value of the kernel.
         */
        double calculate() {

            graphlab::omni_engine<RandomWalkProgram> g1_engine(this->dc, this->g1, engine_type, this->clopts);
            graphlab::omni_engine<RandomWalkProgram> g2_engine(this->dc, this->g2, engine_type, this->clopts);

            g1_engine.transform_vertices(boost::bind(&RandomWalkKernel::init_vertex_data, this, _1, _2));
            g2_engine.transform_vertices(boost::bind(&RandomWalkKernel::init_vertex_data, this, _1, _2));

            const double LIMIT = 1e-5; // Loop until lambda^i becomes smaller than this value.
            int n = (int) (log(LIMIT) / log(lambda)); // Calculate suitable number of times to loop.
            this->dc.cout() << "lambda: " << lambda << ", LIMIT: " << LIMIT << ", n: " << n << std::endl;
            SparseVector u1(n);
            SparseVector u2(n);

            for(int i=0; i<n; i++) {

                // Get the sums for this walk length.
                ULLONG g1_val = g1_engine.map_reduce_vertices<ULLONG>(
                                    boost::bind(&RandomWalkKernel::get_vertex_data, this, _1, _2));
                ULLONG g2_val = g2_engine.map_reduce_vertices<ULLONG>(
                                    boost::bind(&RandomWalkKernel::get_vertex_data, this, _1, _2));
                double lambda_val = pow(lambda, i);

                // Run each vertex program one step.
                g1_engine.signal_all();
                g1_engine.start();
                g2_engine.signal_all();
                g2_engine.start();

                this->dc.cout() << "i=" << i << ", lambda_val=" << lambda_val << ", g1_val=" << g1_val << ", g2_val=" << g2_val << std::endl; 

                u1(i) = g1_val * sqrt(lambda_val);
                u2(i) = g2_val * sqrt(lambda_val);
            }

            return u1.dot(u2) / (u1.norm()*u2.norm());
        }

    private:

        /**
         * \brief Get the data of a vertex.
         */
        ULLONG get_vertex_data(graphlab::omni_engine<RandomWalkProgram>::icontext_type& context, const random_walk_graph_type::vertex_type& vertex) {
            return vertex.data();
        }

        /**
         * \brief Initializes the data of a vertex.
         */
        void init_vertex_data(graphlab::omni_engine<RandomWalkProgram>::icontext_type& context, random_walk_graph_type::vertex_type& vertex) {
            vertex.data() = vertex.num_in_edges();
        }
        
        /*
         * VARIABLES
         */
        double lambda; // path length decay factor. lambda \in [0,1]
};

#endif
