/**
 * @file
 * @author  Tommi Kerola <kerola@student.chalmers.se>
 * @author  Linus Hermansson <hlinus@student.chalmers.se>
 * @version 1.0
 *
 * @section LICENSE
 *
 * Copyright [2013] [Tommi Kerola, Linus Hermansson]
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an "AS
 *  IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 *  express or implied.  See the License for the specific language
 *  governing permissions and limitations under the License.
 *
 * @section DESCRIPTION
 *
 * Vertex program for calculating the number of random walks.
 */
#ifndef RANDOM_WALK_PROGRAM_HPP
#define RANDOM_WALK_PROGRAM_HPP

#include "graphlabchi.hpp"
#include "helper.hpp"

#include "shortest_path_utility.hpp"

typedef unsigned long long int ULLONG;

/**
 * \brief The number of walks of the current length to this vertex.
 */
class rw_vertex_data
#ifdef USING_GRAPHCHI
: graphlab::IS_POD_TYPE
#endif
{
    public:
        size_t cluster_id;
        bool distinguished_vertex;

        ULLONG num_walks;
        rw_vertex_data(size_t _cluster_id = 0, bool _distinguished_vertex = false) : cluster_id(_cluster_id), distinguished_vertex(_distinguished_vertex), num_walks(0) { }

#ifdef USING_GRAPHCHI
#else
        void save(graphlab::oarchive& oarc) const {
            oarc << num_walks << cluster_id << distinguished_vertex;
        }
        void load(graphlab::iarchive& iarc) {
            iarc >> num_walks >> cluster_id >> distinguished_vertex;
        }
#endif

}; // end of vertex data

typedef graphlab::distributed_graph<rw_vertex_data, double> random_walk_graph_type;

/**
 * \brief Vertex program for calculating the number of random walks.
 */
class RandomWalkProgram :
    public graphlab::ivertex_program<
    random_walk_graph_type,
    ULLONG> // gather_type NOTE: Cannot be graphlab::empty when using GraphChi.
#ifdef USING_GRAPHCHI
    , graphlab::IS_POD_TYPE
#endif
{
    public:

        RandomWalkProgram() {}
        virtual ~RandomWalkProgram() {}

        edge_dir_type gather_edges(icontext_type& context,
                const vertex_type& vertex) const {
            return MTLT_DIRECTED_GRAPH ? graphlab::IN_EDGES : graphlab::ALL_EDGES;
        }

        gather_type gather(icontext_type& context, const vertex_type& vertex,
                edge_type& edge) const {

            const vertex_type other = get_other_vertex(edge, vertex);

            return other.data().num_walks;
        }

        void apply(icontext_type& context, vertex_type& vertex,
                const gather_type& total) {

            vertex.data().num_walks = total;
        }

        edge_dir_type scatter_edges(icontext_type& context,
                const vertex_type& vertex) const {
            return graphlab::NO_EDGES;
        }

        /**
         * \brief The scatter function does nothing.
         */
        void scatter(icontext_type& context, const vertex_type& vertex,
                edge_type& edge) const {
        } // end of scatter

#ifdef USING_GRAPHCHI
#else
        void save(graphlab::oarchive& oarc) const {
        }
        void load(graphlab::iarchive& iarc) {
        }
#endif
};

#endif

