/**
 * @file
 * @author  Tommi Kerola <kerola@student.chalmers.se>
 * @author  Linus Hermansson <hlinus@student.chalmers.se>
 * @version 1.0
 *
 * @section LICENSE
 *
 * Copyright [2013] [Tommi Kerola, Linus Hermansson]
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an "AS
 *  IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 *  express or implied.  See the License for the specific language
 *  governing permissions and limitations under the License.
 *
 * @section DESCRIPTION
 *
 * Shortest path graph kernel.
 */
#ifndef SHORTEST_PATH_KERNEL_HPP
#define SHORTEST_PATH_KERNEL_HPP

#include "sparse_vector.hpp"

#include "graph_kernel.hpp"
//#include "shortest_path_program.hpp"
#include "shortest_path_types.hpp"
#include "histogram.hpp"

#include <boost/bind.hpp>

class ShortestPathProgram;

class ShortestPathKernel : public GraphKernel<vertex_data, edge_data> {
    
    typedef graphlab::distributed_graph<vertex_data, edge_data> GraphType;
    
    public:

        /**
         * Constructs a new graph kernel for comparing the two graphs g1 and g2.
         *
         * params:
         * g1, g2 - the graphs
         * dc     - graphlab::distributed_control object.
         * k      - kernel parameter, the maximum length of a path.
         */
        ShortestPathKernel(GraphType& _g1, GraphType& _g2,
                graphlab::distributed_control& _dc,
                std::string _engine_type,
                graphlab::command_line_options& _clopts,
                unsigned int _k) :
            GraphKernel<vertex_data, edge_data>(_g1, _g2, _dc, _engine_type, _clopts), k(_k) {

            ShortestPathProgram::max_path_length = _k;
        }

        /**
         * Calculates the value of the kernel.
         */
        double calculate() {

            SparseVector u1 = create_binned_vector(this->g1);
            SparseVector u2 = create_binned_vector(this->g2);

            double cos_theta = u1.dot(u2) / (u1.norm() * u2.norm());

            return cos_theta;
        }

    private:

        /**
         * Creates a vector with binned shortest path values for a graph.
         */
        SparseVector create_binned_vector(GraphType& graph) {

            graphlab::omni_engine<ShortestPathProgram> engine(this->dc, graph, engine_type, this->clopts);
            engine.signal_all();
            engine.start();

            const float runtime = engine.elapsed_seconds();
            this->dc.cout() << "Finished Running engine in " << runtime << " seconds." << std::endl;

            // -------------------------
            // Finished calculating shortest paths, now create the hash tables.
            // -------------------------
            Histogram h = graph.map_reduce_vertices<Histogram>(
                    boost::bind(&ShortestPathKernel::create_sp_histogram, this, _1));

            SparseVector u1 = h.to_vector();

            return u1;
        }
        
        /**
         * Creates a histogram of binned shortests path values.
         */
        Histogram create_sp_histogram(const sp_graph_type::vertex_type& vertex) {

            Histogram h;

            sp_map_type source_dists = vertex.data().source_dists;
            FOR_EACH(sp_map_type, source_dists, it) {

                distance_type sp_dist = it->second;

                h.add(sp_dist);
            }

            return h;
        }

        /*
         * VARIABLES
         */
        unsigned int k; // The length of the restricted shortest path
};

#endif
