/**
 * @file
 * @author  Tommi Kerola <kerola@student.chalmers.se>
 * @author  Linus Hermansson <hlinus@student.chalmers.se>
 * @version 1.0
 *
 * @section LICENSE
 *
 * Copyright [2013] [Tommi Kerola, Linus Hermansson]
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an "AS
 *  IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 *  express or implied.  See the License for the specific language
 *  governing permissions and limitations under the License.
 *
 * @section DESCRIPTION
 *
 * Vertex program for calculating all-pairs shortest paths (APSP).
 */
#ifndef SHORTEST_PATH_PROGRAM_HPP
#define SHORTEST_PATH_PROGRAM_HPP

#include "graphlabchi.hpp"
#include "helper.hpp"
#include "shortest_path_types.hpp"
#include "shortest_path_utility.hpp"
#include "shortest_path_vertex.hpp"
#include "shortest_path_edge.hpp"

/**
 * \brief Vertex program for calculating the shortest path between all nodes.
 */
class ShortestPathProgram :
		public graphlab::ivertex_program<sp_graph_type,
                        sp_collection_type> // gather_type NOTE: Cannot be graphlab::empty when using GraphChi.
#ifdef USING_GRAPHCHI
		, graphlab::IS_POD_TYPE
#endif
{
public:

	bool changed;

    static distance_type max_path_length;

			 ShortestPathProgram() : changed(false) {}
	virtual ~ShortestPathProgram() {}

	edge_dir_type gather_edges(icontext_type& context,
								const vertex_type& vertex) const {
		return MTLT_DIRECTED_GRAPH ? graphlab::IN_EDGES : graphlab::ALL_EDGES;
	}

	gather_type gather(icontext_type& context, const vertex_type& vertex,
					edge_type& edge) const {
        const vertex_type other = get_other_vertex(edge, vertex);
        return sp_collection_type(other.data().source_dists, edge.data().length);
	}

	void apply(icontext_type& context, vertex_type& vertex,
				const gather_type& msg) {
        changed = false;

		if(vertex.data().has_sent_own_id == false) {
            vertex.data().has_sent_own_id = true;
            set_new_dist(vertex, vertex.id(), 0);
		}

        sp_map_type msg_dist = msg.dist;
        FOR_EACH(sp_map_type, msg_dist, it) {

			graphlab::vertex_id_type    from_id = it->first;
			distance_type               newd    = it->second;

			bool shouldAddWithoutComparing = vertex.data().source_dists.count(from_id) == 0;

			if(shouldAddWithoutComparing || vertex.data().source_dists[from_id] > newd) {

                if(newd <= max_path_length) // Make sure we obey the set maximum path length.
	    			set_new_dist(vertex, from_id, newd);
			}
		}
	}

	void set_new_dist(vertex_type& vertex, graphlab::vertex_id_type from_id, distance_type newd) {
        changed = true;
        vertex.data().source_dists[from_id] = newd;
	}

	edge_dir_type scatter_edges(icontext_type& context,
									const vertex_type& vertex) const {
		if(changed)
			return MTLT_DIRECTED_GRAPH ? graphlab::OUT_EDGES : graphlab::ALL_EDGES;
		else
            return graphlab::NO_EDGES;
	}

	/**
	 * \brief The scatter function just signals adjacent vertices.
	 */
	void scatter(icontext_type& context, const vertex_type& vertex,
			edge_type& edge) const {
		const vertex_type other = get_other_vertex(edge, vertex);
		context.signal(other);
	} // end of scatter

#ifdef USING_GRAPHCHI
#else
	void save(graphlab::oarchive& oarc) const {
		oarc << changed;
	}
	void load(graphlab::iarchive& iarc) {
		iarc >> changed;
	}
#endif
};

#endif

