/**
 * @file
 * @author  Tommi Kerola <kerola@student.chalmers.se>
 * @author  Linus Hermansson <hlinus@student.chalmers.se>
 * @version 1.0
 *
 * @section LICENSE
 *
 * Copyright [2013] [Tommi Kerola, Linus Hermansson]
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an "AS
 *  IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 *  express or implied.  See the License for the specific language
 *  governing permissions and limitations under the License.
 *
 * @section DESCRIPTION
 *
 * Types used for shortest paths.
 */
#ifndef SHORTEST_PATH_TYPES_HPP
#define SHORTEST_PATH_TYPES_HPP

#include <map>

#include "graphlabchi.hpp"
#include "shortest_path_vertex.hpp"
#include "shortest_path_edge.hpp"
#include "helper.hpp"

typedef float distance_type;
typedef std::map<graphlab::vertex_id_type, distance_type> sp_map_type;
typedef graphlab::distributed_graph<vertex_data, edge_data> sp_graph_type;
typedef unsigned long long int counter_type;

/**
 * \brief This class is used as the gather type.
 * It contains a collection with shortest paths.
 */
class sp_collection_type
#ifdef USING_GRAPHCHI
		: graphlab::IS_POD_TYPE
#endif
		  {
public:

  ::sp_map_type dist;

  sp_collection_type() { }

  /**
   * \brief Creates a message with the specified length added to all entries.
   */
  sp_collection_type(sp_map_type _dist,
               distance_type length_to_add)
  {
	  dist = _dist;

	  // Add length
      FOR_EACH(sp_map_type, dist, it) {
          it->second += length_to_add;
      }
  }

  sp_collection_type& operator+=(const sp_collection_type& other) {
	// Add the other message's data to this message.
	FOR_EACH_CONST(sp_map_type, other.dist, other_it) {
		graphlab::vertex_id_type    other_id    = other_it->first;
		distance_type               other_dist  = other_it->second;

		// If we already have a distance from other_id, set it to the smallest one of the two.
		sp_map_type::const_iterator my_it = dist.find(other_id);
		if(my_it != dist.end()) { // If found
		    distance_type my_dist = my_it->second;

//		    logstream(LOG_DEBUG) << "my_dist = " << my_dist << ", other_dist = " << other_dist << std::endl;
			other_dist = std::min(my_dist, other_dist);
//			logstream(LOG_DEBUG) << "other_dist is now " << other_dist << std::endl;
		}

		// Add the distance.
		dist[other_id] = other_dist;
	}

    return *this;
  }

#ifdef USING_GRAPHCHI
#else
  void save(graphlab::oarchive& oarc) const {
		  oarc << dist;
  }
  void load(graphlab::iarchive& iarc) {
		  iarc >> dist;
  }
#endif
};


#endif

