/**
 * @file
 * @author  Tommi Kerola <kerola@student.chalmers.se>
 * @author  Linus Hermansson <hlinus@student.chalmers.se>
 * @version 1.0
 *
 * @section LICENSE
 *
 * Copyright [2013] [Tommi Kerola, Linus Hermansson]
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an "AS
 *  IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 *  express or implied.  See the License for the specific language
 *  governing permissions and limitations under the License.
 *
 * @section DESCRIPTION
 *
 * Linear algebraic vector in sparse format.
 */
#ifndef SPARSE_VECTOR_HPP
#define SPARSE_VECTOR_HPP

#include <cassert>
#include <map>

#include "graphlabchi.hpp"

class SparseVector
#ifdef USING_GRAPHCHI
: graphlab::IS_POD_TYPE
#endif
{
    public:
        typedef std::map<size_t, double> map_type;

        /*
         * \brief Creates a sparse vector with the specified size.
         */
        SparseVector(size_t _size = 1) {
            assert(_size > 0);

            vec_size = _size;
        }

        SparseVector(const SparseVector& other) {
            if(this == &other)
                return;

            vec_size = other.size();
            values = map_type(other.values);
        }

        ~SparseVector() {
        }

        SparseVector& operator=(const SparseVector& other) {
            if(this == &other)
                return *this;

            vec_size = other.size();
            values = map_type(other.values);

            return *this;
        }

        /**
         * \brief Returns the dot product between this SparseVector and another.
         */
        double dot(const SparseVector& other) const {
            assert(vec_size == other.vec_size);
            
            double result = 0;
            for(map_type::const_iterator it=values.begin();
                it!=values.end();
                it++) {

                map_type::const_iterator other_it = other.values.find(it->first);
                if(other_it != other.values.end())
                    result += it->second * other_it->second;
            }

            return result;
        }

        double& operator[](size_t index) {
            assert(index < vec_size);
            return values[index];
        }
        double& operator()(size_t index) {
            assert(index < vec_size);
            return values[index];
        }
        double const operator[](size_t index) const {
            if(values.count(index) > 0)
                return values.at(index);
            else
                return 0;
        }
        double const operator()(size_t index) const {
            if(values.count(index) > 0)
                return values.at(index);
            else
                return 0;
        }

        SparseVector operator/(double denominator) const {
            assert(denominator != 0);
            return (*this)*(1/denominator);
        }

        SparseVector operator*(double factor) const {

            SparseVector result(vec_size);
            for(map_type::const_iterator it=values.begin();
                it!=values.end();
                it++) {

                result[it->first] = it->second * factor;
            }

            return result;
        }

        SparseVector operator-(double term) const {
            return (*this)+(-term);
        }

        SparseVector operator+(double term) const {

            SparseVector result(vec_size);
            for(map_type::const_iterator it=values.begin();
                it!=values.end();
                it++) {

                result[it->first] = it->second + term;
            }

            return result;
        }

        SparseVector operator-(const SparseVector& term) const {
            return (*this)+(-term);
        }

        SparseVector operator+(const SparseVector& term) const {
            assert(vec_size == term.vec_size);

            SparseVector result(*this);
            for(map_type::const_iterator it=term.values.begin();
                it!=term.values.end();
                it++) {

                map_type::const_iterator my_it = values.find(it->first);
                if(my_it != values.end())
                    result[it->first] += it->second;
                else
                    result[it->first]  = it->second;
            }
            return result;
        }

        SparseVector operator-() const {
            return operator*(-1.0);
        }

        SparseVector& operator-=(double term) {
            return operator+=(-term);
        }

        SparseVector& operator+=(double term) {
            for(map_type::iterator it=values.begin();
                it!=values.end();
                it++) {

                it->second += term;
            }
            return *this;
        }

        SparseVector& operator/=(double denominator) {
            assert(denominator != 0);
            return operator*=(1.0/denominator);
        }

        SparseVector& operator*=(double factor) {
            for(map_type::iterator it=values.begin();
                it!=values.end();
                it++) {

                it->second *= factor;
            }
            return *this;
        }

        SparseVector& operator-=(const SparseVector& term) {
            return operator+=(-term);
        }

        SparseVector& operator+=(const SparseVector& term) {
            assert(vec_size == term.vec_size);

            for(map_type::const_iterator it=term.values.begin();
                it!=term.values.end();
                it++) {

                map_type::iterator my_it = values.find(it->first);
                if(my_it != values.end())
                    my_it->second += it->second;
                else
                    values[it->first] = it->second;
            }
            return *this;
        }

        inline double normSqr() const {
            return dot(*this);
        }

        inline double norm() const {

            return sqrt(normSqr());
        }

        inline size_t size() const {
            return vec_size;
        }

        /**
         * \brief Resizes the vector and keeps former values, if possible.
         */
        void conservativeResize(size_t new_size) {
            assert(new_size > 0);

            if(vec_size == new_size)
                return;

            vec_size = new_size;
            values = map_type(values.begin(), values.find(new_size-1));
        }

        /**
         * \brief Check whether this vector contains only zeros.
         */
        bool isZero() const {
            return values.size() == 0;
        }

        friend std::ostream& operator<<(std::ostream& stream, const SparseVector& sv) {
            stream << "[";
            for(size_t i=0;i<sv.vec_size;i++) {
                if(sv.values.count(i) > 0)
                    stream << sv.values.at(i);
                else
                    stream << 0;
                if(i<sv.vec_size-1)
                    stream << ", ";
            }
            stream << "]";

            return stream;
        }


#ifdef USING_GRAPHCHI
#else
        void save(graphlab::oarchive& oarc) const {
            oarc << vec_size;
            oarc << values;
        }
        void load(graphlab::iarchive& iarc) {
            iarc >> vec_size;
            iarc >> values;
        }
#endif

    private:

        size_t vec_size;
        map_type values;

};

#endif
