/**
 * @file
 * @author  Tommi Kerola <kerola@student.chalmers.se>
 * @author  Linus Hermansson <hlinus@student.chalmers.se>
 * @version 1.0
 *
 * @section LICENSE
 *
 * Copyright [2013] [Tommi Kerola, Linus Hermansson]
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an "AS
 *  IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 *  express or implied.  See the License for the specific language
 *  governing permissions and limitations under the License.
 *
 *
 * @section DESCRIPTION
 *
 * Program for training the Pegasos SVM using a GraphLab graph kernel.
 */
#ifndef TRAIN_SVM_HPP
#define TRAIN_SVM_HPP

// Standard includes
#include <iostream>
#include <fstream>
#include <map>
#include <cstdlib>
#include <algorithm>
#include <streambuf>
#include <boost/lexical_cast.hpp>
#include <boost/algorithm/string.hpp>

#include "exception.hpp"

#include <sparse_vector.hpp>

#include "helper.hpp"
#include "graph_loader.hpp"

#include "shortest_path_program.hpp"
#include "random_walk_program.hpp"

#include "half_shortest_path_kernel.hpp"
#include "half_random_walk_kernel.hpp"

#include "pegasos_half_kernel_single.hpp"
#include "pegasos_naive_kernel_single.hpp"
#include "pegasos_half_kernel_active_learning.hpp"

enum KernelType {RSP, DP};

int myrandom(int i) { return rand()%i; }

class Program
{
    public:
    void parse_cmd_opts(std::string& trainfile, std::string& engine_type, std::string& svmlight_outputfile, std::string& halfkernels_outputfile, std::string& svmlight_inputfile, std::string& example_names_file, bool& use_names, KernelType& kerneltype, std::string& precompute_mode,
            std::string& format, double& lambda, bool& dp_distinguished_only, int& k, std::string& rsp_binning_type, size_t& rsp_num_bins, bool& inv_edge_lengths, bool& log_edge_lengths, bool& normalize_half_kernels,
            int& num_folds, bool& cv_just_once, bool& no_cv,
            double& pegasos_lambda, int& pegasos_max_iter, int& pegasos_exam_per_iter, std::string& pegasos_impl,
            graphlab::distributed_control& dc,
            int argc, char** argv,
            graphlab::command_line_options& clopts) {

        std::string kerneltype_str;

        clopts.attach_option("trainfile", trainfile, "File containing training data file locations. Required ");
        clopts.add_positional("trainfile");
        clopts.attach_option("kerneltype", kerneltype_str, "Type of graph kernel to use. Possible values are \"rsp\" (restricted shortest path) or \"dp\" (direct product kernel). Required ");
        clopts.add_positional("kerneltype");
        clopts.attach_option("format", format, "The graph file format");
        clopts.attach_option("precompute_mode", precompute_mode, "The mode for precomputing the graph kernels. (batch or serial or pairwise)");
        clopts.attach_option("normalize_half_kernels", normalize_half_kernels, "Whether half-kernels should be normalized or not (true or false)");
        clopts.attach_option("enginetype", engine_type, "Type of GraphLab engine (sync, async or semi).");

        // Restricted shortest path kernel
        clopts.attach_option("rsp_k", k, "Restricted shortest path kernel parameter. Default: k=10 000");
        clopts.attach_option("rsp_binning", rsp_binning_type, "RSP kernel binning type. Can either be lin ( x=floor(n*s/(s_max+epsilon)) ) or log ( x=floor((n+1)^(s/(s_max+epsilon))-1) ) or revlog ( x=n-floor((n+1)^((s_max-s)/(s_max+epsilon))) ) or default (x=y). Default value=lin");
        clopts.attach_option("rsp_num_bins", rsp_num_bins, "RSP kernel number of bins.");

        // Direct product graph kernel
        clopts.attach_option("dp_lambda", lambda, "Random walk kernel parameter. Default: lambda=0.6");
        clopts.attach_option("dp_distinguished_only", dp_distinguished_only, "If true, only count walks from the distinguished \"Chris Anderson\" vertex with id 0.");

        // Edge transformation options (c_ij is the orignal edge length in the input graph)
        clopts.attach_option("invedges", inv_edge_lengths, "Sets w_ij = 1 / c_ij");
        clopts.attach_option("logedges", log_edge_lengths, "Sets w_ij = max(0, log max_c - log c_ij)");
        clopts.attach_option("directed",  MTLT_DIRECTED_GRAPH, "Whether the graph kernels should treat the graphs as directed or not.");

        // Input/output options
        clopts.attach_option("o", svmlight_outputfile, "If specified, precomputed kernel values will be written to this file in svm-light precomputed kernel format.");
        clopts.attach_option("i", svmlight_inputfile, "If specified, precomputed kernel values will be read from this file in svm-light format and the graph kernels will NOT be computed.");
        clopts.attach_option("ho", halfkernels_outputfile, "If specified, half-kernel values will be written to this file in svm-light format.");
        clopts.attach_option("namesfile", example_names_file, "If specified, names for each of the examples in the trailfile will be read from this file.");
	clopts.attach_option("use_names", use_names,"If set to true, will use the names to increase prediction performance.");

        // Cross-validation options
        clopts.attach_option("v", num_folds, "Set --v n for n-fold cross-validation mode.");
        clopts.attach_option("once", cv_just_once, "Set --once true to make the cross-validation loop run just once.");
        clopts.attach_option("no_cv", no_cv, "Set --no_cv true to not do any cross-validation. Useful when combined with --ho.");

        // Pegasos options
        clopts.attach_option("pegasos_lambda", pegasos_lambda, "Pegasos regularization parameter (default = 0.001)");
        clopts.attach_option("pegasos_iter", pegasos_max_iter, "Pegasos number of iterations (default = 100/lambda)");
        clopts.attach_option("pegasos_k", pegasos_exam_per_iter, "Pegasos size of block for stochastic gradient (default = 1)");
        clopts.attach_option("pegasos_impl", pegasos_impl, "Pegasos implementation to use (half_kernel or naive_kernel or half_kernel_active_learning");

        if(!clopts.parse(argc, argv)) {
            throw MtltQuitException("Error in parsing command line arguments.");
        }
        if (trainfile == "") {
            throw MtltQuitException("Training file not specified. Cannot continue\n");
        }

        // Check kernel type.
        if (kerneltype_str == "rsp") {
            kerneltype = RSP;
        }
        else if (kerneltype_str == "dp") {
            kerneltype = DP;
        }
        else {
            throw MtltQuitException("Unknown kernel type \"" + kerneltype_str + "\" specified. Cannot continue\n");
        }

    }

    void parse_trainfile(std::string trainfile, std::vector<std::string>& graph_files, std::vector<int>& graph_labels,
            graphlab::distributed_control& dc) {

        std::ifstream file(trainfile.c_str());
        if (file.is_open()) {
            std::string line;
            bool first_time = true;
            while (std::getline(file, line)) {

                if(line[0] == '#') // Skip comments
                    continue;

                if(first_time) { // Read labels
                    std::ifstream labelfile(line.c_str());
                    if(labelfile.is_open()) {
                        while (std::getline(labelfile, line)) {
                            if(line[0] == '#') // Skip comments
                                continue;
                            graph_labels.push_back(atoi(line.c_str()));
                        }
                    }
                    labelfile.close();
                    first_time = false;
                }
                else { // Read graph filenames
                    graph_files.push_back(line);
                }
            }
            file.close();
        }

        if (graph_files.empty()) {
            throw MtltQuitException("No graphs specified in trainfile. Cannot continue\n");
        }
        if (graph_labels.empty()) {
            throw MtltQuitException("No labels specified in labelfile. Cannot continue\n");
        }
        if (graph_files.size() != graph_labels.size()) {
            throw MtltQuitException("The number of graphs and labels do not match. Cannot continue\n");
        }
        for(std::vector<int>::iterator it=graph_labels.begin();it!=graph_labels.end();it++) {
            if(*it != 1 && *it != -1) {
                int v = *it;
                std::string msg = std::string("Labels can only be 1 or -1. Found label: ") + boost::lexical_cast<std::string>(v) + std::string(". Cannot continue\n");
                throw MtltQuitException(msg);
            }
        }
    }

    /**
     * \brief Stores the precomputed value from the upper diagonal
     * of the kernel matrix correctly.
     */
    void store_svm_value(std::vector<std::vector<double > >& svm_matrix,
            unsigned int i, unsigned int j, double val) {

        svm_matrix[i][j] = val;
        svm_matrix[j][i] = val;
    }

    void precompute_kernels(
            //out
            std::vector<int>& graph_labels, std::vector<std::vector<double> >& svm_matrix,
            std::vector<SparseVector>& half_kernel_vals, int& num_graphs,
            // in
            KernelType kerneltype, std::string precompute_mode, std::string trainfile, std::string engine_type, double lambda, bool dp_distinguished_only, double k,
            std::string& rsp_binning_type, size_t& rsp_num_bins, bool inv_edge_lengths, bool log_edge_lengths, bool normalize_half_kernels,
            graphlab::distributed_control& dc, graphlab::command_line_options& clopts, std::string format) {
        
        // Read trainfile
        dc.cout() << "Parsing trainfile..." << std::endl;
        std::vector<std::string> graph_files;
        
        parse_trainfile(trainfile, graph_files, graph_labels, dc);
        dc.cout() << graph_files.size() << " graphs loaded." << std::endl;
        dc.cout() << "Starting kernel precomputation phase." << std::endl;

        // Redirect cout and cerr.
        OutputMuter muter(false);

        num_graphs = graph_files.size();
        double kernel_value;
        unsigned int i,j;

        // each row contains [y1, k11, k12, k13, ...]
        svm_matrix = std::vector<std::vector<double> >();
        for(int i=0;i<num_graphs;i++)
            svm_matrix.push_back(std::vector<double>(num_graphs));

        graphlab::timer timer;
        half_kernel_vals = std::vector<SparseVector>(num_graphs);
        
        dc.barrier();
        timer.start();
       
        if(precompute_mode == "serial") {
            dc.cout() << "Precomputing graphs using serial mode. " << std::endl;
            for(size_t i=0;i<num_graphs;i++) {

                dc.cout() << "graph " << i << std::endl;
                std::vector<std::string> tempVec;
                tempVec.push_back(graph_files[i]);
                // Create and calculate half-kernel
                switch(kerneltype) {
                    case DP:
                        {
                            ClusterGraphLoader<random_walk_graph_type> loader1;
                            loader1.load(tempVec, format, dc, clopts);
                            HalfRandomWalkKernel kernel(*loader1.get_graph(), dc, engine_type, clopts, lambda, dp_distinguished_only);
                            half_kernel_vals[i] = kernel.calculate(num_graphs)[0];
                        }   break;
                    case RSP:
                        {
                            ClusterGraphLoader<sp_graph_type> loader1;
                            loader1.load(tempVec, format, dc, clopts);
                            HalfShortestPathKernel kernel(*loader1.get_graph(), dc, engine_type, clopts, k, rsp_binning_type, rsp_num_bins, inv_edge_lengths, log_edge_lengths);
                            half_kernel_vals[i] = kernel.calculate(num_graphs)[0];
                        }   break;
                }
            }
            
            // Resize
            size_t maxsize = 0;
            for(size_t i=0;i<num_graphs;i++) {
                maxsize = std::max(maxsize, half_kernel_vals[i].size());
            }
            for(size_t i=0;i<num_graphs;i++) {
                half_kernel_vals[i].conservativeResize(maxsize);
            }
            dc.cout() << "Resized half-kernels to size " << maxsize << std::endl;

        } else if(precompute_mode == "batch") {

            // Create and calculate half-kernels using batch mode
            switch(kerneltype) {
                case DP:
                    {
                        ClusterGraphLoader<random_walk_graph_type> loader1;
                        loader1.load(graph_files, format, dc, clopts);
                        HalfRandomWalkKernel kernel(*loader1.get_graph(), dc, engine_type, clopts, lambda, dp_distinguished_only);
                        half_kernel_vals = kernel.calculate(num_graphs);
                    }   break;
                case RSP:
                    {
                        ClusterGraphLoader<sp_graph_type> loader1;
                        loader1.load(graph_files, format, dc, clopts);
                        HalfShortestPathKernel kernel(*loader1.get_graph(), dc, engine_type, clopts, k, rsp_binning_type, rsp_num_bins, inv_edge_lengths, log_edge_lengths);
                        half_kernel_vals = kernel.calculate(num_graphs);
                    }   break;
            }
        } else if(precompute_mode == "pairwise") {
            // Naïve pairwise computaton mode
            dc.cout() << "Precomputing graphs using pairwise mode. " << std::endl;
            muter.mute();
            for(size_t i=0;i<num_graphs;i++) {
                muter.cout() << i << " ";
                for(size_t j=i+1; j<num_graphs; j++) {

                    dc.cout() << "graphs " << i << "," << j << std::endl;
                    muter.cout() << ".";
                    muter.cout().flush();
                    std::vector<std::string> tempVec;

                    // Simulate pairwise computation
                    // Graph i
                    tempVec.push_back(graph_files[i]);
                    // Create and calculate half-kernel
                    switch(kerneltype) {
                        case DP:
                            {
                                ClusterGraphLoader<random_walk_graph_type> loader1;
                                loader1.load(tempVec, format, dc, clopts);
                                HalfRandomWalkKernel kernel(*loader1.get_graph(), dc, engine_type, clopts, lambda, dp_distinguished_only);
                                half_kernel_vals[i] = kernel.calculate(num_graphs)[0];
                            }   break;
                        case RSP:
                            {
                                ClusterGraphLoader<sp_graph_type> loader1;
                                loader1.load(tempVec, format, dc, clopts);
                                HalfShortestPathKernel kernel(*loader1.get_graph(), dc, engine_type, clopts, k, rsp_binning_type, rsp_num_bins, inv_edge_lengths, log_edge_lengths);
                                half_kernel_vals[i] = kernel.calculate(num_graphs)[0];
                            }   break;
                    }

                    // Graph j
                    tempVec.clear();
                    tempVec.push_back(graph_files[j]);
                    // Create and calculate half-kernel
                    switch(kerneltype) {
                        case DP:
                            {
                                ClusterGraphLoader<random_walk_graph_type> loader1;
                                loader1.load(tempVec, format, dc, clopts);
                                HalfRandomWalkKernel kernel(*loader1.get_graph(), dc, engine_type, clopts, lambda, dp_distinguished_only);
                                half_kernel_vals[j] = kernel.calculate(num_graphs)[0];
                            }   break;
                        case RSP:
                            {
                                ClusterGraphLoader<sp_graph_type> loader1;
                                loader1.load(tempVec, format, dc, clopts);
                                HalfShortestPathKernel kernel(*loader1.get_graph(), dc, engine_type, clopts, k, rsp_binning_type, rsp_num_bins, inv_edge_lengths, log_edge_lengths);
                                half_kernel_vals[j] = kernel.calculate(num_graphs)[0];
                            }   break;
                    }
                }
                muter.cout() << std::endl;
            }
            muter.restore();
            
            // Resize
            size_t maxsize = 0;
            for(size_t i=0;i<num_graphs;i++) {
                maxsize = std::max(maxsize, half_kernel_vals[i].size());
            }
            for(size_t i=0;i<num_graphs;i++) {
                half_kernel_vals[i].conservativeResize(maxsize);
            }
            dc.cout() << "Resized half-kernels to size " << maxsize << std::endl;

        } else {
            throw MtltQuitException("Unknown precomputation mode: " + precompute_mode);
        }

        // Normalize
        if(normalize_half_kernels) {
            for(size_t i=0;i<num_graphs;i++) {
                const double norm = half_kernel_vals[i].norm();
                if(norm != 0)
                    half_kernel_vals[i] = half_kernel_vals[i] / norm;
            }
        }

        if(dc.procid() == 0) {
            muter.cout() << std::endl;
            muter.cout() << "Computing kernel matrix." << std::endl;
        }
        // The kernel matrix is symmetric, so only compute the upper diagonal.
        for(i=0; i<num_graphs; i++) {
            for(j=i+1; j<num_graphs; j++) {

                // Create the real kernel value by taking the inner product of the half-kernels.
                kernel_value = half_kernel_vals[i].dot(half_kernel_vals[j]);
                // Store values in svm_matrix
                store_svm_value(svm_matrix, i, j, kernel_value);
            }
        }

        // Fill diagonal with 1.0
        for(i=0; i<num_graphs; i++) {
            svm_matrix[i][i] = 1.0;
        }

        dc.cout() << "Kernel precomputation complete! (" << timer.current_time() << " secs)" << std::endl;

        // Restore old cout and cerr.
        muter.restore();
    }

    void write_outputfile(std::string svmlight_outputfile, int num_graphs,
            const std::vector<int>& graph_labels, const std::vector<std::vector<double> >& svm_matrix,
            graphlab::distributed_control& dc) {

        dc.cout() << "Writing precomputed kernel data to \"" << svmlight_outputfile << "\"." << std::endl;
        std::ofstream outfile(svmlight_outputfile.c_str());
        if(outfile.is_open()) {
            for(int i=0;i<num_graphs;i++) {
                for(int j=0;j<num_graphs+2;j++)
                    if(j==0) // label
                        outfile << graph_labels[i] << " ";
                    else if(j==1) // index
                        outfile << "0:" << i+1 << " ";
                    else if(svm_matrix[i][j-2] != 0)
                        outfile << j-1 << ":" << svm_matrix[i][j-2] << " ";
                outfile << std::endl;
            }

            outfile.close();
        }
    }

    void write_halfkernels_outputfile(std::string halfkernels_outputfile, int num_graphs,
            const std::vector<int>& graph_labels, const std::vector<SparseVector>& half_kernels,
            graphlab::distributed_control& dc) {

        dc.cout() << "Writing half-kernel data to \"" << halfkernels_outputfile << "\"." << std::endl;
        size_t dimension = half_kernels[0].size();
        std::ofstream outfile(halfkernels_outputfile.c_str());
        if(outfile.is_open()) {
            for(int i=0;i<num_graphs;i++) {
                for(int j=0;j<dimension+1;j++)
                    if(j==0) // label
                        outfile << graph_labels[i] << " ";
                    else if(!half_kernels[i].isZero()) // half_kernels
                        outfile << j << ":" << half_kernels[i][j-1] << " ";
                outfile << std::endl;
            }

            outfile.close();
        }
    }

    void read_names(std::string example_names_file, std::vector<std::string>& names) {
        
        std::ifstream file(example_names_file.c_str());
        if (file.is_open()) {
            std::string line;
            while (std::getline(file, line)) {

                if(line[0] == '#') // Skip comments
                    continue;

                names.push_back(line);
            }
            file.close();

            std::cout << "Read example_names_file successfully." << std::endl;
        }
    }

    void read_names_frequency(std::map<std::string, double>& last_name_frequency, std::map<std::string, double>& male_name_frequency, std::map<std::string, double>& female_name_frequency)
    {

	
	//open file hardcoded that the frequency files should be available
	//last names
	std::ifstream file("../dataset/name_data/dist.all.last");
	if(file.is_open()) {
	    std::string line;
	    while(std::getline(file,line)) {
		if(line[0] == '#') // Skip comments
		    continue;


		std::string buf; //buffer string
		std::stringstream ss(line);
		std::vector<std::string> words;

		while(ss >> buf)
		    words.push_back(buf);
		
		std::string name = words.at(0);
		boost::algorithm::to_lower(name);
		std::string sfreq = words.at(1);
		double freq = ::atof(sfreq.c_str());
		last_name_frequency[name] = freq;
	    }
	    file.close();
	    
	}

	//male names
	std::ifstream file2("../dataset/name_data/dist.male.first");
	if(file2.is_open()) {
	    std::string line;
	    while(std::getline(file2,line)) {
		if(line[0] == '#') // Skip comments
		    continue;


		std::string buf; //buffer string
		std::stringstream ss(line);
		std::vector<std::string> words;

		while(ss >> buf)
		    words.push_back(buf);
		
		std::string name = words.at(0);	
		boost::algorithm::to_lower(name);
		std::string sfreq = words.at(1); 
		double freq = ::atof(sfreq.c_str());
		male_name_frequency[name] = freq;
	    }
	    file2.close();
	    
	}

	//female names
	std::ifstream file3("../dataset/name_data/dist.female.first");
	if(file3.is_open()) {
	    std::string line;
	    while(std::getline(file3,line)) {
		if(line[0] == '#') // Skip comments
		    continue;


		std::string buf; //buffer string
		std::stringstream ss(line);
		std::vector<std::string> words;

		while(ss >> buf)
		    words.push_back(buf);
		
		std::string name = words.at(0);
		boost::algorithm::to_lower(name);
		std::string sfreq = words.at(1);
		double freq = ::atof(sfreq.c_str());
		female_name_frequency[name] = freq;
	    }
	    file3.close();
	    
	}
	std::cout << "Read name frequencies successfully." << std::endl;
    }

    void read_inputfile(
                // out
                std::vector<int>& graph_labels, std::vector<std::vector<double> >& svm_matrix,
                std::vector<SparseVector>& half_kernel_vals, int& num_graphs,
                // in
                std::string svmlight_inputfile) {


        std::ifstream file(svmlight_inputfile.c_str());
        if (file.is_open()) {
            std::string line;
            size_t i=0;
            while (std::getline(file, line)) {

                if(line[0] == '#') // Skip comments
                    continue;

                // Replace : with whitespace.
                int n=0;
                for (size_t pos=0; pos < line.size(); ++pos)
                    if (line[pos] == ':') {
                        n++; line[pos] = ' ';
                    }

                std::istringstream iss(line);
                size_t label;
                iss >> label;
                if(label != 1 && label != -1) {
                    throw MtltQuitException("Error: Inputfile labels must be either 1 or -1. Found label: " + label);
                }
                graph_labels.push_back(label);

                size_t j;
                double val;
                SparseVector x_val(n);
                // Read rest of line
                while(iss >> j >> val) {    // the index j stored in the file is 1-based
                    x_val(j-1) = val;
                }
                half_kernel_vals.push_back(x_val);

                i++;
            }
            file.close();

            std::cout << "Read svmlight_inputfile successfully." << std::endl;
        }

        num_graphs = half_kernel_vals.size();
        // each row contains [y1, k11, k12, k13, ...]
        svm_matrix = std::vector<std::vector<double> >();
        for(int i=0;i<num_graphs;i++)
            svm_matrix.push_back(std::vector<double>(num_graphs));

        // The kernel matrix is symmetric, so only compute the upper diagonal.
        for(size_t i=0; i<num_graphs; i++) {
            for(size_t j=i+1; j<num_graphs; j++) {

                // Create the real kernel value by taking the inner product of the half-kernels.
                double kernel_value = half_kernel_vals[i].dot(half_kernel_vals[j]);
                // Store values in svm_matrix
                store_svm_value(svm_matrix, i, j, kernel_value);
            }
        }

        // Fill diagonal with 1.0
        for(size_t i=0; i<num_graphs; i++) {
            svm_matrix[i][i] = 1.0;
        }
    }


    /*
     * \brief Adds lines from svm_matrix to dataset.
     */
    void add_examples(std::vector<std::vector<std::vector<double> > >& dataset, 
            int begin_index, int end_index,
            const std::vector<std::vector<double> >& svm_matrix,
            const std::vector<int>& graph_labels) {

                // add train set examples
                // num_val .. num_graphs-1
                std::vector<std::vector<double > > fold;
                for(int i=begin_index;i<end_index;i++) {
                    
                    std::vector<double> row_vec;
                    // Add label
                    row_vec.push_back(graph_labels.at(i));
                    // Add row
                    for(int j=begin_index;j<end_index;j++) {
                        row_vec.push_back(svm_matrix.at(i).at(j));
                    }
                    fold.push_back(row_vec);
                }
                
                dataset.push_back(fold);
    }

    /*
     * \brief Shuffles the svm_matrix so that rows and cols are permutated
     * identically. Since the labels also need to be updated when the svm_matrix
     * is changed, it is taken care of here as well.
     */
    void symmetric_shuffle_matrix(std::vector<std::vector<double> >& svm_matrix,
                    std::vector<int>& graph_labels) {

        //Use the same seed so that both the colums and the rows are shuffled the same 
        unsigned int seed = time(NULL);

        // Shuffle rows
        srand(seed);
        std::random_shuffle(svm_matrix.begin(), svm_matrix.end(), myrandom);
        // Shuffle labels
        srand(seed);
        std::random_shuffle(graph_labels.begin(), graph_labels.end(), myrandom);
        // Shuffle cols
        for(std::vector<std::vector<double> >::iterator it=svm_matrix.begin();it!=svm_matrix.end();it++) {
            srand(seed);
            std::random_shuffle(it->begin(), it->end(), myrandom);
        }
    }

    /*
     * \brief Shuffles the half-kernels and labels in sync.
     */
    void symmetric_shuffle_half_kernels(unsigned int seed, std::vector<SparseVector>& half_kernel_vals,
                    std::vector<int>& graph_labels, std::vector<std::string>& names) {

        // Shuffle half-kernels
        srand(seed);
        std::random_shuffle(half_kernel_vals.begin(), half_kernel_vals.end(), myrandom);
        // Shuffle labels
        srand(seed);
        std::random_shuffle(graph_labels.begin(), graph_labels.end(), myrandom);
        // Shuffle names
        srand(seed);
        std::random_shuffle(names.begin(), names.end(), myrandom);
    }

    void start(int argc, char** argv) {
        // Distributed communication layer which is needed by a number
        // of the core GraphLab objects, whether you are running
        // distributed or not.
        graphlab::distributed_control dc;
        //unsigned int seed = time(NULL);
        unsigned int seed = 1234;
        srand(seed);
        std::cout << "[" << dc.procid() << "] Using random seed: " << seed << std::endl;
        dc.cout() << "numprocs: " << dc.numprocs() << std::endl;
        graphlab::timer program_timer;
        program_timer.start();

        /*
         *  When running in training mode, a trainfile should be specified in the following format,
         *  with the label file on the first line, followed by graph files.
         *
         *  Example with 100 graphs:
         *
         *  $ cat trainfile.txt
         *  labels.txt              // Contains files with 100 y-values (0 or 1, see below)
         *  graph1.txt              // Path to graph file in GraphLab snap format (GraphChi edgelist format)
         *  graph2.txt
         *  graph3.txt
         *  .
         *  .
         *  .
         *  graph100.txt
         *
         *  Where labels.txt has the following format:
         *  $ cat labels.txt
         *  1
         *  0
         *  1
         *  1
         *  .
         *  .
         *  .
         *  0
         *  1
         *
         *  The number of lines in labels.txt and the number of graphs specified must match.
         *
         */

        std::string trainfile, svmlight_outputfile, halfkernels_outputfile, svmlight_inputfile, example_names_file, engine_type = "sync", format = "snap";
	std::map<std::string,double> last_name_frequency;
	std::map<std::string,double> male_name_frequency;
	std::map<std::string,double> female_name_frequency;
        double lambda = 0.6;
        bool dp_distinguished_only = false;
        int k = 10000;
        std::string rsp_binning_type = "lin";
        size_t rsp_num_bins = 10;
        int num_folds = 10;
        bool cv_just_once = false;
        bool no_cv = false;
        KernelType kerneltype;
	bool use_names = true;
        std::string precompute_mode = "batch";
        bool inv_edge_lengths = false;
        bool log_edge_lengths = false;
        double pegasos_lambda = 0.001;
        int pegasos_max_iter = int(100/pegasos_lambda);
        int pegasos_exam_per_iter = 1;
        std::string pegasos_impl = "half_kernel";
        bool normalize_half_kernels = true;

        // Parse command line options
        graphlab::command_line_options clopts("train_svm");
        parse_cmd_opts(trainfile, engine_type, svmlight_outputfile, halfkernels_outputfile, svmlight_inputfile, example_names_file, use_names,kerneltype, precompute_mode, format, lambda, dp_distinguished_only, k, rsp_binning_type, rsp_num_bins, inv_edge_lengths, log_edge_lengths, normalize_half_kernels,
                num_folds, cv_just_once, no_cv, pegasos_lambda, pegasos_max_iter, pegasos_exam_per_iter, pegasos_impl,
                dc, argc, argv, clopts);

        omp_set_num_threads(clopts.get_ncpus());
        dc.cout() << "omp_set_num_threads(" << clopts.get_ncpus() << ")" << std::endl;
        // Create Pegasos interface
        PegasosInterface* pegasosInterface;
        if(pegasos_impl == "half_kernel")
            pegasosInterface = new PegasosHalfKernelSingle(pegasos_lambda, pegasos_max_iter, pegasos_exam_per_iter, dc, clopts, engine_type);
        else if(pegasos_impl == "naive_kernel")
            pegasosInterface = new PegasosNaiveKernelSingle(pegasos_lambda, pegasos_max_iter, pegasos_exam_per_iter, dc, clopts, engine_type);
        else if(pegasos_impl == "half_kernel_active_learning")
            pegasosInterface = new PegasosHalfKernelActiveLearning(pegasos_lambda, pegasos_max_iter, pegasos_exam_per_iter, dc, clopts, engine_type);
        else
            throw MtltQuitException("Unknown Pegasos implementation type specified: " + pegasos_impl);

        std::vector<int> graph_labels;
        std::vector<std::vector<double> > svm_matrix;
        std::vector<SparseVector> half_kernel_vals;
        std::vector<std::string> names;
        int num_graphs;

        if(svmlight_inputfile == "") {
            precompute_kernels(graph_labels, svm_matrix, half_kernel_vals, num_graphs,
                    kerneltype, precompute_mode, trainfile, engine_type, lambda, dp_distinguished_only, k, rsp_binning_type, rsp_num_bins, inv_edge_lengths, log_edge_lengths, normalize_half_kernels,
                    dc, clopts, format);
        } else {
            // Read kernel values from file
            read_inputfile(graph_labels, svm_matrix, half_kernel_vals, num_graphs, 
                    svmlight_inputfile);
        }

        if(example_names_file != "") {
            // Read names for the examples in the trainfile
            read_names(example_names_file, names);
            if(names.size() != half_kernel_vals.size())
                throw new MtltQuitException("The number of names in example_names_file and the number of training examples does not match.");
	


	    //If we should use the names to predict
	    if(use_names)
	    {
		//Insert bool if we should use freq
		read_names_frequency(last_name_frequency, male_name_frequency, female_name_frequency);
		//Extend the vector to include name frequency
		int vec_index = 0;
	
	      /*for( std::map<std::string,double>::iterator ii=male_name_frequency.begin(); ii!=male_name_frequency.end(); ++ii)
		{
		    std::cout << (*ii).first << ": " <<(*ii).second << std::endl; 
		} */

		for(std::vector<std::string>::iterator it=names.begin(); it!=names.end(); it++)
		{
		    std::string name = *it;
		    boost::algorithm::to_lower(name); 
		
		    std::string buf; //buffer string
		    std::stringstream ss(name);
		    std::vector<std::string> words;
		    while(ss >> buf)
			words.push_back(buf);
		
		    double freq = 0.0;
		    double freq2 = 0.0;

		    if(words.size() == 1)
		    {
			std::string first_name = words.at(0);
			if(male_name_frequency.count(first_name) == 1)
			    freq += male_name_frequency.at(first_name);
			if(female_name_frequency.count(first_name) == 1)
			    freq += female_name_frequency.at(first_name);
			else
			        freq = 0.0;
			freq2 = 0.0;

			
	    	    }
		    else if(words.size() == 2)
		    {
			std::string first_name = words.at(0);
			std::string last_name = words.at(1);
			if(male_name_frequency.count(first_name) == 1)
			    freq += male_name_frequency.at(first_name);
			if(female_name_frequency.count(first_name) == 1)
			    freq += female_name_frequency.at(first_name);
			else
			    freq = 0.0;

			if(last_name_frequency.count(last_name) == 1)
			    freq2 = last_name_frequency.at(last_name);
			else
			    freq2 = 0.0;
		    }
		    //More than 2 names for one person.
		    //In this case set the frequency of the names to 0 since it means that 
		    //we have at least one middle name (should be an uncommon name and unambiguious entity)
		    //Note that this is not a clean solution supported by theory, but rather a quick fix.
		    else
		    {
			freq = 0.0;
			freq2 = 0.0;
		    }
		    SparseVector& sv =  half_kernel_vals[vec_index];

		    /*for(int i=0; i<sv.size();i++)
		    {
			dc.cout() << sv[i] << " old vector" << std::endl; 
		    }*/
		    
		    sv.conservativeResize(sv.size()+2);
		    /*for(int i=0; i<sv.size();i++)
		    {
			dc.cout() << sv[i] << " new size vector" << std::endl; 
		    }*/

		    sv[sv.size()-2] = freq;
		    sv[sv.size()-1] = freq2;
		    /*for(int i=0; i<sv.size();i++)
		    {
			dc.cout() << sv[i] << " complete vector" << std::endl; 
		    }*/

		    SparseVector sv2 = half_kernel_vals[vec_index];
		    /*for(int i=0; i<sv2.size();i++)
		    {
			dc.cout() << sv2[i] << " sent vector" << std::endl; 
		    }*/

		    vec_index++;	
		}
	    }
   	
	}






        if(svmlight_outputfile != "") {
            write_outputfile(svmlight_outputfile, num_graphs, graph_labels, svm_matrix, dc);
        }
        if(halfkernels_outputfile != "") {
            write_halfkernels_outputfile(halfkernels_outputfile, num_graphs, graph_labels, half_kernel_vals, dc);
        }

        if(!no_cv) {
            double avg_val_acc = 0,
                   min_val_acc =  std::numeric_limits<double>::max(), 
                   max_val_acc = -std::numeric_limits<double>::max();

            int num_train = (1.0-(1.0/num_folds)) * num_graphs;
            int num_val = num_graphs - num_train;
            dc.cout() << "Starting " << num_folds << "-fold cross-validation phase." << std::endl;
            graphlab::timer timer;
            timer.start();
            Classification tot_c;

            dc.cout() << "The names are:" << std::endl;
            for(std::vector<std::string>::iterator it=names.begin();it!=names.end();it++)
                dc.cout() << *it << std::endl;

            for(int val_index=0;val_index<num_folds;val_index++) {

                // Shuffle data
                symmetric_shuffle_half_kernels(seed, half_kernel_vals, graph_labels, names);

                // Setup iterators
                std::vector<SparseVector>::iterator border = half_kernel_vals.begin();
                std::advance(border, num_val);
                std::vector<int>::iterator border_labels = graph_labels.begin();
                std::advance(border_labels, num_val);

                // add val set examples
                // 0 .. num_val-1
                std::vector<SparseVector> valset(half_kernel_vals.begin(), border);
                std::vector<int> valset_labels(graph_labels.begin(), border_labels);

                // add train set examples
                // num_val .. num_graphs-1
                std::vector<SparseVector> trainset(border, half_kernel_vals.end());
                std::vector<int> trainset_labels(border_labels, graph_labels.end());

                dc.cout() << val_index+1 << " of " << num_folds << std::endl;
                dc.cout() << "trainset (labels): " << trainset.size() << " (" << trainset_labels.size() << ")" << std::endl;
                dc.cout() << "valset (labels): "   <<   valset.size() << " (" <<   valset_labels.size() << ")" << std::endl;

                // Call Pegasos
                Classification c = pegasosInterface->train(trainset, valset, trainset_labels, valset_labels);

                if(names.size() > 9) {
                    // Print rankings
                    std::vector<std::pair<size_t, double> > ranks;
                    size_t rank_pos;

                    // Pos ranks
                    ranks = c.get_ranked_pos_predictions();
                    rank_pos = 1;
                    dc.cout() << " ============================ " << std::endl;
                    dc.cout() << "        - Pos. ranks -        " << std::endl;
                    dc.cout() << " ============================ " << std::endl;
                    for(std::vector<std::pair<size_t, double> >::iterator it=ranks.begin();it!=ranks.end();it++)
                        dc.cout() << rank_pos++ << ":\t\"" << names.at(it->first) << "\"\t\t" << it->second << "\t\t" << graph_labels.at(it->first) << std::endl;

                    // Neg ranks
                    ranks = c.get_ranked_neg_predictions();
                    rank_pos = 1;
                    dc.cout() << " ============================ " << std::endl;
                    dc.cout() << "        - Neg. ranks -        " << std::endl;
                    dc.cout() << " ============================ " << std::endl;
                    for(std::vector<std::pair<size_t, double> >::iterator it=ranks.begin();it!=ranks.end();it++)
                        dc.cout() << rank_pos++ << ":\t\"" << names.at(it->first) << "\"\t\t" << it->second << "\t\t" << graph_labels.at(it->first) << std::endl;
                }

                tot_c += c;
                avg_val_acc += c.get_accuracy();
                min_val_acc = std::min(min_val_acc, c.get_accuracy());
                max_val_acc = std::max(max_val_acc, c.get_accuracy());

                if(cv_just_once)
                    break;
            }

            avg_val_acc /= num_folds;

            double f1_div = tot_c.get_precision() + tot_c.get_recall();
            double f1_score = 0;
            if(f1_div != 0)
                f1_score = 2 * tot_c.get_precision() * tot_c.get_recall() / f1_div;

            dc.cout() << "CV complete! (" << timer.current_time() << " secs) Val. acc.:\nAvg: " << avg_val_acc << "\nMin: " << min_val_acc << "\nMax: " << max_val_acc << std::endl;
            dc.cout() << std::endl << std::setiosflags(std::ios::left) <<
                std::setw(15) << "Accuracy: "       << std::setiosflags(std::ios::left) << std::setw(10) << tot_c.get_accuracy()     << std::endl <<
                std::setw(15) << "Precision: "      << std::setiosflags(std::ios::left) << std::setw(10) << tot_c.get_precision()    << std::endl <<
                std::setw(15) << "Recall: "         << std::setiosflags(std::ios::left) << std::setw(10) << tot_c.get_recall()       << "\tAbility to identify positives (multi-persons)" << std::endl <<
                std::setw(15) << "Specificity: "    << std::setiosflags(std::ios::left) << std::setw(10) << tot_c.get_specificity()  << "\tAbility to identify negatives (single persons)" << std::endl <<
                std::setw(15) << "NPV: "            << std::setiosflags(std::ios::left) << std::setw(10) << tot_c.get_negative_predictive_value()  << std::endl <<
                std::setw(15) << "F1 score: "       << std::setiosflags(std::ios::left) << std::setw(10) << f1_score  << std::endl <<
                std::endl;

            dc.cout() << "Avg. train time: " << tot_c.get_avg_train_time_millis() << " ms" << std::endl;

        }

        dc.cout() << "[" << dc.procid() << "] Using random seed: " << seed << std::endl;
        dc.cout() << "svm_train program time: " << program_timer.current_time() << " secs." << std::endl;

        SAFE_DELETE(pegasosInterface);

    }

    private:
};

#endif
